!  shnmsim
!  
!  Copyright 2022 Igor J. Chybicki <igorchy@ukw.edu.pl>
!
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  (at your option) any later version.
!  
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!  
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
!  MA 02110-1301, USA.
!  
!  The code has been successfully compiled using GCC compiler GFortran
!  10.2.0 using the command:
!  gfortran -Wall -O3 -ftree-vectorize -o shnmsim shnmsim.f90
!  
!  To install GFortran, see https://gcc.gnu.org/wiki/GFortranBinaries

program shnmsim
   
   implicit none
   
   integer :: ii,ni
   integer :: np,no,nl,na,nz,nh
   double precision :: pden,oden,piconst
   integer, allocatable :: pgen(:,:,:),ogen(:,:,:),phap(:),ohap(:)
   double precision, allocatable :: px(:),py(:),pz(:,:),ox(:),oy(:),mlam(:),flam(:),mbeta(:),fbeta(:)
   double precision :: ms,mp,sp,msig,fsig,ds,dp,bs,bp
   character(1) :: tab = char(9)
   character(len=256) :: fn
   integer, dimension(8) :: values
   integer*8 :: seed,jsr
   
   call date_and_time(values=values)
   piconst=dacos(dble(-1.0))
   
   !RNG initialization
   seed=-1
   if (seed<0) seed=values(5)*3600+values(6)*60+values(7)
   jsr=seed
   
   np=100              !number of parents
   no=300              !number of progeny
   nl=16               !number of genetic markers
   na=8                !number of alleles per marker (uniformly distributed)
   nz=3                !number of phenotypic characters
   nh=1                !number of haplotypes (maternally inherited; if zero, each parent has their own haplotype)
   pden=20.D0          !density (ind/ha) of parental population
   oden=60.D0          !density (ind/ha) of progeny population
   fsig=0.5D0          !standard deviation for female log-fitness
   msig=0.5D0          !standard deviation for male log-fitness
   allocate(fbeta(nz)) !selection gradients for female fitness:
   fbeta(1)=1.D0
   fbeta(2)=-0.5D0
   fbeta(3)=0.D0
   allocate(mbeta(nz)) !selection gradients for male fitness:
   mbeta(1)=1.D0
   mbeta(2)=-0.5D0
   mbeta(3)=0.D0
   ms=0.05D0           !seed immigration
   mp=0.1D0            !pollen immigration
   sp=0.025D0          !selfing
   ds=25.D0            !average forward dispersal distance (seeds)
   dp=50.D0            !average forward dispersal distance (pollen)
   bs=1.D0             !shape of the eponential-power dispersal kernel (seeds)
   bp=0.5D0            !shape of the eponential-power dispersal kernel (pollen)
   
   ni=25
   do ii=1,ni
     seed=jsr
     write (fn,"('nmpi_Simulation',i0,'.txt')") ii
     call alloc_mem
     call parents
     call progeny
     call save_data
     write (*,"('Replicate ',i0,' completed')") ii
   enddo
   
 contains
 
 !----------------------------------------------------------------------
 subroutine alloc_mem
   if (allocated(px)) deallocate(px)
   if (allocated(py)) deallocate(py)
   if (allocated(pz)) deallocate(pz)
   if (allocated(ox)) deallocate(ox)
   if (allocated(oy)) deallocate(oy)
   if (allocated(pgen)) deallocate(pgen)
   if (allocated(ogen)) deallocate(ogen)
   if (allocated(mlam)) deallocate(mlam)
   if (allocated(flam)) deallocate(flam)
   if (allocated(phap)) deallocate(phap)
   if (allocated(ohap)) deallocate(ohap)
   
   allocate(px(np))
   allocate(py(np))
   allocate(pz(np,nz))
   allocate(pgen(np,nl,2))
   allocate(phap(np))
   allocate(mlam(np))
   allocate(flam(np))
   allocate(ox(no))
   allocate(oy(no))
   allocate(ogen(no,nl,2))
   allocate(ohap(no))
 end subroutine
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine parents
   integer :: i,l
   double precision :: rad,t,r,m(nz),s(nz)
   
   !parental locations, phenotypes and genotypes
   rad=100.D0*dsqrt(dble(np)/pden/piconst)
   m=0.D0
   s=0.D0
   do i=1,np
     t=rnd()*2.D0*piconst  !random direction from the population center
     r=dsqrt(rnd())*rad    !random distance from the population center
     px(i)=dcos(t)*r       !Cartesian xy coordinates within a circle
     py(i)=dsin(t)*r
     do l=1,nz
       pz(i,l)=normaldev(0.D0,1.D0) !observed phenotype drawn from the Normal distrib.
       m(l)=m(l)+pz(i,l)
     enddo
     do l=1,nl
       pgen(i,l,1)=int(rnd()*na)+1 !observed genotype
       pgen(i,l,2)=int(rnd()*na)+1
     enddo
     phap(i)=i
     if (nh>0) phap(i)=int(rnd()*nh)+1 !observed chloroplast haplotype
   enddo
   m=m/dble(np)
   do i=1,np
     do l=1,nz
       s(l)=s(l)+(pz(i,l)-m(l))*(pz(i,l)-m(l))
     enddo
   enddo
   s=dsqrt(s/dble(np))
   
   !parental fecundities
   do i=1,np
     flam(i)=0.D0
     mlam(i)=0.D0
     do l=1,nz
       pz(i,l)=(pz(i,l)-m(l))/s(l)      !phenotype standardization
       flam(i)=flam(i)+fbeta(l)*pz(i,l) !expected female log-facundity
       mlam(i)=mlam(i)+mbeta(l)*pz(i,l) !expected male log-facundity
     enddo
     if (fsig>0.D0) flam(i)=normaldev(flam(i),fsig) !observed female log-facundity drawn from the normal distrib.
     if (msig>0.D0) mlam(i)=normaldev(mlam(i),msig) !observed male log-facundity drawn from the normal distrib.
   enddo
 end subroutine
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine progeny
   integer :: i,j,l,om,of
   double precision :: psi(np),phi(np)
   double precision :: rad,t,r,as,ap,spsi,sphi
   
   !scale of the exponential-power dispersal kernel
   as=dexp(dlog(ds)+dlgama(dble(2)/bs)-dlgama(dble(3)/bs))
   ap=dexp(dlog(dp)+dlgama(dble(2)/bp)-dlgama(dble(3)/bp))
   
   !progeny locations
   rad=100.D0*dsqrt(dble(no)/oden/piconst)
   do i=1,no
     t=rnd()*2.D0*piconst  !random direction from the population center
     r=dsqrt(rnd())*rad    !random distance from the population center
     ox(i)=dcos(t)*r
     oy(i)=dsin(t)*r
   enddo
   
   !progeny genealogies
   do i=1,no
     om=-1
     of=-1
     if (rnd()>=ms) then
       spsi=0.D0
       do j=1,np
         t=hypot(ox(i)-px(j),oy(i)-py(j))
         psi(j)=dexp(-((t/as)**bs))*dexp(flam(j)) !seed dispersal kernel x female fecundity
         spsi=spsi+psi(j)
       enddo
       r=rnd()
       t=0.D0
       do j=1,np
         t=t+psi(j)/spsi
         if (r<t) then
           om=j
           exit
         endif
       enddo
       
       r=rnd()
       if (r<sp) then
         of=om
       else if (r<1.D0-mp) then
         sphi=0.D0
         do j=1,np
           phi(j)=0.D0
           if (j==om) cycle
           t=hypot(px(om)-px(j),py(om)-py(j))
           phi(j)=dexp(-((t/ap)**bp))*dexp(mlam(j)) !pollen dispersal kernel x male fecundity
           sphi=sphi+phi(j)
         enddo
         r=rnd()
         t=0.D0
         do j=1,np
           t=t+phi(j)/sphi
           if (r<t) then
             of=j
             exit
           endif
         enddo
       endif
     endif
     
     !progeny genotypes
     do l=1,nl
       if (om>-1) then
         ogen(i,l,1)=pgen(om,l,int(rnd()*2)+1)
       else
         ogen(i,l,1)=int(rnd()*na)+1
       endif
       if (of>-1) then
         ogen(i,l,2)=pgen(of,l,int(rnd()*2)+1)
       else
         ogen(i,l,2)=int(rnd()*na)+1
       endif
     enddo
     ohap(i)=-1
     if (om>-1) then
       ohap(i)=phap(om)
     else
       if (nh>0) ohap(i)=int(rnd()*nh)+1
     endif
   enddo
 end subroutine
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 subroutine save_data
   integer :: i,j,ofil
   character(len=10) :: ind
   
   ofil=123
   open (unit=ofil,file=trim(fn),action='write',status='replace')
   write (ofil,*) np,tab,no,tab,nl,tab,nz
   do i=1,np
     write (ofil,*) 0,tab,i,tab,px(i),tab,py(i),tab,phap(i),(tab,pgen(i,j,1),tab,pgen(i,j,2),j=1,nl),(tab,pz(i,j),j=1,nz),tab,0.5
   enddo
   do i=1,no
     write (ofil,*) 1,tab,i,tab,ox(i),tab,oy(i),tab,ohap(i),(tab,ogen(i,j,1),tab,ogen(i,j,2),j=1,nl),tab,-1,tab,-1,tab,1
   enddo
   write (ofil,*) 'np',tab,np
   write (ofil,*) 'no',tab,no
   write (ofil,*) 'nl',tab,nl
   write (ofil,*) 'na',tab,na
   write (ofil,*) 'nz',tab,nz
   write (ofil,*) 'pden',tab,pden
   write (ofil,*) 'oden',tab,oden
   write (ofil,*) 'fsig',tab,fsig
   write (ofil,*) 'msig',tab,msig
   write (ofil,*) 'fbeta',(tab,fbeta(i),i=1,nz)
   write (ofil,*) 'mbeta',(tab,mbeta(i),i=1,nz)
   write (ofil,*) 'ms',tab,ms
   write (ofil,*) 'mp',tab,mp
   write (ofil,*) 'sp',tab,sp
   write (ofil,*) 'ds',tab,ds
   write (ofil,*) 'dp',tab,dp
   write (ofil,*) 'bs',tab,bs
   write (ofil,*) 'bp',tab,bp
   write (ofil,*) 'seed',tab,seed
   do i=1,np
      write (ind,"(i0)") i
      write (ofil,*) 'lambda'//trim(ind),tab,flam(i),tab,mlam(i)
   enddo
   close (ofil)
 end subroutine
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 function shift_xor(val,shift)
   integer*8 :: shift_xor
   integer*8, intent(in) :: val, shift
   
   shift_xor = ieor(val,ishft(val,shift))
 end function
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function rnd()
   jsr = shift_xor(shift_xor(shift_xor(jsr, 13_8),-17_8),43_8)
   rnd = jsr * 2685821657736338717_8 * 5.42101086242752217D-20 + 0.5D0
 end function rnd
 !----------------------------------------------------------------------
 
 !----------------------------------------------------------------------
 double precision function normaldev(m,s)
   double precision, intent(in) :: m,s
   double precision :: v1,v2,r
   
   do
     v1 = dble(2) * rnd() - dble(1)
     v2 = dble(2) * rnd() - dble(1)
     r  = v1**2 + v2**2
     if (r < dble(1) .and. r > dble(0)) exit
   enddo
   normaldev = v2 * dsqrt(-dble(2) * dlog(r) / r) * s + m
   
 end function normaldev
 !----------------------------------------------------------------------
 
end program
