!  nmpi2.f90
!  
!  Copyright 2016-2022 Igor J. Chybicki <igorchy@ukw.edu.pl>
!  
!  NMpi 2.0 implements estimators of parameters
!  of the neighbourhood model (Burczyk et al. 2006).
!  ML estimates are obtained using the Newton method, with the
!  gradient and the Hessian being computed numerically. In consequence,
!  the implemented model is quite flexible in terms of possible
!  modifications, however at cost of somewhat sub-optimal estimation time 
!  and stability. Bayesian estimates are obtained using a combination
!  of the Metropolis-Hastings and the Gibbs sampler (Chybicki et al. 2021)
!
!  This program is free software; you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation; either version 2 of the License, or
!  (at your option) any later version.
!  
!  This program is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!  
!  You should have received a copy of the GNU General Public License
!  along with this program; if not, write to the Free Software
!  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
!  MA 02110-1301, USA.
!  
!  The code has been successfully compiled using GCC compiler GFortran
!  10.2.0 using the command:
!  gfortran -Wall -O3 -ftree-vectorize -fopenmp -o nmpi2 nmpi2.f90
!  
!  To install GFortran, see https://gcc.gnu.org/wiki/GFortranBinaries
!
!  TODO (the order does not reflect priority):
!  1. constant transition probabilities if mistyping errors are constant
!     (should increase analysis speed significantly)
!     this would require global variables:
!     c_pr_b(:) (dim=no)
!     c_pr_mb(:,:) (dim=no,np)
!     c_pr_mm(:,:) (dim=no,np)
!     c_pr_mf(:,:,:) (dim=no,np,np)
!     const_probs as logical enables/disables this mode
!     Initialization includes a subroutine to compute constant probs
!  2. dispersal distance treated as a random variable, dependent on
!     individual characteristics (e.g. location, phentype): implemented as
!     either a fixed-effects model (ML) or random-effects model (Bayes).
!     d_j = d_0 * exp(c_j), where d_0 is a base dispersal distance and
!     c_j is a coefficient for the j-th adult. For the fixed-effects model,
!     c_j = w_j1 * x_j1 + ... + x_jZ * x_jZ. For the random-effects model,
!     c_j ~ N(u_j, s), where u_j = w_j1 * x_j1 + ... + x_jZ * x_jZ, and
!     s is the std.dev. of the normal distribution.
!     Alternative implementation: d_j = d_0 / (1+exp(-c_j))), where d_0 is
!     the maximum threshold mean distance, and c_j is a logit.
!  3. separate genetic data for migrant sources plus the appropriate 
!     migration model, e.g. m_s * (m_s1 * Pr(O|B_1) + ... + m_sK * Pr(O|B_K)),
!     where m_s is a 'global' seed immigration rate, m_sk is the seed
!     migration rate from the k-th source, and Pr(O|B_k) is the transition
!     probability for offspring genotype given that it is a migrant from
!     the k-th source. In this model, m_s1 + m_s2 + ... + m_sK = 1.
!     The same model structure would be for pollen immigration.

program nmpi2
   use omp_lib
   
   implicit none
   
   type parents
     integer :: oid
     integer :: mo1,fa1
     double precision :: pr1
     integer :: mo2,fa2
     double precision :: pr2
     double precision :: prms,prnms
   end type
   
   type(parents), allocatable :: parass(:)
   
   ! global variables
   logical :: preprocess_phenotypes
   character(len=10) :: LASTUPDATE,VERSION
   character(len=1) :: tab = char(9)
   integer :: maxnth,usernth,nincl,lackconv,maxlackconv
   integer :: i,j,maxit,maxpit,nit,outfile,fh,autostart,autostop,writecov
   !integer :: write_pcdf,write_scdf
   double precision :: mtpl,old_logl,new_logl,mep,my2pi,mypi
   integer, dimension(8) :: values
   character(len=256) :: fi,fd,fo,fe,fp,fg,fm,fmf,fmm  ! file names
   real(8) :: tim0,tim1                                ! start and stop time
   integer :: dist_typ                                 ! if dist_typ =1: 1/ds and 1/dp, if dist_typ=0: ds and dp
   
   ! data
   integer :: np,no,nl,nf                              ! number of parents, offspring, loci, phenotypic variables
   integer :: maxl                                     ! maximum number of alleles per locus
   integer, allocatable :: pgen(:,:,:)                 ! genotypes of parents
   integer, allocatable :: ogen(:,:,:)                 ! genotypes of offspring
   integer, allocatable :: odisp(:)                    ! binary indicator whether dispersed or not (1/0)
   integer, allocatable :: allele(:,:)                 ! array of indetified alleles
   integer, allocatable :: nall(:)                     ! number of alleles at the i-th locus
   integer, allocatable :: cyhap(:)                    ! vector of identified haplotypes
   integer :: nhap                                     ! number of haplotypes
   integer, allocatable :: pid(:),oid(:)               ! id numbers
   double precision, allocatable :: px(:),py(:)        ! parental coordinates
   double precision, allocatable :: ox(:),oy(:)        ! offspring coordinates
   integer, allocatable :: om(:),of(:)                 ! known offspring parents
   double precision, allocatable :: pfen(:,:)          ! parental phenotype at multiple characters
   double precision, allocatable :: pzfen(:,:,:)       ! normalized phenotypes for two sexes: pz(0,:,:) - females, pz(1,:,:) males
   double precision, allocatable :: mfen(:,:)          ! phenotypic means for two sexes: m(1,:) - females, m(2,:) males
   double precision, allocatable :: vfen(:,:)          ! phenotypic variances for two sexes: v(1,:) - females, v(2,:) males
   integer, allocatable :: nfen(:,:)                   ! phenotypic counts for two sexes: n(1,:) - females, n(2,:) males
   double precision :: misval                          ! missing phenotype
   integer, allocatable :: pcy(:)                      ! paternal cyDNA haplotype
   integer, allocatable :: ocy(:)                      ! offspring cyDNA haplotype
   double precision :: maternal_cy                     ! prob. of maternal transmission of cyDNA
   double precision, allocatable :: pfem(:)            ! parental femaleness
   ! note that for hermaphrodites (or ambigous cases in dioecious species) pfem(i) should be 0.5
   
   !constant transition probabilities
   double precision, allocatable :: c_pro_b(:),c_pro_mb(:,:),c_pro_m(:,:),c_pro_mf(:,:,:)
   logical :: do_const_probs
   
   ! individual probabilities to be female and male parent for a given seed and mother, resp.
   double precision, allocatable :: ksi(:,:),phi(:,:),sksi(:),sphi(:)
   
   ! parameters of the neighbourhood model
   double precision :: neighborhood                    ! radius of the neighborhood
   integer, allocatable :: onb(:,:),pnb(:,:)           ! binary matrix of neighbors
   double precision, allocatable :: plm(:,:)           ! global nuclear allele frequencies
   double precision, allocatable :: pcm(:)             ! global cyDNA haplotype frequencies
   double precision, allocatable :: mtyp(:)            ! mistyping error rates for nuclear loci
   double precision, allocatable :: vmtyp(:)           ! SE for mistyping rates
   double precision :: mcytyp,vmcytyp                  ! mistyping error rate for cytDNA and SE
   double precision, allocatable :: theta(:),vtheta(:) ! NM-model paramater values and SE
   character(len=10), allocatable :: ntheta(:)         ! NM-model paramater names
   double precision, dimension(7) :: minth,maxth       ! constraints for parameters
   double precision :: minmtyp                         ! minimum mistyping error not different than zero
   double precision :: minmcy                          ! minimum maternal_cy not different than zero
   ! theta(1)=seed immigration (m_s),theta(2)=selfing (s),theta(3)=pollen immigration (m_p),
   ! theta(4)=1/d_s,theta(5)=1/d_p,theta(6)=b_s,theta(7)=b_p
   ! theta(8)=kappa_s,theta(9)=kappa_p,theta(10)=azimuth_s,theta(11)=azimuth_p
   ! theta(12:11+nf)=gamma,theta(12+nf:11+2nf)=beta (for nf characters)
   ! azimuth is given in [2*pi radians] so it takes values from 0 to 1
   
   integer, allocatable :: incl(:)                     ! indicates whether a parameter is estimable
   integer :: incl_mtyp,incl_mcytyp                    ! indicates whether mtyp and mcytyp is estimable
   integer, allocatable :: fen_mfdist(:)               ! indicates whether a mother-father phenotype distance is computed
   integer :: seed_kernel,pollen_kernel,linked_migr
   ! pollen_kernel=0: pollen and seed dispersal follows the single kernel (linked_kernels)
   ! kernel=1: exponential-power (default)
   ! kernel=2: Weibull
   ! kernel=3: Tufto's Wind Threshold Model
   ! kernel=4: lognormal
   ! kernel=5: power-law
   ! if linked_migr=1 then a single parameter is used for ms and mp, otherwise separate parameters
   character(len=31), dimension(0:5) :: kernel_type    ! kernel name
   
   double precision, allocatable :: grad(:)            ! gradient
   double precision, allocatable :: hess(:,:)          ! Hessian
   
   ! maternal_cy as estimable parameter
   integer :: incl_maternal_cy                         ! indicates whether maternal_cy is estimable
   double precision :: vmaternal_cy                    ! standard error for maternal_cy
   
   integer, allocatable :: ondg(:),odg(:)              ! stores non-dispersed and dispersed progeny indices
   integer :: nndg,ndg                                 ! number of non-dispersed and dispersed progeny
   
   !Random Number Generator (Xorshift 64)
   integer*8 :: seed,jsr
   integer*8, allocatable :: pjsr(:)
   
   !Bayesian estimation
   !Hierarchical Neighborhood Model
   logical :: fixedeffects                             ! if TRUE then hierarchical model is enabled
   integer,allocatable :: excl(:)                      ! phenotypic variables that remain a priori excluded from the analysis
   integer :: nexcl_g,nexcl_b                          ! number of excluded variables for female and male fitness
   double precision, allocatable :: femfec(:),malfec(:)! female and male fecundity
   double precision, allocatable :: mlogfemfec(:),mlogmalfec(:) ! female and male fecundity expectations
   double precision, allocatable :: proi(:)            ! the total probability for individual progeny (needed for WAIC)
   double precision :: femsigm,malsigm,femr2,malr2,m_e,y_e
   integer, allocatable :: geneal_m(:,:)               ! inferred genealogy (mother) per offspring
   integer, allocatable :: geneal_f(:,:)               ! inferred genealogy (father) per offspring
   
   !Auxiliary variables
   integer :: nfm,nlm,nls,nlo
   integer, allocatable :: noij(:,:),npjk(:,:)
   
   !Proposal distribution-related variables (acceptance and rejection numbers across MCMC, proposal distribution parameters)
   double precision :: ar_g, rr_g, v_g                          ! female selection gradients
   double precision :: ar_b, rr_b, v_b                          ! male selection gradients
   double precision :: ar_ds, rr_ds, v_ds                       ! mean distance of seed dispersal
   double precision :: ar_dp, rr_dp, v_dp                       ! mean distance of pollen dispersal
   double precision :: ar_bs, rr_bs, v_bs                       ! shape of seed dispersal kernel
   double precision :: ar_bp, rr_bp, v_bp                       ! shape of pollen dispersal kernel
   double precision :: ar_ks, rr_ks, v_ks                       ! anisotropy rate of seed dispersal kernel
   double precision :: ar_kp, rr_kp, v_kp                       ! anisotropy rate of pollen dispersal kernel
   double precision :: ar_as, rr_as, v_as                       ! prevailing direction of seed dispersal
   double precision :: ar_ap, rr_ap, v_ap                       ! prevailing direction of pollen dispersal
   double precision :: ar_ff, rr_ff, v_ff                       ! female fecundity
   double precision :: ar_fm, rr_fm, v_fm                       ! male fecundity
   double precision :: ar_sf, rr_sf, v_sf                       ! std.dev. of female fecundity distribution
   double precision :: ar_sm, rr_sm, v_sm                       ! std.dev. of male fecundity distribution
   double precision,allocatable :: ar_e(:), rr_e(:), d_e(:)     ! mistyping error rates
   double precision :: ar_ecy, rr_ecy, d_ecy                    ! mistyping error rate for cyDNA
   double precision :: ar_mcy, rr_mcy, d_mcy                    ! maternal transmission prob. of cyDNA
   double precision :: ar_m_e, rr_m_e, d_m_e                    ! mean of mistyping error rate
   double precision :: ar_y_e, rr_y_e, d_y_e                    ! dispersion of mistyping error rate
   
   !Parameters of the normal distribution of regression slopes used to draw new values in the RJ-MCMC sampling scheme
   double precision, allocatable :: mq_g(:),vq_g(:),mq_b(:),vq_b(:)
   
   !Lists of parameter values across MCMC
   double precision, allocatable :: lg(:,:),lb(:,:),lr2g(:),lr2b(:),lvg(:),lvb(:),llogfemfec(:,:),llogmalfec(:,:),ltheta(:,:), &
    lmtyp(:,:),lmcytyp(:),lmat_cy(:),lm_e(:),ly_e(:),lproi(:,:)
   integer, allocatable :: lmodg(:),lmodb(:),modng(:),modfg(:),modnb(:),modfb(:),frb(:),frg(:)
   
   !The number of regression models for female and male fitness
   integer :: nmodg,nmodb
   
   !The number of individual fecundity values saved to the mcmc output file
   integer :: nfecsaved = 1000
   
   !---------------------------------------------------------------------------------------------------------
   ! The main program starts here
   
   !====================================================================
   ! VERSION HISTORY:
   LASTUPDATE='30.11.2022'
   VERSION='2.0'
   !====================================================================
   
   !some initializations
   minmtyp=dble(0.0001)   !minimum non-zero mistyping error rate (values < mintyp are treated as zero in the ML analysis)
   minmcy=dble(0.0001)    !minimum non-zero maternal_cy (values < minmcy are treated as zero in the ML analysis)
   linked_migr=0          !pollen immigration can be set equal to seed immigration (works for the ML analysis only)
   maxlackconv=3          !the maximum number of Newton iterations showing 'divergence' behavior
   lackconv=0             !indicator = 1 if the successive Newton iterations show 'divergence' behavior
   fixedeffects=.true.    !FALSE if the Bayesian approach is used
   mtpl=100000            !multiplier used to stabilize floating point arithmetics (for likelihood computed in parallel)
   mypi=dacos(dble(-1.0)) !local pi constant
   my2pi=dble(2)*mypi     !local 2pi constant
   fi='info.txt'          !default input file name
   autostart=1            !if this binary flag indicator is 1 the analysis starts immidiately
   autostop=1             !if this binary flag indicator is 1 the program ends immidiately once the analysis is completeted
   misval = -9999.D0      !missing value indicator for phenotypic variables
   preprocess_phenotypes = .false. !is set to yes, phenotype data will be normalized; otherwise, they will be used untouched
   
   !dispersal kernel names
   kernel_type(0)='Linked to seed dispersal kernel'
   kernel_type(1)='Exponential-power'
   kernel_type(2)='Weibull'
   kernel_type(3)='Tufto'
   kernel_type(4)='Lognormal'
   kernel_type(5)='Power-law'
   
   call date_and_time(values=values)
   write (*,"(a)") '-------------------------------------------------'
   write (*,"(a)",advance="no") '        NMpi (run on '
   write (*,"(i4,2('-',i0.2),1x,i2,2(':',i0.2))",advance='no') (values(i),i=1,3),(values(i),i=5,7)
   write (*,"(a)") ')'
   write (*,"(a)") '                      version: '//trim(VERSION)
   write (*,"(a)") '                 last update: '//trim(LASTUPDATE)
   write (*,"(a)") '-------------------------------------------------'
   write (*,*)
   
   !rng seeded with the time (in milliseconds) since midnight
   seed=((values(5)*60+values(6))*60+values(7))*1000+values(8)
   jsr=seed
   call init_rnd
   
   !read info file
   call read_info
   
   !openmp initialization
   maxnth = omp_get_max_threads()
   if (usernth.gt.maxnth) then
     usernth=maxnth
     write (*,"(a, i3)") 'Number of threads reduced to the maximum of:', usernth
   else
     if (usernth.lt.1) then
       usernth=1
     endif
     write (*,"(a, i3)") 'Number of threads: ', usernth
   endif
   
   call init_prnd
   
   call read_data
   call normalize_phenotypes
   call identify_alleles
   call order_alleles
   call introduce_genotyping_errors(0.05D0,0.025D0)
   call allele_frequencies
   call identify_parents
   call progeny_groups
   call neighbors
   call read_errors
   if (nf>0) call read_gradients
   call initialize_model
   
   if (linked_migr==1) write (*,"(a)") 'Seed and pollen migration rates: linked'
   write (*,"(a)",advance='no') 'Seed dispersal kernel: '
   write (*,"(a)") kernel_type(seed_kernel)
   write (*,"(a)",advance='no') 'Pollen dispersal kernel: '
   write (*,"(a)") kernel_type(pollen_kernel)
      
   if (autostart==0) then
     write (*,"(a)") 'Press <ENTER> to continue...'
     read (*,*)
   endif
   
   tim0 = omp_get_wtime()
   
   !Initial linear search for optimal parameter values
   !(helps in convergence but may fail sometimes)
   if (maxpit>0) call powell_theta(maxpit)
   
   nit=0
   
   if (maxit>0) then
     ! Maximum Likelihood approach
     write (*,"(a)") 'Initializing...'
     old_logl = full_logl()
     write (*,*)
     write (*,*) 'Iter ','log-Likelihood ',(ntheta(j),j=1,11),(ntheta(11+j),ntheta(11+j+nf),j=1,nf)
     if (dist_typ==1) then
       write (*,"(i5,f15.2,*(f10.5))") nit,old_logl,(theta(j),j=1,11),(theta(11+j),theta(11+j+nf),j=1,nf)
     else
       write (*,"(i5,f15.2,3(f10.5),2(f10.2),*(f10.5))") nit,old_logl,(theta(j),j=1,11), &
        (theta(11+j),theta(11+j+nf),j=1,nf)
     endif
     if (incl_mtyp==1) write (*,"(a10,*(f10.5))") ' mtyp_nu:',(mtyp(j),j=1,nl)
     if (incl_mcytyp==1) write (*,"(a10,*(f10.5))") ' mtyp_cy:',mcytyp
     if (incl_maternal_cy==1) write (*,"(a10,*(f10.5))") ' mater_cy:',maternal_cy
     
     fh=1
     
     if (nincl>0) then
       do i=1,maxit
         call derivatives(fh)
         call invert_hess
         call update_theta
         if (incl_mtyp==1) call update_mtyp
         if (incl_mcytyp==1) call update_mcytyp
         if (incl_maternal_cy==1) call update_maternal_cy
         
         new_logl = full_logl()
         if (dist_typ==1) then
           write (*,"(i5,f15.2,*(f10.5))") i,new_logl,(theta(j),j=1,11),(theta(11+j),theta(11+j+nf),j=1,nf)
         else
           write (*,"(i5,f15.2,3(f10.5),2(f10.2),*(f10.5))") i,new_logl,(theta(j),j=1,11), &
            (theta(11+j),theta(11+j+nf),j=1,nf)
         endif
         if (incl_mtyp==1) write (*,"(a10,*(f10.5))") ' mtyp_nu:',(mtyp(j),j=1,nl)
         if (incl_mcytyp==1) write (*,"(a10,*(f10.5))") ' mtyp_cy:',mcytyp
         if (incl_maternal_cy==1) write (*,"(a10,*(f10.5))") ' mater_cy:',maternal_cy
         
         !the hack that allows the Hessian to contain only diagonal terms when the parameters are far from the ML solution
         if (dabs(new_logl - old_logl) < 100.0) then
           fh=1 
         else
           fh=0
         endif
         
         ! CONVERGENCE CRITERION HERE
         if (dabs(new_logl - old_logl) < 0.001) exit
         if (new_logl-old_logl < dble(0)) lackconv=lackconv+1
         if (ISNAN(new_logl)) lackconv=maxlackconv
         if (lackconv==maxlackconv) exit
         if (dabs(new_logl) > huge(new_logl)) lackconv=maxlackconv
         old_logl = new_logl
         nit=nit+1
       enddo
     endif
     
     tim1 = omp_get_wtime()
     
     if (lackconv==maxlackconv) then
       write(*,"(a)") '>>Lack of convergence<< criterion met. Analysis aborted! Possible problems with initial values.'
     else
       
       outfile = 66
       open (unit=outfile,file=fo,action='write',status='replace')
       
       write(outfile,"(a,f10.2)") 'Final log-L: ',new_logl
       write(outfile,"(a)") 'Mating model parameters:'
       write(outfile,"(*(a))") 'Param.:  ',(ntheta(j),j=1,11),(ntheta(11+j),ntheta(11+j+nf),j=1,nf)
       if (dist_typ==1) then
         write(outfile,"(a,*(f10.5))") 'Estim.:  ',(theta(j),j=1,11),(theta(11+j),theta(11+j+nf),j=1,nf)
         write(outfile,"(a,*(f10.5))") 'SE:      ',(vtheta(j),j=1,11),(vtheta(11+j),vtheta(11+j+nf),j=1,nf)
       else
         write(outfile,"(a,3(f10.5),2(f10.2),*(f10.5))") 'Estim.:  ',(theta(j),j=1,11),(theta(11+j),theta(11+j+nf),j=1,nf)
         write(outfile,"(a,3(f10.5),2(f10.2),*(f10.5))") 'SE:      ',(vtheta(j),j=1,11),(vtheta(11+j),vtheta(11+j+nf),j=1,nf)
       endif
       write(outfile,"(a)") 'Per-locus mistyping rates:'
       write(outfile,"(a,*(f10.5))") 'Estim.:  ',(mtyp(i),i=1,nl)
       write(outfile,"(a,*(f10.5))") 'SE:      ',(vmtyp(i),i=1,nl)
       write(outfile,"(a)") 'Mistyping rate for cyDNA:'
       write(outfile,"(a,*(f10.5))") 'Estim.:  ',mcytyp
       write(outfile,"(a,*(f10.5))") 'SE:      ',vmcytyp
       write(outfile,"(a)") 'Probability of maternal transmission of cyDNA:'
       write(outfile,"(a,*(f10.5))") 'Estim.:  ',maternal_cy
       write(outfile,"(a,*(f10.5))") 'SE:      ',vmaternal_cy
       
       if (linked_migr==1) write (outfile,"(a)") 'Seed and pollen migration rates: linked'
       write(outfile,"(a)",advance='no') 'Seed dispersal kernel: '
       write(outfile,"(a)") kernel_type(seed_kernel)
       write(outfile,"(a)",advance='no') 'Pollen dispersal kernel: '
       write(outfile,"(a)") kernel_type(pollen_kernel)
       
       if (neighborhood<dble(0)) then
         write(outfile,"(a)") 'Neighborhood radius: unlimited'
       else 
         write(outfile,"(a,f10.2)") 'Neighborhood radius: ',neighborhood
       endif
       
       flush(outfile)
       close(outfile)
       
       call write_info()
       call write_gradients()
       call write_errors()
       if (writecov==1) call write_cov()
       call db_confidence_bounds()
       
       write(*,*)
       write(*,"(a)") 'Analysis finished normally...'
       write(*,"(a)") 'Results written to the output file: '//trim(fo)
       write(*,*)
       write(*,"(a,f15.2)") 'Total time (in sec.): ',(tim1-tim0)
       write(*,*)
       
       write(*,*)
       write(*,"(a)") 'Performing parentage assignment...'
       call conditional_probs
       write(*,"(a)") 'Parentage results written to the output file: '//trim(fp)
     endif
   else
     ! Bayesian approach
     call run_pilotmcmc(20000) !The fixed-length pilot run (20K iterations)
     call run_mcmc(-maxit,20)
     tim1 = omp_get_wtime()
     open(unit=789, file=trim(fm), status="old", position="append", action="write")
     write(789,"('CPU time (in sec.)',a,f15.1)") tab,(tim1-tim0)
     close(789)
     write(*,*)
     write(*,"(a,f15.2)") 'Total time (in sec.): ',(tim1-tim0)
     write(*,*)
   endif
   
   if (autostop==0) then
     write(*,"(a)") 'Press <ENTER> to exit...'
     read(*,*)
   endif
   ! The main program ends here
   !---------------------------------------------------------------------------------------------------------
 
 ! All routines used in the program are given below
  
 contains
 
 !-----------------------------------------------------------------------------------------------------------
 ! This routine reads a text file containing initial settings
 subroutine read_info
   character(len=256) :: n1,n2,n3,n4,n5
   integer :: infile,i
   
   infile=44
   
   open(unit=infile, file=fi)
   
   nincl = 0
   
   allocate(theta(11))
   allocate(incl(11))
   
   do i=1,7
     read(infile,*) theta(i),incl(i),minth(i),maxth(i)
     nincl=nincl+incl(i)
   enddo
   
   do i=8,11
     read(infile,*) theta(i),incl(i)
     nincl=nincl+incl(i)
   enddo
   
   
   read(infile,*) dist_typ,seed_kernel,pollen_kernel ! seed and pollen dispersal kernel
   read(infile,*) maternal_cy,incl_maternal_cy ! pr. of maternal transmission of cyDNA
   read(infile,*) neighborhood! radius of the neighborhood
   read(infile,*) maxpit,maxit! maximum number of iterations allowed
   read(infile,*) usernth     ! the number of threads to be used for parallel computing
   read(infile,*) mep         ! machine error precision
   read(infile,*) writecov    ! binary indicator of whether write out the cov matrix
   read(infile,*) n1          ! data file name (max=256 characters!)
   read(infile,*) n2          ! output file name (max=256 characters!)
   read(infile,*) n3          ! error rates file name (max=256 characters!)
   read(infile,*) n4          ! selection gradients file name
   read(infile,*) n5          ! file for the output of parentage assignment
   fd = n1
   fo = n2
   fe = n3
   fg = n4
   fp = n5
   close(infile)
   
 end subroutine read_info
 !-----------------------------------------------------------------------------------------------------------
 
 !-----------------------------------------------------------------------------------------------------------
 ! This routine creates a text file containing initial settings
 subroutine write_info
   character(len=256) :: finf
   integer :: outinf,i
   
   finf='info.out'
   outinf=48
   
   open(unit=outinf,file=finf,action='write',status='replace')
   
   do i=1,7
     write(outinf,"(f15.5,i5,f15.5,f15.5)") theta(i),incl(i),minth(i),maxth(i)
   enddo
   
   do i=8,11
     write(outinf,"(f15.5,i5)") theta(i),incl(i)
   enddo
   
   write(outinf,*) dist_typ,seed_kernel,pollen_kernel      ! dispersal kernel modes
   write(outinf,"(f15.5,i5)") maternal_cy,incl_maternal_cy ! pr. of maternal transmission of cyDNA
   write(outinf,*) neighborhood                            ! radius of the neighborhood
   write(outinf,*) maxpit,maxit                            ! max iterations
   write(outinf,*) usernth                                 ! the number of threads to be used for parallel computing
   write(outinf,"(e15.4)") mep                             ! machine precision
   write(outinf,*) writecov                                ! binary indicator of whether write out the cov matrix
   write(outinf,*) trim(fd)                                ! data file name (max=256 characters!)
   write(outinf,*) trim(fo)                                ! output file name (max=256 characters!)
   write(outinf,*) trim(fe)                                ! error rates file name (max=256 characters!)
   write(outinf,*) trim(fg)                                ! selection gradients file name
   write(outinf,*) trim(fp)                                ! file for the output of parentage assignment
   flush(outinf)
   close(outinf)
   
 end subroutine write_info
 !-----------------------------------------------------------------------------------------------------------
 
 !-----------------------------------------------------------------------------------------------------------
 ! This routine reads a text file containing genotyping errors
 subroutine read_errors
   integer :: infile,i
   
   infile=44
   open(unit=infile, file=fe)
   allocate(mtyp(nl))
   read(infile,*) incl_mtyp,incl_mcytyp 
   read(infile,*) (mtyp(i),i=1,nl)   !initial errors
   read(infile,*) mcytyp
   close(infile)
   nincl=nincl+incl_mtyp+incl_mcytyp
   
 end subroutine read_errors
 !-----------------------------------------------------------------------------------------------------------
 
 !-----------------------------------------------------------------------------------------------------------
 ! This routine creates a text file containing genotyping errors
 subroutine write_errors
   integer :: outerr,i
   
   outerr=44
   open(unit=outerr, file='errors.out',action='write',status='replace')
   write(outerr,*) incl_mtyp,incl_mcytyp
   
   do i=1,nl
     write(outerr,"(f15.5)") mtyp(i)
   enddo
   
   write(outerr,"(f15.5)") mcytyp
   flush(outerr)
   close(outerr)
   
 end subroutine write_errors
 !-----------------------------------------------------------------------------------------------------------
 
 !-----------------------------------------------------------------------------------------------------------
 ! This routine reads a text file containing regression slopes
 subroutine read_gradients
   integer :: infile,i
   double precision, allocatable :: tmpt(:)
   integer, allocatable :: tmpi(:)
   
   infile=43
   
   allocate(tmpt(11))
   allocate(tmpi(11))
   
   tmpt=theta
   tmpi=incl
   
   deallocate(theta)
   allocate(theta(11+2*nf))
   
   deallocate(incl)
   allocate(incl(11+2*nf))
   
   allocate(fen_mfdist(nf))
   allocate(excl(2*nf))
   
   do i=1,11
     theta(i)=tmpt(i)
     incl(i)=tmpi(i)
   enddo
   
   open(unit=infile, file=fg)
   
   do i=1,nf
     read(infile,*) theta(11+i),incl(11+i),theta(11+nf+i),incl(11+nf+i),fen_mfdist(i)
   enddo
   close(infile)
   
   do i=1,2*nf
     nincl=nincl+incl(11+i)
   enddo
   
   nexcl_b=0
   nexcl_g=0
   do i=1,nf
     !variables with zero variance have to be excluded regardless settings in the selection gradients file
     if (vfen(1,i)==0.D0) incl(11+i)=0
     if (vfen(2,i)==0.D0) incl(11+nf+i)=0
     
     excl(i)=1-incl(11+i)
     excl(nf+i)=1-incl(11+nf+i)
     nexcl_g=nexcl_g+excl(i)
     nexcl_b=nexcl_b+excl(nf+i)
   enddo
   
 end subroutine read_gradients
 !-----------------------------------------------------------------------------------------------------------
 
 !-----------------------------------------------------------------------------------------------------------
 ! This routine creates a text file containing regression slopes
 subroutine write_gradients
   integer :: outgrad,i
   
   outgrad=49
   
   open(unit=outgrad,file='gradients.out',action='write',status='replace')
   
   do i=1,nf
     write(outgrad,"(2(f15.5,i5),i5)") theta(11+i),incl(11+i),theta(11+nf+i),incl(11+nf+i), &
       fen_mfdist(i)
   enddo
   
   flush(outgrad)
   close(outgrad)
   
 end subroutine write_gradients
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine creates a text file containing covariance matrix
 subroutine write_cov
   integer :: outhess,i,j
   
   outhess=59
   
   open(unit=outhess,file='cov.out',action='write',status='replace')
   
   do i=1,11+2*nf
     write(outhess,"(*(f15.10))") (-hess(i,j),j=1,11+2*nf)
   enddo
   
   flush(outhess)
   close(outhess)
   
 end subroutine write_cov
 !----------------------------------------------------------------------------------------------------------
 
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine reads a data file
 subroutine read_data
   integer :: i,j,k,g
   
   open(unit=99, file=fd)
   read(99, *) np,no,nl,nf  ! number of parents, offspring, markers, phenotypic characters
   allocate(pgen(np,nl,2))  ! parental genotypes, assumes diploidy
   allocate(ogen(no,nl,2))  ! offspring genotypes, assumes diploidy
   allocate(odisp(no))      ! 'dispersed progeny' binary indicator
   allocate(pcy(np))        ! parental cyDNA
   allocate(ocy(no))        ! offspring cyDNA
   allocate(px(np))         ! parental x coordinates
   allocate(py(np))         ! parental y coordinates
   allocate(ox(no))         ! offspring x coordinates
   allocate(oy(no))         ! offspring y coordinates
   allocate(om(no))         ! offspring known mothers
   allocate(of(no))         ! offspring known fathers
   allocate(pid(np))        ! parental ID
   allocate(oid(no))        ! offspring ID
   allocate(pfen(np,nf))    ! parental phenotype
   allocate(pfem(np))       ! parental gender
   allocate(onb(no,np))     ! offspring neighborhoods (seed donors within a defined radius)
   allocate(pnb(np,np))     ! parental neighborhoods (pollen donors within a defined radius)
   
   !read data from file
   do i=1,np
     read(99,*) g,pid(i),px(i),py(i),pcy(i),((pgen(i,j,k),k=1,2),j=1,nl),&
       (pfen(i,j),j=1,nf),pfem(i)
       
     if (g.ne.0) then
       write(*,*) 'Incorrect sample in the parental block...'
     endif
   enddo
   
   do i=1,no
     read(99,*) g,oid(i),ox(i),oy(i),ocy(i),((ogen(i,j,k),k=1,2),j=1,nl),&
       om(i),of(i),odisp(i)
     if (g.ne.1) then
       write(*,*) 'Incorrect sample in the offspring block...'
     endif
   enddo
   close(99)
   
   write(*,"(a)",advance='no') 'Data correctly read from file: '
   write(*,"(a)") trim(fd)
   
 end subroutine read_data
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine normalizes phenotypic variables generating two series of values for two sexes
 subroutine normalize_phenotypes
   integer :: i,j
   character(len=256) :: ffen
   
   allocate(pzfen(2,np,nf))
   allocate(mfen(2,nf))
   allocate(vfen(2,nf))
   allocate(nfen(2,nf))
   
   mfen=0.D0
   vfen=0.D0
   nfen=0
   
   !compute means
   do i=1,np
     do j=1,nf
       if (pfem(i)>0.D0) then
         if (pfen(i,j) .ne. misval) then
           nfen(1,j)=nfen(1,j)+1
           mfen(1,j)=mfen(1,j)+pfen(i,j)
         endif
       endif
       
       if (pfem(i)<1.D0) then
         if (pfen(i,j) .ne. misval) then
           nfen(2,j)=nfen(2,j)+1
           mfen(2,j)=mfen(2,j)+pfen(i,j)
         endif
       endif
     enddo
   enddo
   do j=1,nf
     if (nfen(1,j)>0) mfen(1,j)=mfen(1,j)/dble(nfen(1,j))
     if (nfen(2,j)>0) mfen(2,j)=mfen(2,j)/dble(nfen(2,j))
   enddo
   
   !compute variances
   do i=1,np
     do j=1,nf
       if (pfem(i)>0.D0) then
         if (pfen(i,j) .ne. misval) vfen(1,j)=vfen(1,j)+(pfen(i,j)-mfen(1,j))**2
       endif
       
       if (pfem(i)<1.D0) then
         if (pfen(i,j) .ne. misval) vfen(2,j)=vfen(2,j)+(pfen(i,j)-mfen(2,j))**2
       endif
     enddo
   enddo
   do j=1,nf
     if (nfen(1,j)>0) vfen(1,j)=vfen(1,j)/dble(nfen(1,j))
     if (nfen(2,j)>0) vfen(2,j)=vfen(2,j)/dble(nfen(2,j))
   enddo
   
   !compute normalized scores for two sexes separately
   do i=1,np
     do j=1,nf
       if (pfem(i)>0.D0) then
         if (pfen(i,j) == misval .or. vfen(1,j) == 0.D0) then
           pzfen(1,i,j)=0.D0
         else
           pzfen(1,i,j)=(pfen(i,j)-mfen(1,j))/sqrt(vfen(1,j))
         endif
       endif
       
       if (pfem(i)<1.D0) then
         if (pfen(i,j) == misval .or. vfen(2,j) == 0.D0) then
           pzfen(2,i,j)=0.D0
         else
           pzfen(2,i,j)=(pfen(i,j)-mfen(2,j))/sqrt(vfen(2,j))
         endif
       endif
     enddo
   enddo
   
   !write output file reporting on the normalization process named as the output file but with the extension 'fen'
   if (preprocess_phenotypes) then
     j=0
     do i=1,len(trim(fo))
       if (fo(i:i)=='.') then
         j=i
         write(ffen,"(a,'.fen')") fo(1:i-1)
         exit
       endif
     enddo
     open(unit=456,file=trim(ffen),action="write",status="replace")
     write(456,"(a)") 'Means and variances of phenotypic variables:'
     write(456,"(13a)") 'Var',tab,'Num_f',tab,'Mean_f',tab,'Variance_f',tab,'Num_m',tab,'Mean_m',tab,'Variance_m'
     do i=1,nf
       write(456,"(i0,2(a,i0,2(a,e16.8)))") i,tab,nfen(1,i),tab,mfen(1,i),tab,vfen(1,i),tab,nfen(2,i),tab,mfen(2,i),tab,vfen(2,i)
     enddo
     write(456,"(a)") ''
     write(456,"(a)") 'Normalized phenotypic variables used for female function:'
     do i=1,np
       write(456,"(i0,*(a,e16.8))") pid(i),(tab,pzfen(1,i,j),j=1,nf)
     enddo
     write(456,"(a)") ''
     write(456,"(a)") 'Normalized phenotypic variables used for male function:'
     do i=1,np
       write(456,"(i0,*(a,e16.8))") pid(i),(tab,pzfen(2,i,j),j=1,nf)
     enddo
     close(456)
   else
     pzfen(1,:,:)=pfen
     pzfen(2,:,:)=pfen
   endif
 end subroutine normalize_phenotypes
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine identifies all alleles in genetic data
 subroutine identify_alleles
   integer, allocatable ::  tmpallele(:,:),tmpcyhap(:)
   integer :: i,j,k,l,m,b
   integer, allocatable :: nallp(:)
   
   ! (double -1 used for missing genotype)
   
   allocate(allele(nl,0))
   allocate(nall(nl))
   allocate(tmpallele(nl,0))
   
   do i=1,nl
     nall(i)=0
   enddo
   
   maxl=0
   
   allocate(cyhap(0))
   allocate(tmpcyhap(0))
   nhap = 0
   
   ! find all alleles in parents
   do i=1,np
     do j=1,nl
       do k=1,2
         b=0
         do l=1,nall(j)
           if (pgen(i,j,k)==allele(j,l)) then
             b=b+1
           endif
         enddo
         
         if ((pgen(i,j,k).ne.-1).and.(b==0)) then
           tmpallele=allele
           nall(j)=nall(j)+1
           if (maxl<nall(j)) then
             maxl=nall(j)
             deallocate(allele)
             allocate(allele(nl,maxl))
           endif
           do l=1,nl
             do m=1,nall(l)
               if (l.ne.j) then
                 allele(l,m)=tmpallele(l,m)
               else if (m.lt.nall(l)) then
                 allele(l,m)=tmpallele(l,m)
               endif
             enddo
           enddo
           allele(j,nall(j))=pgen(i,j,k)
           deallocate(tmpallele)
           allocate(tmpallele(nl,maxl))
         endif
       enddo
     enddo
     
     b=0
     do l=1,nhap
       if (pcy(i)==cyhap(l)) then
         b=b+1
       endif
     enddo
     
     if ((pcy(i).ne.-1).and.(b==0)) then
       tmpcyhap=cyhap
       nhap=nhap+1
       deallocate(cyhap)
       allocate(cyhap(nhap))
       
       do l=1,nhap-1
         cyhap(l)=tmpcyhap(l)
       enddo
       cyhap(nhap)=pcy(i)
       deallocate(tmpcyhap)
       allocate(tmpcyhap(nhap))
     endif
   enddo
   
   allocate(nallp(nl))
   do l=1,nl
     nallp(l)=nall(l)
   enddo
   
   ! find additional alleles in progeny
   do i=1,no
     do j=1,nl
       do k=1,2
         b=0
         do l=1,nall(j)
           if (ogen(i,j,k)==allele(j,l)) then
             b=b+1
           endif
         enddo
         
         if ((ogen(i,j,k).ne.-1).and.(b==0)) then
           tmpallele=allele
           nall(j)=nall(j)+1
           if (maxl<nall(j)) then
             maxl=nall(j)
             deallocate(allele)
             allocate(allele(nl,maxl))
           endif
           do l=1,nl
             do m=1,nall(l)
               if (l.ne.j) then
                 allele(l,m)=tmpallele(l,m)
               else if (m.lt.nall(l)) then
                 allele(l,m)=tmpallele(l,m)
               endif
             enddo
           enddo
           allele(j,nall(j))=ogen(i,j,k)
           deallocate(tmpallele)
           allocate(tmpallele(nl,maxl))
         endif
       enddo
     enddo
     
     b=0
     do l=1,nhap
       if (ocy(i)==cyhap(l)) then
         b=b+1
       endif
     enddo
     
     if ((ocy(i).ne.-1).and.(b==0)) then
       tmpcyhap=cyhap
       nhap=nhap+1
       deallocate(cyhap)
       allocate(cyhap(nhap))
       
       do l=1,nhap-1
         cyhap(l)=tmpcyhap(l)
       enddo
       cyhap(nhap)=ocy(i)
       deallocate(tmpcyhap)
       allocate(tmpcyhap(nhap))
     endif
   enddo
   
   b=0
   write(*,"(a28)") 'Number of alleles per locus:'
   write(*,"(a)") 'Locus Parents  Total'
   do i=1,nl
     write(*,"(i3,i6,i8)") i,nallp(i),nall(i)
     b=b+nall(i)
   enddo
   
   write(*,"(a, i0)") 'Total number of alleles: ',b
   write(*,"(a, i0)") 'Number of cyDNA haplotypes: ',nhap
   
   ! now encode alleles using their indices in the allele table
   do i=1,np
     do j=1,nl
       do k=1,2
         do m=1,nall(j)
           if (pgen(i,j,k)==allele(j,m)) then
             pgen(i,j,k)=m
             exit
           endif
         enddo
       enddo
     enddo
     
     do m=1,nhap
       if (pcy(i)==cyhap(m)) then
         pcy(i)=m
         exit
       endif
     enddo
   enddo

   do i=1,no
     do j=1,nl
       do k=1,2
         do m=1,nall(j)
           if (ogen(i,j,k)==allele(j,m)) then
             ogen(i,j,k)=m
             exit
           endif
         enddo
       enddo
     enddo
     
     do m=1,nhap
       if (ocy(i)==cyhap(m)) then
         ocy(i)=m
         exit
       endif
     enddo

   enddo
   
 end subroutine identify_alleles
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine sorts alleles within a genotype
 subroutine order_alleles
   integer :: i,j,k
   
   do i=1,np
     do j=1,nl
       if (pgen(i,j,1) > pgen(i,j,2)) then
         k = pgen(i,j,1)
         pgen(i,j,1) = pgen(i,j,2)
         pgen(i,j,2) = k
       endif
     enddo
   enddo

   do i=1,no
     do j=1,nl
       if (ogen(i,j,1) > ogen(i,j,2)) then
         k = ogen(i,j,1)
         ogen(i,j,1) = ogen(i,j,2)
         ogen(i,j,2) = k
       endif
     enddo
   enddo
 end subroutine order_alleles
 !----------------------------------------------------------------------------------------------------------

 !----------------------------------------------------------------------------------------------------------
 ! This routine creates an index list of known mothers and fathers
 subroutine identify_parents
   integer i,j,m,f
   
   do i=1,no
     if (om(i)<0) of(i)=-1
     m=-1
     f=-1
     do j=1,np
       if (om(i)==pid(j)) m=j
       if (of(i)==pid(j)) f=j
     enddo
     om(i)=m
     of(i)=f
   enddo
   
 end subroutine identify_parents
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 subroutine progeny_groups
   integer :: i
   
   nndg=0
   ndg=0
   allocate(ondg(no))
   allocate(odg(no))
   
   do i=1,no
     
     if (odisp(i)==0) then
       nndg=nndg+1
       ondg(nndg)=i
     else
       ndg=ndg+1
       odg(ndg)=i
     endif
     
   enddo
   
 end subroutine progeny_groups
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the Kronecker delta for two alleles assuming random mistyping error at rate of ep
 double precision function delta_eps(i,j,k,ep)
   integer, intent(in) :: i,j,k
   double precision, intent(in) :: ep
   double precision :: dij
   
   if (k>1) then
     if (i==j) then 
       dij = (dble(1)-ep)**2+ep**2/dble(k-1)
     else 
       dij = ep*(dble(2*(k-1))-ep*dble(k))/(dble(k-1))**2
     endif
   else
     dij = dble(1)
   end if
   
   delta_eps=dij
   
 end function delta_eps
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine generates random errors in genotypes
 ! Per-locus error rates are drawn from the truncated exponential distribution
 ! This subroutine must be run after allele decoding ('identify_alleles' subroutine)
 subroutine introduce_genotyping_errors(mean_dropout,mean_mistyping)
   double precision :: mean_dropout,mean_mistyping,e1(nl),e2(nl),b1,b2
   integer :: i,a
   
   ! Note that this routine is disabled by default!
   return
   
   !the following approximation works for mean<=0.42
   b1=mean_dropout
   b2=mean_mistyping
   do i=1,100
     b1=(1.D0-mean_dropout*(1.D0-dexp(1.D0/b1)))/(dexp(1.D0/b1)-1.D0)
     b2=(1.D0-mean_mistyping*(1.D0-dexp(1.D0/b2)))/(dexp(1.D0/b2)-1.D0)
   enddo
   
   !random error rates
   do i=1,nl
     e1(i)=rnd()
     e2(i)=rnd()
     e1(i)=b1*dlog(1.D0/(e1(i)-dexp(1.D0/b1)*(e1(i)-1.D0)))+1.D0
     e2(i)=b2*dlog(1.D0/(e2(i)-dexp(1.D0/b2)*(e2(i)-1.D0)))+1.D0
   enddo
   
   !errors in parents
   do i=1,np
     do j=1,nl
       !random dropout
       if (pgen(i,j,1)/=pgen(i,j,2)) then
         if (rnd()<e1(j)) then
           a=int(rnd()*2)+1
           pgen(i,j,a)=pgen(i,j,3-a)
         endif
       endif
       
       !random mistyping of the 1st allele
       if (rnd()<e2(j)) then
         do
           a=int(rnd()*nall(j))+1
           if (a/=pgen(i,j,1)) exit
         enddo
         pgen(i,j,1)=a
       endif
         
       !random mistyping of the 2nd allele
       if (rnd()<e2(j)) then
         do
           a=int(rnd()*nall(j))+1
           if (a/=pgen(i,j,2)) exit
         enddo
         pgen(i,j,2)=a
       endif
     enddo
   enddo
   
   !errors in progeny
   do i=1,no
     do j=1,nl
       !random dropout
       if (ogen(i,j,1)/=ogen(i,j,2)) then
         if (rnd()<e1(j)) then
           a=int(rnd()*2)+1
           ogen(i,j,a)=ogen(i,j,3-a)
         endif
       endif
       
       !random mistyping of the 1st allele
       if (rnd()<e2(j)) then
         do
           a=int(rnd()*nall(j))+1
           if (a/=ogen(i,j,1)) exit
         enddo
         ogen(i,j,1)=a
       endif
       
       !random mistyping of the 2nd allele
       if (rnd()<e2(j)) then
         do
           a=int(rnd()*nall(j))+1
           if (a/=ogen(i,j,2)) exit
         enddo
         ogen(i,j,2)=a
       endif
     enddo
   enddo
   
   open(unit=234,file='genotypes_with_errors.out',action='write',status='replace')
   do i=1,np
     write (234,*) pid(i),(allele(j,pgen(i,j,1)),allele(j,pgen(i,j,2)),j=1,nl)
   enddo
   do i=1,no
     write (234,*) oid(i),(allele(j,ogen(i,j,1)),allele(j,ogen(i,j,2)),j=1,nl)
   enddo
   do j=1,nl
     write (234,*) j,e1(j),e2(j)
   enddo
   close(234)
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the probability for genotype u/v given that a parent w/x fertilizes itself
 double precision function pr_mm(u,v,w,x,k,ep)
   integer, intent(in) :: u,v,w,x,k
   double precision, intent(in) :: ep
   double precision :: duv,duw,dux,dvw,dvx
   
   if (k>1) then
       
       duv=delta_eps(u,v,k,ep)
       duw=delta_eps(u,w,k,ep)
       dux=delta_eps(u,x,k,ep)
       dvw=delta_eps(v,w,k,ep)
       dvx=delta_eps(v,x,k,ep)
       
       pr_mm = (dble(2)-duv)*(duw+dux)*(dvw+dvx)/dble(4)
       
   else
     pr_mm = dble(1)
   endif
   
 end function pr_mm
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the probability for genotype u/v given that parents have genotypes w/x and y/z
 double precision function pr_mf(u,v,w,x,y,z,k,ep)
   integer, intent(in) :: u,v,w,x,y,z,k
   double precision, intent(in) :: ep
   double precision :: duv,duw,dux,dvw,dvx,duy,duz,dvy,dvz
   double precision :: duwx,dvwx,duyz,dvyz
   
   if (k > 1) then
       
       duv=delta_eps(u,v,k,ep)
       duw=delta_eps(u,w,k,ep)
       dux=delta_eps(u,x,k,ep)
       dvw=delta_eps(v,w,k,ep)
       dvx=delta_eps(v,x,k,ep)
       duy=delta_eps(u,y,k,ep)
       duz=delta_eps(u,z,k,ep)
       dvy=delta_eps(v,y,k,ep)
       dvz=delta_eps(v,z,k,ep)
       
       duwx = (duw + dux) / dble(2)
       dvwx = (dvw + dvx) / dble(2)
       duyz = (duy + duz) / dble(2)
       dvyz = (dvy + dvz) / dble(2)
       
       pr_mf = (dble(2)-duv)*(duwx*dvyz+dvwx*duyz)/dble(2)
   else
     pr_mf = dble(1)
   endif

 end function pr_mf
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the probability for genotype u/v given that one parent has genotype w/x and
 ! the second parent is unknown (alleles are drawn at random from a gene pool)
 double precision function pr_mb(u,v,w,x,k,ep,l)
   integer, intent(in) :: u,v,w,x,k,l
   double precision, intent(in) :: ep
   double precision :: pr,duv,duw,dux,dvw,dvx,duwx,dvwx,dub,dvb
   
   if (k>1) then
   
       duv=delta_eps(u,v,k,ep)
       duw=delta_eps(u,w,k,ep)
       dux=delta_eps(u,x,k,ep)
       dvw=delta_eps(v,w,k,ep)
       dvx=delta_eps(v,x,k,ep)
       
       duwx = (duw + dux) / dble(2)
       dvwx = (dvw + dvx) / dble(2)
       dub = (dble(1)-ep)*plm(l,u)+ep/dble(k-1)*(dble(1)-plm(l,u))
       dvb = (dble(1)-ep)*plm(l,v)+ep/dble(k-1)*(dble(1)-plm(l,v))
       
       pr = (dble(2)-duv)*(duwx*dvb+dvwx*dub)/dble(2)
   else
     pr=dble(1)
   endif
   
   pr_mb = pr
 end function pr_mb
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the HW probability for genotype u/v
 double precision function pr_b(u,v,k,ep,l)
   integer, intent(in) :: u,v,k,l
   double precision, intent(in) :: ep
   double precision :: pr,duv,dub,dvb
   
   if (k>1) then
       duv=delta_eps(u,v,k,ep)
       
       dub = (dble(1)-ep)*plm(l,u)+ep/dble(k-1)*(dble(1)-plm(l,u))
       dvb = (dble(1)-ep)*plm(l,v)+ep/dble(k-1)*(dble(1)-plm(l,v))
       pr = (dble(2)-duv)*dub*dvb
   else
     pr=dble(1)
   endif
   
   pr_b = pr
   
 end function pr_b
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine identifies "neighbors", i.e. excludes candidate mothers and fathers that are located further
 ! than the limit distance (neighborhood radius)
 subroutine neighbors
   double precision :: d
   integer :: i,j
   
   
   do i=1,no
     do j=1,np
     d = dsqrt((px(j)-ox(i))**2 + (py(j)-oy(i))**2)
     if (d <= neighborhood .or. neighborhood<dble(0)) then
       onb(i,j)=1
     else
       onb(i,j)=0
     endif
     enddo
   enddo
   
   do i=1,np
     do j=1,np
     d = dsqrt((px(j)-px(i))**2 + (py(j)-py(i))**2)
     if (d <= neighborhood .or. neighborhood<dble(0)) then
       pnb(i,j)=1
     else
       pnb(i,j)=0
     endif
     enddo
   enddo
   
 end subroutine neighbors
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes the relative proportion of seeds from each candidate father expected for
 ! the i-th seedling's location (seed profile)
 subroutine compute_ksi(i,d_s,b_s)
   integer, intent(in) :: i
   double precision, intent(in) :: d_s,b_s
   integer :: j,k
   double precision :: a_s,d,f,xc,yc
   
   if (odisp(i)==0) then
     ! if offspring is not dispersed it does not provide information
     ! about reproductive success of the mother
     sksi(i) = dble(1)
     do j=1,np
       ksi(i,j)=dble(0)
       if (om(i)==j) ksi(i,j)=dble(1)
     enddo
   else
     sksi(i) = dble(0)
     do j=1,np
       ksi(i,j)=dble(0)
       if (onb(i,j)==1) then
       
         if (pfem(j) > dble(0)) then
           d = dsqrt((px(j)-ox(i))**2 + (py(j)-oy(i))**2)
           
           xc=ox(i)-px(j)
           yc=oy(i)-py(j)
           
           select case (seed_kernel)
             case (2) !Weibull
               a_s = dexp(d_s - DLGAMA(dble(1) + dble(1) / b_s))
               ksi(i,j) = ((d/a_s)**(b_s-dble(2)))*dexp(-((d/a_s)**b_s))
             case (3) !Tufto
               if (d==dble(0)) d=dble(1)
               a_s = dble(1) / (dble(2)*dexp(d_s))
               ksi(i,j) = dexp(-(d*a_s)) / (d**dble(1.5))
             case (4) !Lognormal
               if (d==dble(0)) d=dble(1)
               a_s = dexp(d_s - (b_s**2)/dble(2))
               ksi(i,j) = dexp(-((dlog(d/a_s))**2)/(dble(2)*b_s**2))/d**2
             case (5) !Power-law
              a_s = dexp(d_s) * (b_s-dble(3))/dble(2)
               ksi(i,j) = dble(1)/((a_s+d)/a_s)**b_s
             case default !Exponential-power
               a_s = dexp(d_s + DLGAMA(dble(2) / b_s) - DLGAMA(dble(3) / b_s))
               ksi(i,j) = dexp(-((d/a_s)**b_s))
           end select
           
           if (d>dble(0)) ksi(i,j)=ksi(i,j)*dexp(theta(8)*(xc*dsin(my2pi*theta(10))&
             +yc*dcos(my2pi*theta(10)))/d)
           
           f=dble(0)
           if (fixedeffects) then
             do k=1,nf
               f=f+theta(11+k)*pzfen(1,j,k)
             enddo
           endif
           
         ksi(i,j)=pfem(j)*ksi(i,j)*dexp(f)*femfec(j)
         endif
        
         sksi(i) = sksi(i) + ksi(i,j)
       endif
     enddo
   endif
 end subroutine compute_ksi
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes the relative proportion of pollen from each candidate father expected for
 ! the i-th candidate mother's location (pollen profile)
 subroutine compute_phi(i,d_p,b_p)
   integer, intent(in) :: i
   double precision, intent(in) :: d_p,b_p
   integer :: j,k,pk
   double precision :: d,f,xc,yc,a_p
   
   if (pollen_kernel==0) then 
     pk=seed_kernel
   else
     pk=pollen_kernel
   endif
   
   sphi(i) = dble(0)
   do j=1,np
     
     phi(i,j)=dble(0)
     if (j==i) cycle
     
     if (pnb(i,j)==1) then
       if (dble(1)-pfem(j)>dble(0)) then
         d = dsqrt((px(j)-px(i))**2 + (py(j)-py(i))**2)
         xc=px(i)-px(j)
         yc=py(i)-py(j)
         
         
         select case (pk)
           case (2)
             a_p = dexp(d_p - DLGAMA(dble(1) + dble(1) / b_p))
             phi(i,j) = ((d/a_p)**(b_p-dble(2)))*dexp(-((d/a_p)**b_p))
           case (3)
             if (d==dble(0)) d=dble(1)
             a_p = dble(1) / (dble(2)*dexp(d_p))
             phi(i,j) = dexp(-(d*a_p)) / (d**dble(1.5))
           case (4)
             if (d==dble(0)) d=dble(1)
             a_p = dexp(d_p - (b_p**2)/dble(2))
             phi(i,j) = dexp(-((dlog(d/a_p))**2)/(dble(2)*b_p**2))/d**2
           case (5)
             a_p = dexp(d_p) * (b_p-dble(3))/dble(2)
             phi(i,j) = dble(1)/((a_p+d)/a_p)**b_p
           case default
             a_p = dexp(d_p + DLGAMA(dble(2) / b_p) - DLGAMA(dble(3) / b_p))
             phi(i,j) = dexp(-((d/a_p)**b_p))
         end select
         
         if (d>dble(0)) phi(i,j)=phi(i,j)*dexp(theta(9)*(xc*dsin(my2pi*theta(11))&
           +yc*dcos(my2pi*theta(11)))/d)
           
         f=dble(0)
         if (fixedeffects) then
           do k=1,nf
             d=pzfen(2,j,k)
             if (fen_mfdist(k)==1) d=dabs(d - pzfen(1,i,k))
             f=f+theta(11+nf+k)*d
           enddo
         endif
         
       phi(i,j)=(dble(1)-pfem(j))*phi(i,j)*dexp(f)*malfec(j)
       endif
       
     sphi(i) = sphi(i) + phi(i,j)
     endif
   enddo
 end subroutine compute_phi
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes a matrix of relative female and male fitness indices
 subroutine reproductive_success
   integer :: i,nth,pk
   double precision :: d_s,b_s,d_p,b_p,f_s,f_p
   
   nth=usernth
   
   if (pollen_kernel==0) then 
     if (incl(5)==0) theta(5) = theta(4)
     if (incl(7)==0) theta(7) = theta(6)
     pk=seed_kernel
   else
     pk=pollen_kernel
   endif
   
   select case (dist_typ)
     case(1)
       f_s = dlog(dble(1)/theta(4))
       f_p = dlog(dble(1)/theta(5))
     case default
       f_s = dlog(theta(4))
       f_p = dlog(theta(5))
   end select
      
   select case (seed_kernel)
     case (2)
       b_s = theta(6)
       d_s = f_s
     case (3)
       b_s = theta(6)
       d_s = f_s
     case (4)
       b_s = theta(6)
       d_s = f_s
     case (5)
       b_s = theta(6)
       d_s = f_s
     case default
       b_s = theta(6)
       d_s = f_s
   end select
   
   select case (pk)
     case (2)
       b_p = theta(7)
       d_p = f_p
     case (3)
       b_p = theta(7)
       d_p = f_p
     case (4)
       b_p = theta(7)
       d_p = f_p
     case (5)
       b_p = theta(7)
       d_p = f_p
     case default
       b_p = theta(7)
       d_p = f_p
   end select
   
   !$omp parallel num_threads(nth) private(i) &
   !$omp default(shared)
   !$omp do
   do i=1,no
     call compute_ksi(i,d_s,b_s)
   enddo
   !$omp end do
   !$omp end parallel
   
   !$omp parallel num_threads(nth) private(i) &
   !$omp default(shared)
   !$omp do
   do i=1,np
     call compute_phi(i,d_p,b_p)
   enddo
   !$omp end do
   !$omp end parallel
   
 end subroutine reproductive_success
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes a matrix of relative female fitness indices
 subroutine reproductive_success_ksi
   integer :: i,nth
   double precision :: d_s,b_s,f_s
   
   nth=usernth
   
   select case (dist_typ)
     case(1)
       f_s = dlog(dble(1)/theta(4))
     case default
       f_s = dlog(theta(4))
   end select
      
   select case (seed_kernel)
     case (2)
       b_s = theta(6)
       d_s = f_s
     case (3)
       b_s = theta(6)
       d_s = f_s
     case (4)
       b_s = theta(6)
       d_s = f_s
     case (5)
       b_s = theta(6)
       d_s = f_s
     case default
       b_s = theta(6)
       d_s = f_s
   end select
   
   !$omp parallel num_threads(nth) private(i) &
   !$omp default(shared)
   !$omp do
   do i=1,no
     call compute_ksi(i,d_s,b_s)
   enddo
   !$omp end do
   !$omp end parallel
   
 end subroutine reproductive_success_ksi
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes a matrix of relative male fitness indices
 subroutine reproductive_success_phi
   integer :: i,nth,pk
   double precision :: d_p,b_p,f_p
   
   nth=usernth
   
   if (pollen_kernel==0) then 
     if (incl(5)==0) theta(5) = theta(4)
     if (incl(7)==0) theta(7) = theta(6)
     pk=seed_kernel
   else
     pk=pollen_kernel
   endif
   
   select case (dist_typ)
     case(1)
       f_p = dlog(dble(1)/theta(5))
     case default
       f_p = dlog(theta(5))
   end select
      
   select case (pk)
     case (2)
       b_p = theta(7)
       d_p = f_p
     case (3)
       b_p = theta(7)
       d_p = f_p
     case (4)
       b_p = theta(7)
       d_p = f_p
     case (5)
       b_p = theta(7)
       d_p = f_p
     case default
       b_p = theta(7)
       d_p = f_p
   end select
   
   !$omp parallel num_threads(nth) private(i) &
   !$omp default(shared)
   !$omp do
   do i=1,np
     call compute_phi(i,d_p,b_p)
   enddo
   !$omp end do
   !$omp end parallel
   
 end subroutine reproductive_success_phi
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the likelihood function for the i-th offspring
 double precision function lik(i)
   integer, intent(in) :: i
   integer :: j,k,l,o1,o2,m1,m2,f1,f2
   double precision :: pro_b,pro_m,pro_mb,pro_mf,spro_mf,spro_m_,doup,dom,dof
   
   if (sksi(i)>dble(0)) then
     
     spro_m_=dble(0)
     
     doup = dble(1)
     if (ocy(i).ne.-1 .and. nhap>1) then
       doup = (dble(1)-mcytyp)*pcm(ocy(i))+mcytyp/dble(nhap-1)*(dble(1)-pcm(ocy(i)))
     endif
     
     pro_b=doup
     
     if (om(i)<0) then
       do l=1,nl
         o1=ogen(i,l,1)
         o2=ogen(i,l,2)
         if (o1 > -1) then
           pro_b=pro_b * pr_b(o1,o2,nall(l),mtyp(l),l)
         endif
       enddo
     else
       pro_b=dble(0)
     endif
     
     spro_m_=dble(0)
     do j=1,np
       
       dom=dble(1)
       if (ocy(i).ne.-1) then
         if (pcy(j)==-1) then
           dom = doup
         else
           dom = delta_eps(pcy(j),ocy(i),nhap,mcytyp)
         endif
       endif
       
       pro_mb=dble(dom) * maternal_cy + doup * (dble(1) - maternal_cy)
       pro_m=dble(dom)
       
       if (ksi(i,j)==dble(0) .or. sphi(j)==dble(0)) cycle
       
       if ((om(i)<0) .or. (om(i)==j)) then
         do l=1,nl
           if (pro_mb == dble(0)) then
             exit
           endif
           
           o1=ogen(i,l,1)
           o2=ogen(i,l,2)
           if (o1 > -1) then
             m1=pgen(j,l,1)
             m2=pgen(j,l,2)
             if (m1 > -1) then
               pro_m=pro_m * pr_mm(o1,o2,m1,m2,nall(l),mtyp(l))
               pro_mb=pro_mb * pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
             else
               pro_m=pro_m * pr_b(o1,o2,nall(l),mtyp(l),l)
               pro_mb=pro_mb * pr_b(o1,o2,nall(l),mtyp(l),l)
             endif
           endif
           
         enddo
       else
         pro_m=dble(0)
         pro_mb=dble(0)
       endif
       
       if ((om(i)>=0) .and. (of(i)>=0) .and. (om(i).ne.of(i))) pro_m=dble(0)
       if ((om(i)>=0) .and. (of(i)>=0)) pro_mb=dble(0)
       
       spro_mf=dble(0)
       do k=1,np
         
         dof = dble(1)
         if (ocy(i).ne.-1) then
           if (pcy(k)==-1) then
             dof = doup
           else
             dof = delta_eps(pcy(k),ocy(i),nhap,mcytyp)
           endif
         endif
         
         pro_mf=dble(dom) * maternal_cy + dble(dof) * (dble(1) - maternal_cy)
         if (k==j) pro_mf=dble(0)
          
         if ((om(i)<0 .and. of(i)<0) .or. (om(i)==j .and. (of(i)<0 .or. of(i)==k))) then
           do l=1,nl
             if (pro_mf == dble(0)) then
               exit
             endif
             o1=ogen(i,l,1)
             o2=ogen(i,l,2)
             if (o1 > -1) then
               m1=pgen(j,l,1)
               m2=pgen(j,l,2)
               f1=pgen(k,l,1)
               f2=pgen(k,l,2)
               if (m1 > -1) then
                 if (f1 > -1) then
                   pro_mf=pro_mf * pr_mf(o1,o2,m1,m2,f1,f2,nall(l),mtyp(l))
                 else
                   pro_mf=pro_mf * pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
                 endif
               else
                 if (f1 > -1) then
                   pro_mf=pro_mf * pr_mb(o1,o2,f1,f2,nall(l),mtyp(l),l)
                 else
                   pro_mf=pro_mf * pr_b(o1,o2,nall(l),mtyp(l),l)
                 endif
               endif
             endif
           enddo
         else
           pro_mf=dble(0)
         endif 
         spro_mf=spro_mf + phi(j,k) * pro_mf / sphi(j)
       enddo
       spro_m_=spro_m_ + ksi(i,j) * (theta(2) * pro_m + theta(3) * pro_mb + &
        (dble(1) - theta(2) - theta(3)) * spro_mf) / sksi(i)
     enddo
     
     lik=dble(odisp(i))*theta(1) * pro_b + (dble(1) - dble(odisp(i))*theta(1)) * spro_m_
   else
     lik=dble(1)
   endif
   
   proi(i)=lik
 end function lik
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the likelihood function for the i-th offspring based on constant transition probs
 double precision function c_lik(i)
   integer, intent(in) :: i
   integer :: j,k
   double precision :: spro_mf,spro_m_
   
   if (sksi(i)>dble(0)) then
     
     spro_m_=dble(0)
     spro_m_=dble(0)
     do j=1,np
       if (ksi(i,j)==dble(0) .or. sphi(j)==dble(0)) cycle
       spro_mf=dble(0)
       do k=1,np
         spro_mf=spro_mf + phi(j,k) * c_pro_mf(i,j,k) / sphi(j)
       enddo
       spro_m_=spro_m_ + ksi(i,j) * (theta(2) * c_pro_m(i,j) + theta(3) * c_pro_mb(i,j) + &
        (dble(1) - theta(2) - theta(3)) * spro_mf) / sksi(i)
     enddo
     
     c_lik=dble(odisp(i))*theta(1) * c_pro_b(i) + (dble(1) - dble(odisp(i))*theta(1)) * spro_m_
   else
     c_lik=dble(1)
   endif
   
   proi(i)=c_lik
 end function c_lik
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the full log-likelihood function for data
 double precision function full_logl()
   double precision :: l
   integer :: i
   
   if (linked_migr==1) theta(3)=theta(1)
   
   call reproductive_success
   
   l=dble(0)
   if (do_const_probs) then
     !$omp parallel num_threads(usernth) default(shared)
     !$omp do reduction(+:l)
     do i=1,nndg
       l = l + dlog(c_lik(ondg(i))) * dble(mtpl)
     enddo
     !$omp end do
     !$omp end parallel
     
     !$omp parallel num_threads(usernth) default(shared)
     !$omp do reduction(+:l)
     do i=1,ndg
       l = l + dlog(c_lik(odg(i))) * dble(mtpl)
     enddo
     !$omp end do
     !$omp end parallel
   else
     !$omp parallel num_threads(usernth) default(shared)
     !$omp do reduction(+:l)
     do i=1,nndg
       l = l + dlog(lik(ondg(i))) * dble(mtpl)
     enddo
     !$omp end do
     !$omp end parallel
     
     !$omp parallel num_threads(usernth) default(shared)
     !$omp do reduction(+:l)
     do i=1,ndg
       l = l + dlog(lik(odg(i))) * dble(mtpl)
     enddo
     !$omp end do
     !$omp end parallel
   endif
   
   full_logl = l / dble(mtpl)
 end function full_logl
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the simplified log-likelihood function for data (used in the Bayesian approach to
 ! estimate mtyp)
 double precision function full_logl_mf(k)
   double precision :: l
   integer :: i,k
   
   l=dble(0)
   !!$omp parallel num_threads(usernth) default(shared)
   !!$omp do reduction(+:l)
   do i=1,nndg
     l = l + dlog(lik_mf(ondg(i),k)) * dble(mtpl)
   enddo
   !!$omp end do
   !!$omp end parallel
   
   !!$omp parallel num_threads(usernth) default(shared)
   !!$omp do reduction(+:l)
   do i=1,ndg
     l = l + dlog(lik_mf(odg(i),k)) * dble(mtpl)
   enddo
   !!$omp end do
   !!$omp end parallel
   
   full_logl_mf = l / dble(mtpl)
 end function full_logl_mf
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the simplified log-likelihood function for data (used in the Bayesian approach to
 ! estimate mcytyp and maternal_cy)
 double precision function full_logl_mf_cy()
   double precision :: l
   integer :: i
   
   l=dble(0)
   !!$omp parallel num_threads(usernth) default(shared)
   !!$omp do reduction(+:l)
   do i=1,nndg
  l = l + dlog(lik_mf_cy(ondg(i))) * dble(mtpl)
   enddo
   !!$omp end do
   !!$omp end parallel
   
   !!$omp parallel num_threads(usernth) default(shared)
   !!$omp do reduction(+:l)
   do i=1,ndg
  l = l + dlog(lik_mf_cy(odg(i))) * dble(mtpl)
   enddo
   !!$omp end do
   !!$omp end parallel
   
   full_logl_mf_cy = l / dble(mtpl)
 end function full_logl_mf_cy
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the simplified likelihood function value for the i-th offspring given that parents
 ! are known (used in "full_logl_mf")
 double precision function lik_mf(i,l)
   integer, intent(in) :: i,l
   integer :: o1,o2,m1,m2,f1,f2,m,f
   double precision :: pr_o
   
   pr_o=1.D0
   o1=ogen(i,l,1)
   o2=ogen(i,l,2)
   if (sksi(i)>dble(0)) then !zero candidate mothers (wihtin the neighborhood)
     m=parass(i)%mo1
     f=parass(i)%fa1
     if (o1>-1) then !offspring genotype is known
       if (m>0) then !local mother assigned
         m1=pgen(m,l,1)
         m2=pgen(m,l,2)
         if (f>0 .and. f/=m) then !local father assigned
           f1=pgen(f,l,1)
           f2=pgen(f,l,2)
           if (m1>-1 .and. f1>-1) pr_o=pr_mf(o1,o2,m1,m2,f1,f2,nall(l),mtyp(l))
           if (m1>-1 .and. f1<0) pr_o=pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
           if (m1<0 .and. f1>-1) pr_o=pr_mb(o1,o2,f1,f2,nall(l),mtyp(l),l)
           if (m1<0 .and. f1<0) pr_o=pr_b(o1,o2,nall(l),mtyp(l),l)
           pr_o=pr_o*(1.D0-theta(1))*(1.D0-theta(2)-theta(3))*(phi(m,f)/sphi(m))*(ksi(i,m)/sksi(i))
         else if (f>0 .and. f==m) then !local self-fertilization
           if (m1>-1) pr_o=pr_mm(o1,o2,m1,m2,nall(l),mtyp(l))
           if (m1<0) pr_o=pr_b(o1,o2,nall(l),mtyp(l),l)
           pr_o=pr_o*(1.D0-theta(1))*theta(2)*(ksi(i,m)/sksi(i))
         else !unknown father
           if (m1>-1) then 
             pr_o=pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
           else
             pr_o=pr_b(o1,o2,nall(l),mtyp(l),l)
           endif
           pr_o=pr_o*(1.D0-theta(1))*theta(3)*(ksi(i,m)/sksi(i))
         endif
       else !unknown mother & father
         pr_o=theta(1)*pr_b(o1,o2,nall(l),mtyp(l),l)
       endif
     endif
   else
     if (o1>-1) then
       pr_o=1.D0
     endif
   endif
   
   lik_mf=pr_o
 end function lik_mf
 !----------------------------------------------------------------------------------------------------------
  
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the simplified likelihood function value for the i-th offspring given that parents
 ! are known (used in "full_logl_mf_cy")
 double precision function lik_mf_cy(i)
   integer, intent(in) :: i
   integer :: o1,m1,f1,m,f
   double precision :: pr_o,doup
   
   doup=1.D0
   pr_o=1.D0
   o1=ocy(i)
   if (o1>-1) doup=(dble(1)-mcytyp)*pcm(ocy(i))+mcytyp/dble(nhap-1)*(dble(1)-pcm(ocy(i)))
   if (sksi(i)>dble(0)) then !zero candidate mothers (wihtin the neighborhood)
     m=parass(i)%mo1
     f=parass(i)%fa1
     if (o1>-1) then !offspring genotype is known
       if (m>0) then !local mother assigned
         m1=pcy(m)
         if (f>0 .and. f/=m) then !local father assigned
           f1=pcy(f)
           if (m1>-1 .and. f1>-1) pr_o=delta_eps(pcy(m),ocy(i),nhap,mcytyp)*maternal_cy+delta_eps(pcy(f),ocy(i),nhap,mcytyp)&
            *(dble(1)-maternal_cy)
           if (m1>-1 .and. f1<0) pr_o=delta_eps(pcy(m),ocy(i),nhap,mcytyp)*maternal_cy+(dble(1)-maternal_cy)*doup
           if (m1<0 .and. f1>-1) pr_o=doup*maternal_cy+(dble(1)-maternal_cy)*delta_eps(pcy(f),ocy(i),nhap,mcytyp)
           if (m1<0 .and. f1<0) pr_o=doup
           pr_o=pr_o*(1.D0-theta(1))*(1.D0-theta(2)-theta(3))*(phi(m,f)/sphi(m))*(ksi(i,m)/sksi(i))
         else  if (f>0 .and. f==m) then !self-fertilization
           if (m1>-1) then
             pr_o=delta_eps(pcy(m),ocy(i),nhap,mcytyp)
           else
             pr_o=doup
           endif
           pr_o=pr_o*(1.D0-theta(1))*theta(2)*(ksi(i,m)/sksi(i))
         else !unknown father
           if (m1>-1) then 
             pr_o=delta_eps(pcy(m),ocy(i),nhap,mcytyp)*maternal_cy+(dble(1)-maternal_cy)*doup
           else
             pr_o=doup
           endif
           pr_o=pr_o*(1.D0-theta(1))*theta(3)*(ksi(i,m)/sksi(i))
         endif
       else !unknown mother & father
         pr_o=doup*theta(1)
       endif
     endif
   else
     if (o1>-1) then
       pr_o=1.D0
     endif
   endif
  
   lik_mf_cy=pr_o
 end function lik_mf_cy
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes constant transition probs for the i-th offspring
 subroutine const_probs(i)
   integer, intent(in) :: i
   integer :: j,k,l,o1,o2,m1,m2,f1,f2
   double precision :: doup,dom,dof
   
   if (sksi(i)>dble(0)) then
     doup = dble(1)
     if (ocy(i).ne.-1 .and. nhap>1) then
       doup = (dble(1)-mcytyp)*pcm(ocy(i))+mcytyp/dble(nhap-1)*(dble(1)-pcm(ocy(i)))
     endif
     
     c_pro_b(i)=doup
     
     if (om(i)<0) then
       do l=1,nl
         o1=ogen(i,l,1)
         o2=ogen(i,l,2)
         if (o1 > -1) then
           c_pro_b(i)=c_pro_b(i) * pr_b(o1,o2,nall(l),mtyp(l),l)
         endif
       enddo
     else
       c_pro_b(i)=dble(0)
     endif
     
     do j=1,np
       
       dom=dble(1)
       if (ocy(i).ne.-1) then
         if (pcy(j)==-1) then
           dom = doup
         else
           dom = delta_eps(pcy(j),ocy(i),nhap,mcytyp)
         endif
       endif
       
       c_pro_mb(i,j)=dble(dom) * maternal_cy + doup * (dble(1) - maternal_cy)
       c_pro_m(i,j)=dble(dom)
       
       if (ksi(i,j)==dble(0) .or. sphi(j)==dble(0)) cycle
       
       if ((om(i)<0) .or. (om(i)==j)) then
         do l=1,nl
           if (c_pro_mb(i,j) == dble(0)) then
             exit
           endif
           o1=ogen(i,l,1)
           o2=ogen(i,l,2)
           if (o1 > -1) then
             m1=pgen(j,l,1)
             m2=pgen(j,l,2)
             if (m1 > -1) then
               c_pro_m(i,j)=c_pro_m(i,j) * pr_mm(o1,o2,m1,m2,nall(l),mtyp(l))
               c_pro_mb(i,j)=c_pro_mb(i,j) * pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
             else
               c_pro_m(i,j)=c_pro_m(i,j) * pr_b(o1,o2,nall(l),mtyp(l),l)
               c_pro_mb(i,j)=c_pro_mb(i,j) * pr_b(o1,o2,nall(l),mtyp(l),l)
             endif
           endif
         enddo
       else
         c_pro_m(i,j)=dble(0)
         c_pro_mb(i,j)=dble(0)
       endif
       
       if ((om(i)>=0) .and. (of(i)>=0) .and. (om(i).ne.of(i))) c_pro_m(i,j)=dble(0)
       if ((om(i)>=0) .and. (of(i)>=0)) c_pro_mb(i,j)=dble(0)
       do k=1,np
         dof = dble(1)
         if (ocy(i).ne.-1) then
           if (pcy(k)==-1) then
             dof = doup
           else
             dof = delta_eps(pcy(k),ocy(i),nhap,mcytyp)
           endif
         endif
         
         c_pro_mf(i,j,k)=dble(dom) * maternal_cy + dble(dof) * (dble(1) - maternal_cy)
         if (k==j) c_pro_mf(i,j,k)=dble(0)
          
         if ((om(i)<0 .and. of(i)<0) .or. (om(i)==j .and. (of(i)<0 .or. of(i)==k))) then
           do l=1,nl
             if (c_pro_mf(i,j,k) == dble(0)) then
               exit
             endif
             o1=ogen(i,l,1)
             o2=ogen(i,l,2)
             if (o1 > -1) then
               m1=pgen(j,l,1)
               m2=pgen(j,l,2)
               f1=pgen(k,l,1)
               f2=pgen(k,l,2)
               if (m1 > -1) then
                 if (f1 > -1) then
                   c_pro_mf(i,j,k)=c_pro_mf(i,j,k) * pr_mf(o1,o2,m1,m2,f1,f2,nall(l),mtyp(l))
                 else
                   c_pro_mf(i,j,k)=c_pro_mf(i,j,k) * pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
                 endif
               else
                 if (f1 > -1) then
                   c_pro_mf(i,j,k)=c_pro_mf(i,j,k) * pr_mb(o1,o2,f1,f2,nall(l),mtyp(l),l)
                 else
                   c_pro_mf(i,j,k)=c_pro_mf(i,j,k) * pr_b(o1,o2,nall(l),mtyp(l),l)
                 endif
               endif
             endif
           enddo
         else
           c_pro_mf(i,j,k)=dble(0)
         endif 
       enddo
     enddo
   endif
 end subroutine const_probs
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine makes some initializations (e.g. memory allocation)
 subroutine initialize_model
   integer :: i,k
   character(len=5) :: mystring
   
   
   if (linked_migr==1) then
     incl(3)=0
     theta(3)=theta(1)
   endif
   
   allocate(ntheta(11+2*nf))
   ntheta(1)='    ms    '
   ntheta(2)='    s     '
   ntheta(3)='    mp    '
   ntheta(4)='    ds    '
   ntheta(5)='    dp    '
   ntheta(6)='    bs    '
   ntheta(7)='    bp    '
   ntheta(8)='    ks    '
   ntheta(9)='    kp    '
   ntheta(10)='    as    '
   ntheta(11)='    ap    '
   
   do i=12,11+nf
     write(mystring,"(i5)") (i-11)
     ntheta(i)='    g'//adjustl(mystring)
     ntheta(i+nf)='    b'//adjustl(mystring)
   enddo
   
   allocate(ksi(no,np))
   allocate(phi(np,np))
   allocate(sksi(no))
   allocate(sphi(np))
   allocate(vmtyp(nl))
   allocate(vtheta(11+2*nf))
   allocate(grad(11+2*nf))
   allocate(hess(11+2*nf,11+2*nf))
   
   do i=1,nl
     if (nall(i) == 1) then
       mtyp(i)=dble(0.0)
     endif
     vmtyp(i)=dble(-1)
   enddo
   vmcytyp=dble(-1)
   
   allocate(femfec(np))
   allocate(mlogfemfec(np))
   allocate(malfec(np))
   allocate(mlogmalfec(np))
   do i=1,np
     femfec(i)=1.D0
     malfec(i)=1.D0
     mlogfemfec(i)=0.D0
     mlogmalfec(i)=0.D0
     do k=1,nf
       if (pfem(i)>0.D0) mlogfemfec(i)=mlogfemfec(i)+theta(11+k)*pzfen(1,i,k)
       if (pfem(i)<1.D0) mlogmalfec(i)=mlogmalfec(i)+theta(11+nf+k)*pzfen(2,i,k)
     enddo
   enddo
   
   femsigm=0.5D0
   malsigm=0.5D0
   
   allocate(proi(no))
   
   ar_g=0.D0; rr_g=0.D0; v_g=0.1D0
   ar_b=0.D0; rr_b=0.D0; v_b=0.1D0
   ar_ds=0.D0; rr_ds=0.D0; v_ds=0.1D0
   ar_dp=0.D0; rr_dp=0.D0; v_dp=0.1D0
   ar_bs=0.D0; rr_bs=0.D0; v_bs=0.1D0
   ar_bp=0.D0; rr_bp=0.D0; v_bp=0.1D0
   ar_ks=0.D0; rr_ks=0.D0; v_ks=0.1D0
   ar_kp=0.D0; rr_kp=0.D0; v_kp=0.1D0
   ar_as=0.D0; rr_as=0.D0; v_as=0.1D0
   ar_ap=0.D0; rr_ap=0.D0; v_ap=0.1D0
   ar_ff=0.D0; rr_ff=0.D0; v_ff=0.1D0
   ar_fm=0.D0; rr_fm=0.D0; v_fm=0.1D0
   ar_sf=0.D0; rr_sf=0.D0; v_sf=0.1D0
   ar_sm=0.D0; rr_sm=0.D0; v_sm=0.1D0
   allocate(ar_e(nl))
   allocate(rr_e(nl))
   allocate(d_e(nl))
   ar_e=0.D0; rr_e=0.D0; d_e=0.1D0
   ar_ecy=0.D0; rr_ecy=0.D0; d_ecy=0.1D0
   ar_mcy=0.D0; rr_mcy=0.D0; d_mcy=0.1D0
   m_e=0.5D0
   y_e=1.D0/3.D0
   ar_m_e=0.D0; rr_m_e=0.D0; d_m_e=0.1D0
   ar_y_e=0.D0; rr_y_e=0.D0; d_y_e=0.1D0
   
   call reproductive_success
   
   if (incl_mtyp==0 .and. incl_mcytyp==0 .and. incl_maternal_cy==0) then
     do_const_probs=.true.
     
     allocate(c_pro_b(no))
     allocate(c_pro_mb(no,np))
     allocate(c_pro_m(no,np))
     allocate(c_pro_mf(no,np,np))
     
     !$omp parallel num_threads(usernth) default(shared)
     !$omp do
     do i=1,nndg
       call const_probs(ondg(i))
     enddo
     !$omp end do
     !$omp end parallel
     
     !$omp parallel num_threads(usernth) default(shared)
     !$omp do
     do i=1,ndg
       call const_probs(odg(i))
     enddo
     !$omp end do
     !$omp end parallel
   else
     do_const_probs=.false.
   endif
   
 end subroutine initialize_model
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes allele frequencies based on genetic data
 subroutine allele_frequencies
   integer :: i,l,k,a,nh
   integer :: n(nl)
   
   allocate(plm(nl,maxl))
   allocate(pcm(nhap))
   pcm=dble(0)
   nh=0
   n=0
   plm=dble(0)
   
   do i=1,np
     do l=1,nl
       do k=1,2
         a=pgen(i,l,k)
         if (a > -1) then
           plm(l,a)=plm(l,a)+1.D0
           n(l)=n(l)+1
         endif
       enddo
     enddo
     
     a=pcy(i)
     if (a > -1) then
       pcm(a)=pcm(a)+1.D0
       nh=nh+1
     endif
   enddo
   
   ! alleles not present in parents have tiny proportion of order 1/2n
   do l=1,nl
     do k=1,nall(l)
       if (plm(l,k)==dble(0)) then
         plm(l,k) = 1.D0
         n(l)=n(l) + 1
       endif
     enddo
   enddo
   
   do k=1,nhap
     if (pcm(k)==dble(0)) then
       pcm(k) = 1.D0
       nh=nh+1
     endif
   enddo
   
   do l=1,nl
     plm(l,:) = plm(l,:) / n(l)
   enddo
   
   pcm=pcm/nh
   
 end subroutine allele_frequencies
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes the gradient and the Hessian for the parameter vector (theta)
 ! (derivatives are approximated using finite differences)
 subroutine derivatives(full_hess)
   ! numerical derivatives computed as finite differences
   integer, intent(in) :: full_hess
   integer :: i,j,k
   double precision, allocatable :: fu(:),fl(:),dt(:),th(:)
   double precision :: f,f2u
   
   write(*,"(a)",advance='no') 'Optimizing parameters:'
   
   f=full_logl()
   
   allocate(th(11+2*nf))
   allocate(fu(11+2*nf))
   allocate(fl(11+2*nf))
   allocate(dt(11+2*nf))
   
   th=theta
   
   do i=1,11+2*nf
     if (incl(i)==0) cycle
     do j=1,11+2*nf
       dt(j)=dble(0)
       theta(j)=th(j)
     enddo
     dt(i)=dabs(theta(i)) * mep
     theta(i)=th(i)+dt(i)
     fu(i)=full_logl()
     
     theta(i)=th(i)-dt(i)
     fl(i)=full_logl()
     
     grad(i)=(fu(i)-fl(i)) / (dble(2) * dt(i))
     hess(i,i)=(fu(i) + fl(i) - dble(2) * f) / dt(i)**2
     write(*,"(i2,a2)",advance='no') i*100/(2*(11+2*nf)),'% '
   enddo
   
   do i=1,11+2*nf
     if (incl(i)==0) cycle
     do j=i+1,11+2*nf
       if (incl(j)==0) cycle
       do k=1,11+2*nf
         dt(k)=dble(0)
         theta(k)=th(k)
       enddo
       dt(i)=dabs(theta(i)) * mep
       dt(j)=dabs(theta(j)) * mep
       
       theta(i)=th(i)+dt(i)
       theta(j)=th(j)+dt(j)
       hess(i,j)=dble(0)
       if (full_hess == 1) then
         f2u=full_logl()
         hess(i,j)=(f2u - fu(i) - fu(j) + f) / (dt(i) * dt(j))
       endif
       hess(j,i)=hess(i,j)
     enddo
     if (i<11+2*nf) write(*,"(i2,a2)",advance='no') (i+11+2*nf)*100/(2*(11+2*nf)),'% '
   enddo
 
 write (*,"(a)") '100%'
   
 end subroutine derivatives
 !----------------------------------------------------------------------------------------------------------
 
 !-----------------------------------------------------------------------------------------------------------
 ! This routine computes Hessian^-1
 subroutine invert_hess
   integer :: ia,ja,ka
 
   do ka=1,11+2*nf
     if (incl(ka)==0) cycle
     hess(ka,ka)=dble(1)/hess(ka,ka)
     do ia=1,11+2*nf
       if (ia.ne.ka) then
         if (incl(ia)==0) cycle
         hess(ia,ka)=-hess(ia,ka)*hess(ka,ka)
       endif
     enddo
     do ja=1,11+2*nf
       if (incl(ja)==0) cycle
       if (ja.ne.ka) then
         do ia=1,11+2*nf
           if (ia.ne.ka) then
             if (incl(ia)==0) cycle
             hess(ia,ja)=hess(ia,ja)+hess(ia,ka)*hess(ka,ja)
           endif
         enddo
       hess(ka,ja)=hess(ka,ja)*hess(ka,ka)
       endif
     enddo
   enddo
   
 end subroutine invert_hess
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine performs a single step of the Newton optimization algorithm for the parameter vector theta
 subroutine update_theta
   integer :: i,j
   double precision, allocatable :: dt(:),t(:)
   
   allocate(dt(11+2*nf))
   allocate(t(11+2*nf))
   
   do i=1,11+2*nf
     t(i)=theta(i)
     vtheta(i)=dble(-1)
     
     if (incl(i)==0) cycle
     dt(i)=dble(0)
     
     do j=1,11+2*nf
       if (incl(j)==0) cycle
       dt(i)=dt(i)+hess(j,i)*grad(j)
     enddo
     
     t(i)=t(i)-dt(i)
     if (i > 7) then
       theta(i)=t(i)
       
       if (i==10 .or. i==11) then
          theta(i)=theta(i)-floor(theta(i))
       endif
     endif
     
     if (i < 8) then
       if (t(i) > minth(i) .and. t(i) < maxth(i)) then
         theta(i)=t(i)
       else
         if (t(i) < minth(i)) theta(i)=theta(i)-(theta(i)-minth(i)) / dble(2)
         if (t(i) > maxth(i)) theta(i)=theta(i)+(maxth(i)-theta(i)) / dble(2)
         if (dabs(theta(i)-minth(i))<0.001) then
           theta(i)=minth(i)
           incl(i)=0
         endif
         if (dabs(theta(i)-maxth(i))<0.001) then
           theta(i)=maxth(i)
           incl(i)=0
         endif
       endif
     endif
     
     if (hess(i,i) < dble(0)) vtheta(i)=dsqrt(-hess(i,i))
   enddo
   
 end subroutine update_theta
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine performs a single step of the Newton optimization algorithm for the parameter vector mtyp
 subroutine update_mtyp
   double precision, allocatable :: grad(:),hess(:),x(:)
   integer :: l,c
   double precision :: f,fu,fl,dx,nx
   
   f=full_logl()
   
   allocate(grad(nl))
   allocate(hess(nl))
   allocate(x(nl))
   
   x=mtyp
   vmtyp=dble(-1)
   
   write(*,"(a)",advance='no') 'Updating mtyp for loc: '
   c=0
   do l=1,nl
     dx = dabs(x(l)) * mep
     
     if (nall(l) == 1 .or. dx == 0) cycle
     c=c+1
     mtyp(l)=x(l) + dx
     fu = full_logl()
     mtyp(l)=x(l) - dx
     fl = full_logl()
     grad(l) = (fu - fl) / (dx + dx)
     hess(l) = (fu + fl - f - f) / dx**2
     
     mtyp(l) = x(l)
     write(*,"((i0),(a))",advance='no') l,' '
     
   enddo
   if (c==0) then
     write(*,"(a)") 'all skipped'
   else
     write(*,*)
   endif
   
   do l=1,nl
     if (nall(l) == 1 .or. x(l) == 0.D0) cycle
     
     nx = x(l) - grad(l)/hess(l)
     if (nx <= 0.D0) nx = x(l) * 0.5D0
     if (nx >= 1.D0) nx = x(l) + (1.D0 - x(l)) * 0.5D0
     if (nx > 0.D0 .and. nx < 1.D0) mtyp(l) = nx
     if (mtyp(l) < minmtyp) mtyp(l) = 0.D0
     
     if (hess(l) .ne. 0.D0) vmtyp(l)=dsqrt(-1.D0/hess(l))
   enddo
   
   
 end subroutine update_mtyp
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine performs a single step of the Newton optimization algorithm for the parameter mcytyp
 subroutine update_mcytyp
   double precision :: grad,hess,x
   integer :: c
   double precision :: f,fu,fl,dx,nx
   
   f=full_logl()
   
   x=mcytyp
   vmcytyp=dble(-1)
   
   grad=dble(0)
   hess=dble(1)
   
   write(*,"(a)",advance='no') 'Updating mcytyp: '
   c=0
   dx = dabs(x) * mep
     
   if (nhap > 1 .and. dx > 0.D0) then
     c=1
     mcytyp=x + dx
     fu = full_logl()
     mcytyp=x - dx
     fl = full_logl()
     grad = (fu - fl) / (dx + dx)
     hess = (fu + fl - f - f) / dx**2
     
     mcytyp = x
   endif
   
   if (c==0) then
     write(*,"(a)") 'skipped'
   else
     write(*,"(a)") 'done'
   endif
   
   nx = x - grad/hess
   if (nx <= 0.D0) nx = x * 0.5D0
   if (nx >= 1.D0) nx = x + (1.D0 - x) * 0.5D0
   if (nx > 0.D0 .and. nx < 1.D0) mcytyp = nx
   if (mcytyp < minmtyp) mcytyp = 0.D0
   if (hess < 0.D0) vmcytyp=dsqrt(-1/hess)
   
 end subroutine update_mcytyp
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine performs a single step of the Newton optimization algorithm for the parameter maternal_cy
 subroutine update_maternal_cy
   double precision :: grad,hess,x
   integer :: c
   double precision :: f,fu,fl,dx,nx
   
   f=full_logl()
   
   
   x=maternal_cy
   vmaternal_cy=dble(-1)
   
   grad=dble(0)
   hess=dble(1)
   
   write(*,"(a)",advance='no') 'Updating maternal_cy: '
   c=0
   dx = dabs(x) * mep
     
   if (nhap > 1 .and. dx > 0.D0) then
     c=1
     maternal_cy=x + dx
     fu = full_logl()
     maternal_cy=x - dx
     fl = full_logl()
     grad = (fu - fl) / (dx + dx)
     hess = (fu + fl - f - f) / dx**2
     
     maternal_cy = x
   endif
   
   if (c==0) then
     write(*,"(a)") 'skipped'
   else
     write(*,"(a)") 'done'
   endif
   
   nx = x - grad/hess
   if (nx <= 0) nx = x * 0.5D0
   if (nx >= 1) nx = x + (1.D0 - x) * 2.D0
   if (nx > 0.D0 .and. nx < 1.D0) maternal_cy = nx
   if (maternal_cy < minmcy) maternal_cy = dble(0)
     
   if (hess < dble(0)) vmaternal_cy=dsqrt(-1.D0/hess)
   
 end subroutine update_maternal_cy
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes two the most likely genealogies for the i-th offspring (ML approach)
 subroutine conditional_pr(i)
   integer, intent(in) :: i
   integer :: j,k,l,o1,o2,m1,m2,f1,f2
   double precision :: pro_b,spro_mf,spro_m_,doup,pro,dom,dof
   double precision, allocatable :: pro_m(:),pro_mb(:),pro_mf(:,:)
   double precision :: prb,prm,prgtmp
   integer, dimension(2,2) :: mf
   double precision, dimension(2) :: prg
   
   spro_m_=dble(0)
   
   doup = dble(1)
   if (ocy(i).ne.-1 .and. nhap>1) doup = (dble(1)-mcytyp)*pcm(ocy(i))+mcytyp/dble(nhap-1)*(dble(1)-pcm(ocy(i)))
   
   pro_b=doup
   
   allocate(pro_m(np))
   allocate(pro_mb(np))
   allocate(pro_mf(np,np))
   
   if (om(i)<0) then  
     do l=1,nl
       o1=ogen(i,l,1)
       o2=ogen(i,l,2)
       if (o1 > -1) pro_b=pro_b * pr_b(o1,o2,nall(l),mtyp(l),l)
     enddo
   else
     pro_b=dble(0)
   endif
   
   spro_m_=dble(0)
   do j=1,np
     if (onb(i,j)==1 .or. om(i)==j) then
       
       dom=dble(1)
       if (ocy(i).ne.-1) then
         if (pcy(j)==-1) then
           dom = doup
         else
           dom = delta_eps(pcy(j),ocy(i),nhap,mcytyp)
         endif
       endif
       
       pro_mb(j)=dble(dom) * maternal_cy + doup * (dble(1) - maternal_cy)
       pro_m(j)=dble(dom)
       
       if ((om(i)<0) .or. (om(i)==j)) then
         do l=1,nl
           if (pro_mb(j) == dble(0)) then
             exit
           endif
           
           o1=ogen(i,l,1)
           o2=ogen(i,l,2)
           if (o1 > -1) then
             m1=pgen(j,l,1)
             m2=pgen(j,l,2)
             if (m1 > -1) then
               pro_m(j)=pro_m(j) * pr_mm(o1,o2,m1,m2,nall(l),mtyp(l))
               pro_mb(j)=pro_mb(j) * pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
             else
               pro_m(j)=pro_m(j) * pr_b(o1,o2,nall(l),mtyp(l),l)
               pro_mb(j)=pro_mb(j) * pr_b(o1,o2,nall(l),mtyp(l),l)
             endif
           endif
           
         enddo
       else
         pro_m(j)=dble(0)
         pro_mb(j)=dble(0)
       endif
       
       if ((om(i)>=0) .and. (of(i)>=0) .and. (om(i).ne.of(i))) pro_m(j)=dble(0)
       if ((om(i)>=0) .and. (of(i)>=0)) pro_mb(j)=dble(0)
            
       spro_mf=dble(0)
       do k=1,np
         if (pnb(j,k)==1) then
           dof = dble(1)
           if (ocy(i).ne.-1) then
             if (pcy(k)==-1) then
               dof = doup
             else
               dof = delta_eps(pcy(k),ocy(i),nhap,mcytyp)
             endif
           endif
         
           pro_mf(j,k)=dble(dom) * maternal_cy + dble(dof) * (dble(1) - maternal_cy)
           
           if (k==j) pro_mf(j,k)=dble(0)
           
           if ((om(i)<0 .and. of(i)<0) .or. (om(i)==j .and. (of(i)<0 .or. of(i)==k))) then
             do l=1,nl
               if (pro_mf(j,k) == dble(0)) then
                 exit
               endif
               o1=ogen(i,l,1)
               o2=ogen(i,l,2)
               if (o1 > -1) then
                 m1=pgen(j,l,1)
                 m2=pgen(j,l,2)
                 f1=pgen(k,l,1)
                 f2=pgen(k,l,2)
                 if (m1 > -1) then
                   if (f1 > -1) then
                     pro_mf(j,k)=pro_mf(j,k) * pr_mf(o1,o2,m1,m2,f1,f2,nall(l),mtyp(l))
                   else
                     pro_mf(j,k)=pro_mf(j,k) * pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
                   endif
                 else
                   if (f1 > -1) then
                     pro_mf(j,k)=pro_mf(j,k) * pr_mb(o1,o2,f1,f2,nall(l),mtyp(l),l)
                   else
                     pro_mf(j,k)=pro_mf(j,k) * pr_b(o1,o2,nall(l),mtyp(l),l)
                   endif
                 endif
               endif
             enddo
           else
             pro_mf(j,k)=dble(0)
           endif
             
           spro_mf=spro_mf + phi(j,k) * pro_mf(j,k) / sphi(j)
         endif
       enddo
       spro_m_=spro_m_ + ksi(i,j) * (theta(2) * pro_m(j) + theta(3) * pro_mb(j) + &
       (dble(1) - theta(2) - theta(3)) * spro_mf) / sksi(i)
     endif
   enddo
   
   ! identify two the most likely genealogies
   pro=dble(odisp(i))*theta(1) * pro_b + (dble(1) - dble(odisp(i))*theta(1)) * spro_m_
   
   prb=dble(odisp(i))*theta(1) * pro_b / pro
   prm=spro_m_
   
   prg(1)=prb
   prg(2)=0.D0
   mf(1,1)=-1
   mf(1,2)=-1
   mf(2,1)=-1
   mf(2,2)=-1
   
   do j=1,np
     if (onb(i,j)==1 .or. om(i)==j) then
       !selfing with j-th parent
       prgtmp=(dble(1) - dble(odisp(i))*theta(1)) * ksi(i,j)/sksi(i) * theta(2) * pro_m(j) / pro
       if (prgtmp > prg(1)) then
         prg(2)=prg(1)
         prg(1)=prgtmp
         mf(2,1)=mf(1,1)
         mf(2,2)=mf(1,2)
         mf(1,1)=j
         mf(1,2)=j
       else
         if (prgtmp > prg(2)) then
           prg(2)=prgtmp
           mf(2,1)=j
           mf(2,2)=j
         endif
       endif
       
       !pollen mmigrant with j-th parent
       prgtmp=(dble(1) - dble(odisp(i))*theta(1)) * ksi(i,j)/sksi(i) * theta(3) * pro_mb(j) / pro
       if (prgtmp > prg(1)) then
         prg(2)=prg(1)
         prg(1)=prgtmp
         mf(2,1)=mf(1,1)
         mf(2,2)=mf(1,2)
         mf(1,1)=j
         mf(1,2)=-1
       else
         if (prgtmp > prg(2)) then
           prg(2)=prgtmp
           mf(2,1)=j
           mf(2,2)=-1
         endif
       endif
       
       !local outcrossing with j-th and k-th parent
       do k=1,np
         if (pnb(j,k)==1) then
           if (pro_mf(j,k) == dble(0)) cycle
           if (k==j) cycle
           
           prgtmp=(dble(1) - dble(odisp(i))*theta(1)) * ksi(i,j)/sksi(i) * (dble(1) - theta(2) - theta(3)) &
            * phi(j,k) * pro_mf(j,k)/sphi(j) / pro
           if (prgtmp > prg(1)) then
             prg(2)=prg(1)
             prg(1)=prgtmp
             mf(2,1)=mf(1,1)
             mf(2,2)=mf(1,2)
             mf(1,1)=j
             mf(1,2)=k
           else
             if (prgtmp > prg(2)) then
               prg(2)=prgtmp
               mf(2,1)=j
               mf(2,2)=k
             endif
           endif
         endif
       enddo
     endif
   enddo
   
   if (mf(1,1) > -1) mf(1,1)=pid(mf(1,1))
   if (mf(1,2) > -1) mf(1,2)=pid(mf(1,2))
   if (mf(2,1) > -1) mf(2,1)=pid(mf(2,1))
   if (mf(2,2) > -1) mf(2,2)=pid(mf(2,2))
   
   parass(i)%oid=oid(i)
   parass(i)%mo1=mf(1,1)
   parass(i)%fa1=mf(1,2)
   parass(i)%pr1=prg(1)
   parass(i)%mo2=mf(2,1)
   parass(i)%fa2=mf(2,2)
   parass(i)%pr2=prg(2)
   
 end subroutine conditional_pr
 !----------------------------------------------------------------------------------------------------------
  
 !----------------------------------------------------------------------------------------------------------
 ! This routine creates a text file containing two the most likely offspring genealogies (ML approach)
 subroutine conditional_probs
   integer :: i
   
   call reproductive_success
   allocate(parass(no))
   
   !$omp parallel num_threads(usernth) default(shared)
   !$omp do
   do i=1,nndg
     call conditional_pr(ondg(i))
   enddo
   !$omp end do
   !$omp end parallel
   
   !$omp parallel num_threads(usernth) default(shared)
   !$omp do
   do i=1,ndg
     call conditional_pr(odg(i))
   enddo
   !$omp end do
   !$omp end parallel
   
   outfile = 90
   open(unit=outfile,file=fp,action='write',status='replace')
   write(outfile,"(a1,7(a10))") ' ','Prog','Mo1','Fa1','Pr1','Mo2','Fa2','Pr2'   
   do i=1,no
     write(outfile,"(i10,2(2(i10),f10.4))") parass(i)%oid,parass(i)%mo1,parass(i)%fa1,parass(i)%pr1, &
          parass(i)%mo2,parass(i)%fa2,parass(i)%pr2
     flush(outfile)
   enddo
   close(outfile)
   
 end subroutine conditional_probs
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes posterior probability that the i-th offspring results from seed and pollen immigration
!~  subroutine migr_pr(i)
!~    integer, intent(in) :: i
!~    integer :: j,k,l,o1,o2,m1,m2,f1,f2
!~    double precision :: pro_b,spro_mf,spro_mb,spro_m_,doup,pro,dom,dof
!~    double precision, allocatable :: pro_m(:),pro_mb(:),pro_mf(:,:)
!~    double precision :: prb
   
!~    spro_m_=dble(0)
   
!~    doup = dble(1)
!~    if (ocy(i).ne.-1 .and. nhap>1) then
!~      doup = (dble(1)-mcytyp)*pcm(ocy(i))+mcytyp/dble(nhap-1)*(dble(1)-pcm(ocy(i)))
!~    endif
   
!~    pro_b=doup
   
!~    allocate(pro_m(np))
!~    allocate(pro_mb(np))
!~    allocate(pro_mf(np,np))
   
!~    if (om(i)<0) then  
!~      do l=1,nl
!~        o1=ogen(i,l,1)
!~        o2=ogen(i,l,2)
!~        if (o1 > -1) then
!~          pro_b=pro_b * pr_b(o1,o2,nall(l),mtyp(l),l)
!~        endif
!~      enddo
!~    else
!~      pro_b=dble(0)
!~    endif
   
!~    spro_m_=dble(0)
!~    spro_mb=dble(0)
!~    do j=1,np
     
!~      dom=dble(1)
!~      if (ocy(i).ne.-1) then
!~        if (pcy(j)==-1) then
!~          dom = doup
!~        else
!~          dom = delta_eps(pcy(j),ocy(i),nhap,mcytyp)
!~        endif
!~      endif
     
!~      pro_mb(j)=dble(dom) * maternal_cy + doup * (dble(1) - maternal_cy)
!~      pro_m(j)=dble(dom)
     
!~      if ((om(i)<0) .or. (om(i)==j)) then
!~        do l=1,nl
!~          if (pro_mb(j) == dble(0)) then
!~            exit
!~          endif
         
!~          o1=ogen(i,l,1)
!~          o2=ogen(i,l,2)
!~          if (o1 > -1) then
!~            m1=pgen(j,l,1)
!~            m2=pgen(j,l,2)
!~            if (m1 > -1) then
!~              pro_m(j)=pro_m(j) * pr_mm(o1,o2,m1,m2,nall(l),mtyp(l))
!~              pro_mb(j)=pro_mb(j) * pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
!~            else
!~              pro_m(j)=pro_m(j) * pr_b(o1,o2,nall(l),mtyp(l),l)
!~              pro_mb(j)=pro_mb(j) * pr_b(o1,o2,nall(l),mtyp(l),l)
!~            endif
!~          endif
         
!~        enddo
!~      else
!~        pro_m(j)=dble(0)
!~        pro_mb(j)=dble(0)
!~      endif
     
!~      if ((om(i)>=0) .and. (of(i)>=0) .and. (om(i).ne.of(i))) pro_m(j)=dble(0)
!~      if ((om(i)>=0) .and. (of(i)>=0)) pro_mb(j)=dble(0)
          
!~      spro_mf=dble(0)
!~      do k=1,np
        
!~        dof = dble(1)
!~        if (ocy(i).ne.-1) then
!~          if (pcy(k)==-1) then
!~            dof = doup
!~          else
!~            dof = delta_eps(pcy(k),ocy(i),nhap,mcytyp)
!~          endif
!~        endif
       
!~        pro_mf(j,k)=dble(dom) * maternal_cy + dble(dof) * (dble(1) - maternal_cy)
       
!~        if (k==j) pro_mf(j,k)=dble(0)
       
!~        if ((om(i)<0 .and. of(i)<0) .or. (om(i)==j .and. (of(i)<0 .or. of(i)==k))) then
!~          do l=1,nl
!~            if (pro_mf(j,k) == dble(0)) then
!~              exit
!~            endif
!~            o1=ogen(i,l,1)
!~            o2=ogen(i,l,2)
!~            if (o1 > -1) then
!~              m1=pgen(j,l,1)
!~              m2=pgen(j,l,2)
!~              f1=pgen(k,l,1)
!~              f2=pgen(k,l,2)
!~              if (m1 > -1) then
!~                if (f1 > -1) then
!~                  pro_mf(j,k)=pro_mf(j,k) * pr_mf(o1,o2,m1,m2,f1,f2,nall(l),mtyp(l))
!~                else
!~                  pro_mf(j,k)=pro_mf(j,k) * pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
!~                endif
!~              else
!~                if (f1 > -1) then
!~                  pro_mf(j,k)=pro_mf(j,k) * pr_mb(o1,o2,f1,f2,nall(l),mtyp(l),l)
!~                else
!~                  pro_mf(j,k)=pro_mf(j,k) * pr_b(o1,o2,nall(l),mtyp(l),l)
!~                endif
!~              endif
!~            endif
!~          enddo
!~        else
!~          pro_mf(j,k)=dble(0)
!~        endif
           
!~        spro_mf=spro_mf + phi(j,k) * pro_mf(j,k) / sphi(j)
!~      enddo
     
!~      spro_mb=spro_mb + ksi(i,j) * theta(3) * pro_mb(j) / sksi(i)
!~      spro_m_=spro_m_ + ksi(i,j) * (theta(2) * pro_m(j) + theta(3) * pro_mb(j) + &
!~      (dble(1) - theta(2) - theta(3)) * spro_mf) / sksi(i)
!~    enddo
   
!~    pro=dble(odisp(i))*theta(1) * pro_b + (dble(1) - dble(odisp(i))*theta(1)) * spro_m_
!~    prb=dble(odisp(i))*theta(1) * pro_b / pro
!~    spro_mb=spro_mb * (dble(1) - dble(odisp(i))*theta(1)) / pro

!~    parass(i)%prms= prb
!~    parass(i)%prnms=spro_mb
   
!~  end subroutine migr_pr
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine creates a text file containing posterior probabilities that offspring results from seed
 ! and pollen immigration
!~  subroutine migr_probs
!~    integer :: i
   
!~    call reproductive_success
   
!~    !$omp parallel num_threads(usernth) default(shared)
!~    !$omp do
!~    do i=1,nndg
!~      call migr_pr(ondg(i))
!~    enddo
!~    !$omp end do
!~    !$omp end parallel
   
!~    !$omp parallel num_threads(usernth) default(shared)
!~    !$omp do
!~    do i=1,ndg
!~      call migr_pr(odg(i))
!~    enddo
!~    !$omp end do
!~    !$omp end parallel
   
!~    outfile = 189
!~    open (unit=outfile,file='mirg.out',action='write',status='replace')
!~    write (outfile,"(a1,(3a10))") ' ','Prog','ms','(1-ms)mp'
!~    do i=1,no
!~      write (outfile,*) parass(i)%oid,parass(i)%prms,parass(i)%prnms
!~      flush (outfile)
!~    enddo
!~    close(outfile)
   
!~  end subroutine migr_probs
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine  computes confidence limits for two parameters of dispersal kernel assuming 2D normal
 ! distrubution and the covariance structure between distance and shape taken from the Hessian
 subroutine N2D_confidence_bounds(i,j,dx,dy)
   integer, intent(in) :: i,j
   double precision, intent(inout) :: dx,dy
   double precision :: l1,l2,c11,c12,c22,x,y,chiv,th1,th2
   double precision, dimension(1:2) :: v1,v2
   double precision, dimension(1:4) :: xv,yv
   
   dx=dble(0)
   dy=dble(0)
   
   if (incl(i)==1 .and. incl(j)==1) then
   
     x=theta(i)
     y=theta(j)
     
     chiv=5.991464547108
     
     c11=-hess(i,i)
     c12=-hess(i,j)
     c22=-hess(j,j)
     
     l1=-(dsqrt(c11**2-dble(2)*c11*c22 + dble(4)*c12*c12 + c22**2) - c11 -c22)/dble(2)
     l2=(dsqrt(c11**2-dble(2)*c11*c22 + dble(4)*c12*c12 + c22**2) + c11 +c22)/dble(2)
     
     v1(1)=(dsqrt(c11**2 - dble(2)*c11*c22 + dble(4)*c12*c12 + c22**2) - c11 + c22)/(dble(2)*c12)
     v1(2)=dble(-1)
     
     v2(1)=-(dsqrt(c11**2 - dble(2)*c11*c22 + dble(4)*c12*c12 + c22**2) + c11 - c22)/(dble(2)*c12)
     v2(2)=dble(-1)
     
     th1=datan(v1(2)/v1(1))
     th2=datan(v2(2)/v2(1))
     
     xv(1)=dcos(th1)*dsqrt(chiv*l1)
     xv(2)=-xv(1)
     yv(1)=dsin(th1)*dsqrt(chiv*l1)
     yv(2)=-yv(1)
     
     xv(3)=dcos(th2)*dsqrt(chiv*l2)
     xv(4)=-xv(3)
     yv(3)=dsin(th2)*dsqrt(chiv*l2)
     yv(4)=-yv(3)
     
     dx=dmax1(xv(1),xv(2),xv(3),xv(4))
     dy=dmax1(yv(1),yv(2),yv(3),yv(4))
     
   endif
   
 end subroutine N2D_confidence_bounds
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine creates a text file containing approximate confidence bounds for dispersal kernel parameters
 subroutine db_confidence_bounds
   double precision :: dds,dbs,ddp,dbp,up
   integer :: outconf
   
   if (dist_typ==1) then
     if (incl(4)==1 .and. incl(6)==1) then
       call N2D_confidence_bounds(4,6,dds,dbs)
     else
       dds=dble(0)
       dbs=dble(0)
       if (incl(4)==1) dds=dsqrt(-hess(4,4))*1.959963985
       if (incl(6)==1) dbs=dsqrt(-hess(6,6))*1.959963985
     endif
     
     if (pollen_kernel==0) then
       if (incl(4)==1 .and. incl(6)==1) then
         call N2D_confidence_bounds(4,6,ddp,dbp)
       else
         ddp=dble(0)
         dbp=dble(0)
         if (incl(4)==1) ddp=dsqrt(-hess(4,4))*1.959963985
         if (incl(6)==1) dbp=dsqrt(-hess(6,6))*1.959963985
       endif
     else
       if (incl(5)==1 .and. incl(7)==1) then
         call N2D_confidence_bounds(5,7,ddp,dbp)
       else
         ddp=dble(0)
         dbp=dble(0)
         if (incl(5)==1) ddp=dsqrt(-hess(5,5))*1.959963985
         if (incl(7)==1) dbp=dsqrt(-hess(7,7))*1.959963985
       endif
     endif
     
     outconf=64
     open(unit=outconf,file='kernelconf.out',action='write',status='replace')
     write(outconf,"(a10,*(a15))") '  Param.  ','Estim.','Q[2.5%]','Q[97.5%]'
     up=dble(1)/(theta(4)-dds)
     if (up < 0) up=dble(1)/(theta(4)-theta(4))
     write(outconf,"(a10,*(f15.3))") ntheta(4),dble(1)/theta(4),dble(1)/(theta(4)+dds),up
     write(outconf,"(a10,*(f15.5))") ntheta(6),theta(6),theta(6)-dbs,theta(6)+dbs
     
     up=dble(1)/(theta(5)-ddp)
     if (up < 0) up=dble(1)/(theta(5)-theta(5))
     write(outconf,"(a10,*(f15.3))") ntheta(5),dble(1)/theta(5),dble(1)/(theta(5)+ddp),up
     write(outconf,"(a10,*(f15.5))") ntheta(7),theta(7),theta(7)-dbp,theta(7)+dbp
     
   endif
   
 end subroutine db_confidence_bounds
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a finite difference
 real (kind=8) function divdif2(x,y,fx,fy)
   real (kind=8), intent(in) :: x,y,fx,fy
   
   divdif2 = (fy-fx)/(y-x)
   
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a triple finite difference
 real (kind=8) function divdif3(x,y,z,fx,fy,fz)
   real (kind=8), intent(in) :: x,y,z,fx,fy,fz
   
   divdif3 = (divdif2(y,z,fy,fz)-divdif2(x,y,fx,fy))/(z-x)
   
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine performs a single step of searching along line for a given parameter
 subroutine powell_step(x1,x2,x3,f1,f2,f3,eps,minx,maxx,param,cont)
   real (kind=8), intent(inout) :: x1,x2,x3,f1,f2,f3
   real (kind=8), intent(in) :: eps,minx,maxx
   integer, intent(in) :: param
   integer, intent(inout) :: cont
   real (kind=8) :: x0,f0,dd3,d01,d02,d03,xt,ft
   
   dd3 = divdif3(x1,x2,x3,f1,f2,f3)
   
   !write (*,*) x1,x2,x3,dd3
   
   if (isnan(dd3)) then
     cont=-1
   else
   
     x0 = dble(0.5)*(dd3*(x1+x2)-divdif2(x1,x2,f1,f2))/dd3
     if (x0 < minx .or. x0 > maxx) then
       if (x0 < minx) x0=min(x1,x2,x3)-(min(x1,x2,x3)-minx) / 2.D0
       if (x0 > maxx) x0=max(x1,x2,x3)+(maxx-max(x1,x2,x3)) / 2.D0
     endif
     
     d01=abs(x0-x1)
     d02=abs(x0-x2)
     d03=abs(x0-x3)
     
     theta(param)=x0
     
     f0 = full_logl()
     if (f1<f2 .and. f1<f3) then
       x1=x0
       f1=f0
     else if (f2<f1 .and. f2<f3) then
       x2=x0
       f2=f0
     else
       x3=x0
       f3=f0
     endif
     
     if (x1>x2) then
       xt=x2
       ft=f2
       x2=x1
       f2=f1
       x1=xt
       f1=ft
     endif
     
     if (x1>x3) then
       xt=x3
       ft=f3
       x3=x1
       f3=f1
       x1=xt
       f1=ft
     endif
     
     if (x2>x3) then
       xt=x3
       ft=f3
       x3=x2
       f3=f2
       x2=xt
       f2=ft
     endif
     
     
     if (min(d01,d02,d03)<eps) then
       cont=0
     else
       cont=1
     endif
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine performs a search along line for a given parameter
 subroutine powell_optim(t1,t2,t3,eps,maxl,param)
   real (kind=8), intent(inout) :: t1,t2,t3,maxl
   real (kind=8) :: eps
   real (kind=8) :: l1,l2,l3,mint,maxt,x,y
   integer :: param,cont
   
   theta(param)=t1
   l1=full_logl()
   
   theta(param)=t2
   l2=full_logl()
   
   theta(param)=t3
   l3=full_logl()
   
   if (l1>l2) then
     y=l1
     x=t1
     t1=t2
     l1=l2
     t1=x
     l1=y
   endif
   
   if (l1>l3) then
     y=l1
     x=t1
     t1=t3
     l1=l3
     t3=x
     l3=y
   endif
   
   if (l2>l3) then
     y=l2
     x=t2
     t2=t3
     l2=l3
     t3=x
     l3=y
   endif
   
   if (param<=7) then
     mint=minth(param)
     maxt=maxth(param)
   else
     mint=-10.D0
     maxt=10.D0
   endif
   
   do
     call powell_step(t1,t2,t3,l1,l2,l3,eps,mint,maxt,param,cont)
     if (cont==0) then
       exit
     elseif (cont==-1) then
       write (*,"('Abnormal termination due to divergence. Value of <',a)",advance='no') trim(adjustl(ntheta(param)))
       write (*,"('> that causes /divergence/ error: ')",advance='no')
       write (*,*) theta(param)
       stop
     endif
   enddo
   
   maxl=max(l1,l2,l3)
   if (l1==maxl) then
     theta(param)=t1
   else if (l2==maxl) then
     theta(param)=t2
   else
     theta(param)=t3
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine attepts to find model parameter values which maximize the likelihood
 ! Parameters are optimized sequentially in a random order
 ! (may be helpful if parameter values are hardly known)
 subroutine powell_theta(nit)
   integer :: i,j,k,l,nit
   real (kind=8) :: maxl,omaxl
   real (kind=8), allocatable :: th(:,:),oth(:)
   integer, allocatable :: nth(:)
   
   write(*,"(a)") 'Initial searching for parameter values:'
   allocate(th(11+2*nf,3))
   allocate(nth(11+2*nf))
   allocate(oth(11+2*nf))
   
   do i=1,11+2*nf
     th(i,1)=theta(i)-theta(i)*0.1D0
     th(i,2)=theta(i)
     th(i,3)=theta(i)+theta(i)*0.1D0
     nth(i)=i
     oth(i)=theta(i)
   enddo
   
   omaxl=0.D0
   do i=1,nit
     
     !shuffle order of updates
     do j=11+2*nf,1,-1
       k=nth(j)
       l=int(rnd()*j)+1
       nth(j)=nth(l)
       nth(l)=k
     enddo
     !end of shuffling
     
     write(*,"(a)",advance='no') 'Optimizing parameters (search):'
     do j=1,11+2*nf
       if (incl(nth(j))==1) call powell_optim(th(nth(j),1),th(nth(j),2),th(nth(j),3),dble(0.01),maxl,nth(j))
       if (j<11+2*nf) write (*,"(i2,a2)",advance='no') j*100/(11+2*nf),'% '
     enddo
     write (*,"(a)") '100%'
     
     if (i>1 .and. maxl<omaxl) exit
     
     if (dist_typ==1) then
       write(*,"(i5,f15.2,*(f10.5))") i,maxl,(theta(k),k=1,11),(theta(11+k),theta(11+k+nf),k=1,nf)
     else
       write(*,"(i5,f15.2,3(f10.5),2(f10.2),*(f10.5))") i,maxl,(theta(k),k=1,11), &
        (theta(11+k),theta(11+k+nf),k=1,nf)
     endif
     
     if (i>1 .and. dabs(maxl-omaxl)<0.001D0) exit
     omaxl=maxl
     
     if (incl_mtyp==1) call update_mtyp
     if (incl_mcytyp==1) call update_mcytyp
     if (incl_maternal_cy==1) call update_maternal_cy
     
     do j=1,11+2*nf
       if (oth(j)==theta(j)) then
         th(j,1)=th(j,1)-theta(j)*0.1D0
         th(j,3)=th(j,3)+theta(j)*0.1D0
       endif
       oth(j)=theta(j)
     enddo
     
   enddo
   
   call write_out(maxl,'p')
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------

 !----------------------------------------------------------------------------------------------------------
 ! This routine writes model parameter values (theta) to the text file
 ! (may be helpful for a series of automated analyses)
 subroutine write_out(maxl,ch)
   integer :: k,f
   character (len=1) :: ch
   real (kind=8) :: maxl
   logical :: exist
   
    f=99
   inquire(file="out_"//ch//".txt", exist=exist)
   if (exist) then
     open(f, file="out_"//ch//".txt", status="old", position="append", action="write")
   else
     open(f, file="out_"//ch//".txt", status="new", action="write")
   endif
   
   write(f,*) trim(fo),maxl,(theta(k),k=1,11), &
      (theta(11+k),theta(11+k+nf),k=1,nf)
   close(f)
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws a random genealogy for the i-th offspring
 subroutine genealogy(i,ujsr)
   integer, intent(in) :: i
   integer*8, intent(inout) :: ujsr
   integer :: j,k,l,o1,o2,m1,m2,f1,f2
   double precision :: pro_b,spro_mf,spro_m_,doup,pro,dom,dof,r,cumul_pro
   double precision, allocatable :: pro_m(:),pro_mb(:),pro_mf(:,:)
   double precision :: protmp
   integer, dimension(2,2) :: mf
   logical :: done
   
   done=.false.
   
   spro_m_=dble(0)
   
   doup = dble(1)
   if (ocy(i).ne.-1 .and. nhap>1) then
     doup = (dble(1)-mcytyp)*pcm(ocy(i))+mcytyp/dble(nhap-1)*(dble(1)-pcm(ocy(i)))
   endif
   
   pro_b=doup              !Pr(Oi|B)
   allocate(pro_m(np))     !Pr(Oi|Mj,Mj)
   allocate(pro_mb(np))    !Pr(Oi|Mj,B)
   allocate(pro_mf(np,np)) !Pr(Oi|Mj,Fk)
   
   pro_m=0.D0
   pro_mb=0.D0
   pro_mb=0.D0
   
   !spro_m_ 
   
   if (om(i)<0) then  
     do l=1,nl
       o1=ogen(i,l,1)
       o2=ogen(i,l,2)
       if (o1 > -1) then
         pro_b=pro_b * pr_b(o1,o2,nall(l),mtyp(l),l)
       endif
     enddo
   else
     pro_b=dble(0)
   endif
   
   spro_m_=dble(0)
   do j=1,np
     if (ksi(i,j)==dble(0)) cycle
     
     if (onb(i,j)==1 .or. om(i)==j) then
       
       dom=dble(1)
       if (ocy(i).ne.-1) then
         if (pcy(j)==-1) then
           dom = doup
         else
           dom = delta_eps(pcy(j),ocy(i),nhap,mcytyp)
         endif
       endif
       
       pro_mb(j)=dble(dom) * maternal_cy + doup * (dble(1) - maternal_cy)
       pro_m(j)=dble(dom)
       
       if ((om(i)<0) .or. (om(i)==j)) then
         do l=1,nl
           if (pro_mb(j) == dble(0)) then
             exit
           endif
           
           o1=ogen(i,l,1)
           o2=ogen(i,l,2)
           if (o1 > -1) then
             m1=pgen(j,l,1)
             m2=pgen(j,l,2)
             if (m1 > -1) then
               pro_m(j)=pro_m(j) * pr_mm(o1,o2,m1,m2,nall(l),mtyp(l))
               pro_mb(j)=pro_mb(j) * pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
             else
               pro_m(j)=pro_m(j) * pr_b(o1,o2,nall(l),mtyp(l),l)
               pro_mb(j)=pro_mb(j) * pr_b(o1,o2,nall(l),mtyp(l),l)
             endif
           endif
           
         enddo
       else
         pro_m(j)=dble(0)
         pro_mb(j)=dble(0)
       endif
       
       if ((om(i)>=0) .and. (of(i)>=0) .and. (om(i).ne.of(i))) pro_m(j)=dble(0)
       if ((om(i)>=0) .and. (of(i)>=0)) pro_mb(j)=dble(0)
       
       if (sphi(j)==dble(0)) cycle
       
       spro_mf=dble(0)
       do k=1,np
         if (phi(j,k)==0.D0) cycle
         
         if (pnb(j,k)==1) then
           dof = dble(1)
           if (ocy(i).ne.-1) then
             if (pcy(k)==-1) then
               dof = doup
             else
               dof = delta_eps(pcy(k),ocy(i),nhap,mcytyp)
             endif
           endif
         
           pro_mf(j,k)=dble(dom) * maternal_cy + dble(dof) * (dble(1) - maternal_cy)
           
           if (k==j) pro_mf(j,k)=dble(0)
           
           if ((om(i)<0 .and. of(i)<0) .or. (om(i)==j .and. (of(i)<0 .or. of(i)==k))) then
             do l=1,nl
               if (pro_mf(j,k) == dble(0)) then
                 exit
               endif
               o1=ogen(i,l,1)
               o2=ogen(i,l,2)
               if (o1 > -1) then
                 m1=pgen(j,l,1)
                 m2=pgen(j,l,2)
                 f1=pgen(k,l,1)
                 f2=pgen(k,l,2)
                 if (m1 > -1) then
                   if (f1 > -1) then
                     pro_mf(j,k)=pro_mf(j,k) * pr_mf(o1,o2,m1,m2,f1,f2,nall(l),mtyp(l))
                   else
                     pro_mf(j,k)=pro_mf(j,k) * pr_mb(o1,o2,m1,m2,nall(l),mtyp(l),l)
                   endif
                 else
                   if (f1 > -1) then
                     pro_mf(j,k)=pro_mf(j,k) * pr_mb(o1,o2,f1,f2,nall(l),mtyp(l),l)
                   else
                     pro_mf(j,k)=pro_mf(j,k) * pr_b(o1,o2,nall(l),mtyp(l),l)
                   endif
                 endif
               endif
             enddo
           else
             pro_mf(j,k)=dble(0)
           endif
             
           spro_mf=spro_mf + phi(j,k) * pro_mf(j,k) / sphi(j)
         endif
       enddo
       !sum of all Pr(Oi|M,*)
       spro_m_=spro_m_ + ksi(i,j) * (theta(2) * pro_m(j) + theta(3) * pro_mb(j) + &
       (dble(1) - theta(2) - theta(3)) * spro_mf) / sksi(i)
     endif
   enddo
   
   ! identify two the most likely genealogies
   
   !total probability of the offspring
   pro=dble(odisp(i))*theta(1) * pro_b + (dble(1) - dble(odisp(i))*theta(1)) * spro_m_
   
   !probability that th offspring is due to seed immigration
   cumul_pro=dble(odisp(i))*theta(1) * pro_b / pro
   
   mf(1,1)=-1
   mf(1,2)=-1
   
   r=prnd(ujsr)
   if (r<cumul_pro) then
     !offspring is due to seed immigration
     mf(1,1)=-1
     mf(1,2)=-1
     done=.true.
   else
     !offsprig is due to local seed dispersal (i.e. mother is within local population)
     mf(1,1)=-1
     mf(1,2)=-1
     
     do j=1,np
       if (done) cycle
       if (onb(i,j)==1 .or. om(i)==j) then
         !selfing with j-th parent
         protmp=(dble(1) - dble(odisp(i))*theta(1)) * ksi(i,j)/sksi(i) * theta(2) * pro_m(j) / pro
         cumul_pro=cumul_pro+protmp
         
         if (r<cumul_pro) then
           mf(1,1)=j
           mf(1,2)=j
           done=.true.
           exit
         endif
         
         !pollen immigrant with j-th parent
         protmp=(dble(1) - dble(odisp(i))*theta(1)) * ksi(i,j)/sksi(i) * theta(3) * pro_mb(j) / pro
         cumul_pro=cumul_pro+protmp
         
         if (r<cumul_pro) then
           mf(1,1)=j
           mf(1,2)=-1
           done=.true.
           exit
         endif
         
         !local outcrossing with j-th and k-th parent
         do k=1,np
           if (done) cycle
           if (pnb(j,k)==1) then
             if (pro_mf(j,k) == dble(0)) cycle
             if (k==j) cycle
             
             protmp=(dble(1) - dble(odisp(i))*theta(1)) * ksi(i,j)/sksi(i) * (dble(1) - theta(2) - theta(3)) &
              * phi(j,k) * pro_mf(j,k)/sphi(j) / pro
             
             cumul_pro=cumul_pro+protmp
             
             if (r<cumul_pro) then
               mf(1,1)=j
               mf(1,2)=k
               done=.true.
               exit
             endif
           endif
         enddo
       endif
     enddo
   
   endif
   
   parass(i)%oid=oid(i)
   parass(i)%mo1=mf(1,1)
   parass(i)%fa1=mf(1,2)
   
 end subroutine genealogy
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine draws a random genealogy for the i-th offspring based on constant transition probs
 subroutine c_genealogy(i,ujsr)
   integer, intent(in) :: i
   integer*8, intent(inout) :: ujsr
   integer :: j,k
   double precision :: spro_mf,spro_m_,r,tot_pro,cumul_pro
   double precision :: protmp
   integer, dimension(2,2) :: mf
   logical :: done
   
   done=.false.
   
   spro_m_=dble(0)
   do j=1,np
     if (ksi(i,j)==dble(0) .or. sphi(j)==dble(0)) cycle
     spro_mf=dble(0)
     do k=1,np
       if (phi(j,k)==0.D0) cycle
       spro_mf=spro_mf + phi(j,k) * c_pro_mf(i,j,k) / sphi(j)
     enddo
     !sum of all Pr(Oi|M,*)
     spro_m_=spro_m_ + ksi(i,j) * (theta(2) * c_pro_m(i,j) + theta(3) * c_pro_mb(i,j) + &
       (dble(1) - theta(2) - theta(3)) * spro_mf) / sksi(i)
   enddo
   
   ! draw at random a genealogy:
   
   !total probability of the offspring
   tot_pro=dble(odisp(i))*theta(1) * c_pro_b(i) + (dble(1) - dble(odisp(i))*theta(1)) * spro_m_
   
   !cumulative probability is initialized at the prob thah the offspring is due to seed immigration
   cumul_pro=dble(odisp(i))*theta(1) * c_pro_b(i) / tot_pro
   
   mf(1,1)=-1
   mf(1,2)=-1
   
   r=prnd(ujsr)
   if (r<cumul_pro) then
     !offspring is due to seed immigration
     mf(1,1)=-1
     mf(1,2)=-1
     done=.true.
   else
     !offsprig is due to local seed dispersal (i.e. mother is within local population)
     mf(1,1)=-1
     mf(1,2)=-1
     
     do j=1,np
       if (done) cycle
       if (onb(i,j)==1 .or. om(i)==j) then
         !selfing with j-th parent
         protmp=(dble(1) - dble(odisp(i))*theta(1)) * ksi(i,j)/sksi(i) * theta(2) * c_pro_m(i,j) / tot_pro
         cumul_pro=cumul_pro+protmp
         
         if (r<cumul_pro) then
           mf(1,1)=j
           mf(1,2)=j
           done=.true.
           exit
         endif
         
         !pollen immigrant with j-th parent
         protmp=(dble(1) - dble(odisp(i))*theta(1)) * ksi(i,j)/sksi(i) * theta(3) * c_pro_mb(i,j) / tot_pro
         cumul_pro=cumul_pro+protmp
         
         if (r<cumul_pro) then
           mf(1,1)=j
           mf(1,2)=-1
           done=.true.
           exit
         endif
         
         !local outcrossing with j-th and k-th parent
         do k=1,np
           if (done) cycle
           if (pnb(j,k)==1) then
             if (c_pro_mf(i,j,k) == dble(0)) cycle
             if (k==j) cycle
             
             protmp=(dble(1) - dble(odisp(i))*theta(1)) * ksi(i,j)/sksi(i) * (dble(1) - theta(2) - theta(3)) &
              * phi(j,k) * c_pro_mf(i,j,k)/sphi(j) / tot_pro
             
             cumul_pro=cumul_pro+protmp
             
             if (r<cumul_pro) then
               mf(1,1)=j
               mf(1,2)=k
               done=.true.
               exit
             endif
           endif
         enddo
       endif
     enddo
   
   endif
   
   parass(i)%oid=oid(i)
   parass(i)%mo1=mf(1,1)
   parass(i)%fa1=mf(1,2)
   
 end subroutine c_genealogy
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine first draws random genealogies given the current values of model parameters and then
 ! computes summary statistics (auxiliary variables) that are used in the Metropolis-Hastings algorithm
 ! (see "Bayes_update_parameters")
 subroutine genealogies
   integer :: i,th
   integer*8 :: tjsr
   
   call reproductive_success
   if (.not. (allocated(parass))) allocate(parass(no))
   if (.not. (allocated(noij))) allocate(noij(no,np))
   if (.not. (allocated(npjk))) allocate(npjk(np,np))
   nfm=0
   nlm=0
   nls=0
   nlo=0
   noij=0
   npjk=0
   
   if (do_const_probs) then
     !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
     !$omp do
     do i=1,nndg
       th=omp_get_thread_num()+1
       tjsr=pjsr(th)
       call c_genealogy(ondg(i),tjsr)
       pjsr(th)=tjsr
     enddo
     !$omp end do
     !$omp end parallel
     
     !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
     !$omp do
     do i=1,ndg
       th=omp_get_thread_num()+1
       tjsr=pjsr(th)
       call c_genealogy(odg(i),tjsr)
       pjsr(th)=tjsr
     enddo
     !$omp end do
     !$omp end parallel
   else
     !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
     !$omp do
     do i=1,nndg
       th=omp_get_thread_num()+1
       tjsr=pjsr(th)
       call genealogy(ondg(i),tjsr)
       pjsr(th)=tjsr
     enddo
     !$omp end do
     !$omp end parallel
     
     !$omp parallel num_threads(usernth) default(shared) private(th,tjsr)
     !$omp do
     do i=1,ndg
       th=omp_get_thread_num()+1
       tjsr=pjsr(th)
       call genealogy(odg(i),tjsr)
       pjsr(th)=tjsr
     enddo
     !$omp end do
     !$omp end parallel
   endif
   
   do i=1,no
     if (parass(i)%mo1==-1) then
       nfm=nfm+1
     else
       noij(i,parass(i)%mo1)=noij(i,parass(i)%mo1)+1
       if (parass(i)%fa1==-1) then
         nlm=nlm+1
       else
         if (parass(i)%mo1==parass(i)%fa1) then
           nls=nls+1
         else if (parass(i)%mo1/=parass(i)%fa1) then
           nlo=nlo+1
           npjk(parass(i)%mo1,parass(i)%fa1)=npjk(parass(i)%mo1,parass(i)%fa1)+1
         endif 
       endif
     endif
   enddo
   
 end subroutine genealogies
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine implements a shift-register generator (64-bit version)
 function shift_xor(val,shift)
   integer*8 :: shift_xor
   integer*8, intent(in) :: val, shift
   
   shift_xor = ieor(val,ishft(val,shift))
 end function
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine initializes serial random number generator
 subroutine init_rnd
    
    jsr=shift_xor(shift_xor(shift_xor(jsr, 13_8),-17_8),43_8)
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the uniform (0,1) pdf
 ! (Serial versio)
 double precision function rnd()
   integer*8 :: ojsr
   ojsr=jsr
   jsr = shift_xor(shift_xor(shift_xor(jsr, 13_8),-17_8),43_8)
   rnd = dble(ojsr+jsr)*5.42101086242752217D-20+0.5D0
 end function rnd
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine initializes parallel random number generator
 subroutine init_prnd
   integer*8 :: i
   
   allocate(pjsr(usernth))
   do i=1,usernth
     pjsr(i)=int(dble(300000)*rnd())+1
     pjsr(i)=shift_xor(shift_xor(shift_xor(pjsr(i), 13_8),-17_8),43_8)
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the uniform (0,1) pdf
 ! (Parallel version)
 double precision function prnd(ujsr)
   integer*8, intent(inout) :: ujsr
   integer*8 :: ojsr
   ojsr=ujsr
   ujsr = shift_xor(shift_xor(shift_xor(ujsr, 13_8),-17_8),43_8)
   Prnd = dble(ojsr+ujsr)*5.42101086242752217D-20+0.5D0
 end function prnd
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the standard gamma pdf
 double precision function rgamma(a)
   double precision, intent(in) :: a
   double precision :: x,v,u,c,d
   
   d=a-dble(1)/dble(3)
   c=dble(1)/dsqrt(dble(9)*d)
   
   do
     do
       x=dsqrt(-dble(2)*dlog(rnd()))*dcos(my2pi*rnd())
       v=(dble(1)+c*x)
       if (v > dble(0)) exit
     enddo
     
     v=v**3
     u=rnd()
     
     if (u<(dble(1)-dble(0.0331)*x**4)) exit
     if (dlog(u)<(dble(0.5)*x**2+d*(dble(1)-v+dlog(v)))) exit
   enddo
   
   rgamma=d*v
 end function rgamma
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the standard gamma pdf
 double precision function gammadev(a)
   double precision, intent(in) :: a
  
   if (a<dble(1)) then
     gammadev=rgamma(a+dble(1))*rnd()**(dble(1)/a)
   else
     gammadev=rgamma(a)
   endif
  
 end function gammadev
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the normal pdf (mean m and std.dev. s)
 ! (Serial version)
 double precision function normaldev(m,s)
   double precision, intent(in) :: m,s
   double precision :: v1,v2,r
   
   do
     v1 = dble(2) * rnd() - dble(1)
     v2 = dble(2) * rnd() - dble(1)
     r  = v1**2 + v2**2
     if (r < dble(1) .and. r > dble(0)) exit
   enddo
   normaldev = v2 * dsqrt(-dble(2) * dlog(r) / r) * s + m
   
 end function normaldev
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a random draw from the normal pdf (mean m and std.dev. s)
 ! (Parallel version)
 double precision function pnormaldev(m,s,tjsr)
   integer*8, intent(inout) :: tjsr
   double precision, intent(in) :: m,s
   double precision :: v1,v2,r
   
   do
     v1 = dble(2) * prnd(tjsr) - dble(1)
     v2 = dble(2) * prnd(tjsr) - dble(1)
     r  = v1**2 + v2**2
     if (r < dble(1) .and. r > dble(0)) exit
   enddo
   pnormaldev = v2 * dsqrt(-dble(2) * dlog(r) / r) * s + m
   
 end function pnormaldev
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a new (0,1)-constrained value using the rectangular distr. (reflection assures a positive)
 ! (Serial version)
 double precision function rectdev(x,d)
   !returns a new value
   !using the rectangular distr. reflection
   double precision, intent(in) :: x,d
   double precision :: nx
   
   nx = x + dble(dble(0.5) - rnd()) * d
   if (nx < dble(0)) nx = -nx
   if (nx > dble(1)) nx = dble(int(nx)+1) - nx
   rectdev = nx
   
 end function rectdev
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns a new (0,1)-constrained value using the rectangular distr. (reflection assures a positive)
 ! (Parallel version)
 double precision function prectdev(x,d,ujsr)
   double precision, intent(in) :: x,d
   integer*8, intent(inout) :: ujsr
   double precision :: nx
   
   nx = x + dble(dble(0.5) - prnd(ujsr)) * d
   if (nx < dble(0)) nx = -nx
   if (nx > dble(1)) nx = dble(int(nx)+1) - nx
   
   prectdev = nx
 end function prectdev
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine re-computes female relative fitness after the change of the j-th indivdual's female fitness
 subroutine recomp_ksi(j,fjo,fjn)
   integer :: i,j
   double precision :: fjo,fjn
   
   do i=1,no
     if (odisp(i)==0) cycle
     sksi(i)=sksi(i)-ksi(i,j)
     ksi(i,j)=ksi(i,j)/fjo*fjn
     sksi(i)=sksi(i)+ksi(i,j)
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine re-computes male relative fitness after the change of the j-th indivdual's male fitness
 subroutine recomp_phi(j,fjo,fjn)
   integer :: i,j
   double precision :: fjo,fjn
   
   do i=1,np
     if (i==j) cycle
     sphi(i)=sphi(i)-phi(i,j)
     phi(i,j)=phi(i,j)/fjo*fjn
     sphi(i)=sphi(i)+phi(i,j)
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine re-computes the expected log-female fitness after the change of the k-th variable's slope
 subroutine compute_mlogfemfec(k,obk,nbk)
   integer :: i,k
   double precision :: obk,nbk
   
   do i=1,np
     if (pfem(i)>0.D0) mlogfemfec(i)=mlogfemfec(i)+(nbk-obk)*pzfen(1,i,k)
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine re-computes the expected log-male fitness after the change of the k-th variable's slope
 subroutine compute_mlogmalfec(k,obk,nbk)
   integer :: i,k
   double precision :: obk,nbk
   
   do i=1,np
     if (pfem(i)<1.D0) mlogmalfec(i)=mlogmalfec(i)+(nbk-obk)*pzfen(2,i,k)
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes the Bayesian R^2 as a proportion of variance explained by the regression model
 subroutine bayes_r2
   integer :: i,nfem,nmal,mod_f,mod_m
   double precision :: malmpred,malmres,malvpred,malvres
   double precision :: femmpred,femmres,femvpred,femvres
   
   call ytoint(mod_f,mod_m)
   
   femmpred=0.D0
   femmres=0.D0
   femvpred=0.D0
   femvres=0.D0
   malmpred=0.D0
   malmres=0.D0
   malvpred=0.D0
   malvres=0.D0
   nfem=0
   nmal=0
   do i=1,np
     if (pfem(i)>0.D0) then
       femmpred=femmpred+mlogfemfec(i)
       femvpred=femvpred+mlogfemfec(i)**2
       femmres=femmres+dlog(femfec(i))-mlogfemfec(i)
       femvres=femvres+(dlog(femfec(i))-mlogfemfec(i))**2
       nfem=nfem+1
     endif
     if (pfem(i)<1.D0) then
       malmpred=malmpred+mlogmalfec(i)
       malvpred=malvpred+mlogmalfec(i)**2
       malmres=malmres+dlog(malfec(i))-mlogmalfec(i)
       malvres=malvres+(dlog(malfec(i))-mlogmalfec(i))**2
       nmal=nmal+1
     endif
   enddo
   
   if (nfem>0 .and. mod_f>0) then
     femvpred=(femvpred-femmpred**2/dble(nfem))
     femvres=(femvres-femmres**2/dble(nfem))
     femr2=femvpred/(femvpred+femvres)
   else
     femr2=0.D0
   endif
   
   if (nmal>0 .and. mod_m>0) then
     malvpred=(malvpred-malmpred**2/dble(nmal))
     malvres=(malvres-malmres**2/dble(nmal))
     malr2=malvpred/(malvpred+malvres)
   else
     malr2=0.D0
   endif
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine returns the natural logarithm of the Dirichlet pdf
 double precision function log_DirichletD(x,b)
   integer :: n
   double precision, intent(in) :: x(:),b(:)
   integer :: i
   double precision :: pg,sb,p
   
   n=size(x)
   pg=dble(0)
   sb=dble(0)
   p=dble(0)
   do i=1,n
     sb=sb+b(i)
     pg=pg+DLGAMA(b(i))
     p=p+(b(i)-dble(1))*dlog(x(i))
   enddo
   log_DirichletD=p-pg+DLGAMA(sb)
   
 end function log_DirichletD
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine implements a single iteration of MCMC run. Parameters are updated using either the Gibbs
 ! sampler or the Metropolis-Hastings sampler.
 subroutine Bayes_update_parmeters
   integer :: i,j,k,th,nth
   integer*8 :: tjsr
   double precision :: a,b,c,l_o,l_n,p_o,p_n,x_o,x_n
   double precision :: bo(2),bn(2),xx(2),bb(2)
   
   !seed immigration drawn from the beta distribution with parameters a=# seed immigrants+1 b=# all dispersed progeny+1
   if (incl(1)==1) then
     a=rgamma(nfm+1.D0)
     b=rgamma(ndg-nfm+1.D0)
     theta(1)=a/(a+b)
   endif
   
   !selfing and pollen immigration drawn from the Dirichlet distribution
   if (incl(2)==1 .and. incl(3)==1) then
     !both
     a=gammadev(nls+1.D0)
     b=gammadev(nlm+1.D0)
     c=gammadev(nlo+1.D0)
     theta(2)=a/(a+b+c)
     theta(3)=b/(a+b+c)
   else if (incl(2)==1 .and. incl(3)==0) then
     !selfing only
     a=gammadev(nls+1.D0)
     c=gammadev(nlo+nlm+1.D0)
     theta(2)=a/(a+c)
   else if (incl(2)==0 .and. incl(3)==1) then
     !pollen immigration only
     b=gammadev(nlm+1.D0)
     c=gammadev(nlo+nls+1.D0)
     theta(3)=b/(b+c)
   endif
   
   !seed dispersal kenel
   if (incl(4)==1 .or. incl(6)==1 .or. incl(8)==1 .or. incl(10)==1) then
     l_o=0.D0
     do i=1,no
       do j=1,np
       if (noij(i,j)>0) l_o=l_o+noij(i,j)*dlog(ksi(i,j)/sksi(i))
       enddo
     enddo
     
     !dispersal distance
     if (incl(4)==1) then
       x_o=theta(4)
       theta(4)=x_o/dexp(normaldev(0.D0,v_ds))
       if (theta(4)<minth(4) .or. theta(4)>maxth(4)) then
         theta(4)=x_o
         rr_ds=rr_ds+1.D0
       else
         call reproductive_success_ksi
         l_n=0.D0
         do i=1,no
           do j=1,np
           if (noij(i,j)>0) l_n=l_n+noij(i,j)*dlog(ksi(i,j)/sksi(i))
           enddo
         enddo
         
         !if (dlog(rnd())>=l_n-l_o+dlog(1.D0/theta(4))-dlog(1.D0/x_o)) then !uniform prior
         if (dlog(rnd())>=l_n-l_o) then !reciprocal prior (generally works better!)
           rr_ds=rr_ds+1.D0
           theta(4)=x_o
           call reproductive_success_ksi
         else
           l_o=l_n
           ar_ds=ar_ds+1.D0
         endif
       endif
     endif
     
     !kernel shape
     if (incl(6)==1) then
       x_o=theta(6)
       theta(6)=x_o*dexp(normaldev(0.D0,v_bs))
       if (theta(6)<minth(6) .or. theta(6)>maxth(6)) then
         theta(6)=x_o
         rr_bs=rr_bs+1.D0
       else
         call reproductive_success_ksi
         l_n=0.D0
         do i=1,no
           do j=1,np
           if (noij(i,j)>0) l_n=l_n+noij(i,j)*dlog(ksi(i,j)/sksi(i))
           enddo
         enddo
         !if (dlog(rnd())>=l_n-l_o+dlog(theta(6))-dlog(x_o)) then !uniform prior
         if (dlog(rnd())>=l_n-l_o) then !reciprocal prior (generally works better!)
           rr_bs=rr_bs+1.D0
           theta(6)=x_o
           call reproductive_success_ksi
         else
           l_o=l_n
           ar_bs=ar_bs+1.D0
         endif
       endif
     endif
     
     !intensity of ansitropy
     if (incl(8)==1) then
       x_o=theta(8)
       theta(8)=x_o+normaldev(0.D0,v_ks)
       if (theta(8)<0.D0) theta(8)=-theta(8) !reflection to keep kappa positive
       call reproductive_success_ksi
       p_o=-(x_o**2)/(2*10.D0)
       p_n=-(theta(9)**2)/(2*10.D0)
       l_n=0.D0
       do i=1,no
         do j=1,np
         if (noij(i,j)>0) l_n=l_n+noij(i,j)*dlog(ksi(i,j)/sksi(i))
         enddo
       enddo
       if (dlog(rnd())>=l_n-l_o+p_n-p_o) then
         rr_ks=rr_ks+1.D0
         theta(8)=x_o
         call reproductive_success_ksi
       else
         l_o=l_n
         ar_ks=ar_ks+1.D0
       endif
     endif
     
     !prevailing direction (% pi)
     if (incl(10)==1) then
       x_o=theta(10)
       theta(10)=x_o*my2pi+normaldev(0.D0,v_as)
       theta(10)=atan2(sin(theta(10)),cos(theta(10)))
       if (theta(10)<0.D0) theta(10)=my2pi+theta(10) !kind of reflrection to keep angle within (0,2pi)
       theta(10)=theta(10)/my2pi
       
       call reproductive_success_ksi
       l_n=0.D0
       do i=1,no
         do j=1,np
         if (noij(i,j)>0) l_n=l_n+noij(i,j)*dlog(ksi(i,j)/sksi(i))
         enddo
       enddo
       if (dlog(rnd())>=l_n-l_o) then
         theta(10)=x_o
         call reproductive_success_ksi
         rr_as=rr_as+1.D0
       else
         ar_as=ar_as+1.D0
       endif
     endif
   endif
   
   !pollen dispersal kernel
   if (incl(5)==1 .or. incl(7)==1 .or. incl(9)==1 .or. incl(11)==1) then
     l_o=0.D0
     do i=1,np
       do j=1,np
       if (npjk(i,j)>0) l_o=l_o+npjk(i,j)*dlog(phi(i,j)/sphi(i))
       enddo
     enddo
     
     !dispersal distance
     if (incl(5)==1) then
       x_o=theta(5)
       theta(5)=x_o/dexp(normaldev(0.D0,v_dp))
       if (theta(5)<minth(5) .or. theta(5)>maxth(5)) then
         theta(5)=x_o
         rr_dp=rr_dp+1.D0
       else
         call reproductive_success_phi
         l_n=0.D0
         do i=1,np
           do j=1,np
             if (npjk(i,j)>0) l_n=l_n+npjk(i,j)*dlog(phi(i,j)/sphi(i))
           enddo
         enddo
         !if (dlog(rnd())>=l_n-l_o+dlog(1.D0/theta(5))-dlog(1.D0/x_o)) then !uniform prior
         if (dlog(rnd())>=l_n-l_o) then
           theta(5)=x_o
           call reproductive_success_phi
           rr_dp=rr_dp+1.D0
         else
           l_o=l_n
           ar_dp=ar_dp+1.D0
         endif
       endif
     endif
     
     !kernel shape
     if (incl(7)==1) then
       x_o=theta(7)
       theta(7)=x_o*dexp(normaldev(0.D0,v_bp))
       if (theta(7)<minth(7) .or. theta(7)>maxth(7)) then
         theta(7)=x_o
         rr_bp=rr_bp+1.D0
       else
         call reproductive_success_phi
         l_n=0.D0
         do i=1,np
           do j=1,np
             if (npjk(i,j)>0) l_n=l_n+npjk(i,j)*dlog(phi(i,j)/sphi(i))
           enddo
         enddo
         !if (dlog(rnd())>=l_n-l_o+dlog(theta(7))-dlog(x_o)) then !uniform prior
         if (dlog(rnd())>=l_n-l_o) then !reciprocal prior
           theta(7)=x_o
           call reproductive_success_phi
           rr_bp=rr_bp+1.D0
         else
           l_o=l_n
           ar_bp=ar_bp+1.D0
         endif
       endif
     endif
     
     !intensity of anisotropy
     if (incl(9)==1) then
       x_o=theta(9)
       theta(9)=x_o+normaldev(0.D0,v_kp)
       if (theta(9)<0.D0) theta(9)=-theta(9) !reflection to keep kappa positive
       call reproductive_success_phi
       p_o=-(x_o**2)/(2*10.D0)
       p_n=-(theta(9)**2)/(2*10.D0)
       l_n=0.D0
       do i=1,np
         do j=1,np
           if (npjk(i,j)>0) l_n=l_n+npjk(i,j)*dlog(phi(i,j)/sphi(i))
         enddo
       enddo
       if (dlog(rnd())>=l_n-l_o+p_n-p_o) then
         theta(9)=x_o
         call reproductive_success_phi
         rr_kp=rr_kp+1.D0
       else
         l_o=l_n
         ar_kp=ar_kp+1.D0
       endif
     endif
     
     !prevailing direction
     if (incl(11)==1) then
       x_o=theta(11)
       theta(11)=x_o*my2pi+normaldev(0.D0,v_ap)
       theta(11)=atan2(sin(theta(11)),cos(theta(11)))
       if (theta(11)<0.D0) theta(11)=my2pi+theta(11) !kind of reflrection to keep angle within (0,2pi)
       theta(11)=theta(11)/my2pi
       
       call reproductive_success_phi
       l_n=0.D0
       do i=1,np
         do j=1,np
           if (npjk(i,j)>0) l_n=l_n+npjk(i,j)*dlog(phi(i,j)/sphi(i))
         enddo
       enddo
       if (dlog(rnd())>=l_n-l_o) then
         theta(11)=x_o
         call reproductive_success_phi
         rr_ap=rr_ap+1.D0
       else
         l_o=l_n
         ar_ap=ar_ap+1.D0
       endif
     endif
   endif
   
   !female fecundity
   if (ndg>0) then
     l_o=0.D0
     do i=1,no
       do j=1,np
         if (noij(i,j)>0) l_o=l_o+noij(i,j)*dlog(ksi(i,j)/sksi(i))
       enddo
     enddo
     
     do k=1,np
       if (pfem(k)==0.D0) cycle
       x_o=femfec(k)
       p_o=-((dlog(x_o)-mlogfemfec(k))**2)/(2.D0*femsigm**2)
       x_n=dexp(dlog(x_o)+normaldev(0.D0,v_ff))
       call recomp_ksi(k,x_o,x_n)
       p_n=-((dlog(x_n)-mlogfemfec(k))**2)/(2.D0*femsigm**2)
       
       l_n=0.D0
       do i=1,no
         do j=1,np
         if (noij(i,j)>0) l_n=l_n+noij(i,j)*dlog(ksi(i,j)/sksi(i))
         enddo
       enddo
       
       if (dlog(rnd())>=l_n-l_o+p_n-p_o) then
         call recomp_ksi(k,x_n,x_o)
         rr_ff=rr_ff+1.D0
       else
         femfec(k)=x_n
         l_o=l_n
         ar_ff=ar_ff+1.D0
       endif
     enddo
   endif
   
   if (ndg>0) then
     x_o=femsigm
     femsigm=femsigm+normaldev(0.D0,v_sf)
     if (femsigm<0.D0) femsigm=-femsigm !reflection to get positive sigma
     l_o=0.D0
     l_n=0.D0
     do i=1,np
       if (pfem(i)==0.D0) cycle
       l_o=l_o-((dlog(femfec(i))-mlogfemfec(i))**2)/(2.D0*x_o**2)-dlog(x_o)
       l_n=l_n-((dlog(femfec(i))-mlogfemfec(i))**2)/(2.D0*femsigm**2)-dlog(femsigm)
     enddo
     
     if (dlog(rnd())>=l_n-l_o) then !uniform prior for sigma
     !if (dlog(rnd())>=l_n-l_o-1.D0/femsigm**2+1.D0/x_o**2) then !inv.gamma(1,1) prior for sigma^2
       femsigm=x_o
       rr_sf=rr_sf+1.D0
     else
       ar_sf=ar_sf+1.D0
     endif
   endif
   
   if (ndg>0) then
     do k=1,nf
       if (incl(11+k)==0) cycle
       
       l_o=0.D0
       do i=1,np
         if (pfem(i)==0.D0) cycle
         l_o=l_o-((dlog(femfec(i))-mlogfemfec(i))**2)/(2.D0*femsigm**2)
       enddo
       x_o=theta(11+k)
       p_o=-(x_o**2)/(20.D0) !note that the normal prior has a variance of 10
       theta(11+k)=x_o+normaldev(0.D0,v_g)
       p_n=-(theta(11+k)**2)/(20.D0)
       call compute_mlogfemfec(k,x_o,theta(11+k))
       l_n=0.D0
       do i=1,np
         if (pfem(i)==0.D0) cycle
         l_n=l_n-((dlog(femfec(i))-mlogfemfec(i))**2)/(2.D0*femsigm**2)
       enddo
       if (dlog(rnd())>=l_n-l_o+p_n-p_o) then
         call compute_mlogfemfec(k,theta(11+k),x_o)
         theta(11+k)=x_o
         rr_g=rr_g+1.D0
       else
         ar_g=ar_g+1.D0
       endif
     enddo
   endif
   
   !male fecundity
   if (.true.) then
     l_o=0.D0
     do i=1,np
       do j=1,np
       if (npjk(i,j)>0) l_o=l_o+npjk(i,j)*dlog(phi(i,j)/sphi(i))
       enddo
     enddo
     
     do k=1,np
       if (pfem(k)==1.D0) cycle
       x_o=malfec(k)
       p_o=-((dlog(x_o)-mlogmalfec(k))**2)/(2.D0*malsigm**2)
       x_n=dexp(dlog(x_o)+normaldev(0.D0,v_fm))
       call recomp_phi(k,x_o,x_n)
       p_n=-((dlog(x_n)-mlogmalfec(k))**2)/(2.D0*malsigm**2)
       
       l_n=0.D0
       do i=1,np
         do j=1,np
         if (npjk(i,j)>0) l_n=l_n+npjk(i,j)*dlog(phi(i,j)/sphi(i))
         enddo
       enddo
       
       if (dlog(rnd())>=l_n-l_o+p_n-p_o) then
         call recomp_phi(k,x_n,x_o)
         rr_fm=rr_fm+1.D0
       else
         malfec(k)=x_n
         l_o=l_n
         ar_fm=ar_fm+1.D0
       endif
     enddo
   endif
   
   if (.true.) then
     x_o=malsigm
     malsigm=malsigm+normaldev(0.D0,v_sm)
     if (malsigm<0.D0) malsigm=-malsigm
     l_o=0.D0
     l_n=0.D0
     do i=1,np
       if (pfem(i)==1.D0) cycle
       l_o=l_o-((dlog(malfec(i))-mlogmalfec(i))**2)/(2.D0*x_o**2)-dlog(x_o)
       l_n=l_n-((dlog(malfec(i))-mlogmalfec(i))**2)/(2.D0*malsigm**2)-dlog(malsigm)
     enddo
     
     if (dlog(rnd())>=l_n-l_o) then !uniform prior for sigma (works better)
     !if (dlog(rnd())>=l_n-l_o-1.D0/malsigm**2+1.D0/x_o**2) then !inv.gamma(1,1) prior for sigma^2
       malsigm=x_o
       rr_sm=rr_sm+1.D0
     else
       ar_sm=ar_sm+1.D0
     endif
   endif
   
   if (.true.) then
     do k=1,nf
       if (incl(11+nf+k)==0) cycle
       
       l_o=0.D0
       do i=1,np
         if (pfem(i)==1.D0) cycle
         l_o=l_o-((dlog(malfec(i))-mlogmalfec(i))**2)/(2.D0*malsigm**2)
       enddo
       x_o=theta(11+nf+k)
       p_o=-(x_o**2)/(20.D0)
       theta(11+nf+k)=x_o+normaldev(0.D0,v_b)
       p_n=-(theta(11+nf+k)**2)/(20.D0)
       call compute_mlogmalfec(k,x_o,theta(11+nf+k))
       l_n=0.D0
       do i=1,np
         if (pfem(i)==1.D0) cycle
         l_n=l_n-((dlog(malfec(i))-mlogmalfec(i))**2)/(2.D0*malsigm**2)
       enddo
       if (dlog(rnd())>=l_n-l_o+p_n-p_o) then
         call compute_mlogmalfec(k,theta(11+nf+k),x_o)
         theta(11+nf+k)=x_o
         rr_b=rr_b+1.D0
       else
         ar_b=ar_b+1.D0
       endif
     enddo
   endif
   
   nth=usernth
   if (nl<nth) nth=nl
   
   !genotyping errors for nuclear markers
   if (incl_mtyp==1) then
     bb(1)=m_e*(dble(1)-y_e)/y_e
     bb(2)=(dble(1)-m_e)*(dble(1)-y_e)/y_e
     
     !$omp parallel num_threads(usernth) default(shared) private(k,th,tjsr,x_o,x_n,l_o,l_n,p_o,p_n,xx)
     !$omp do
     do k=1,nl
       if (mtyp(k)==0.D0) cycle
       th=omp_get_thread_num()+1
       tjsr=pjsr(th)
       xx(1)=mtyp(k)
       xx(2)=dble(1)-mtyp(k)
       p_o = log_DirichletD(xx,bb)
       x_o=mtyp(k)
       l_o=full_logl_mf(k)
       x_n=prectdev(x_o,d_e(k),tjsr)
       mtyp(k)=x_n
       xx(1)=mtyp(k)
       xx(2)=dble(1)-mtyp(k)
       p_n = log_DirichletD(xx,bb)
       l_n=full_logl_mf(k)
       if (dlog(prnd(tjsr))>=l_n-l_o+p_n-p_o) then
         rr_e(k)=rr_e(k)+1.D0
         mtyp(k)=x_o
       else
         ar_e(k)=ar_e(k)+1.D0
       endif
       pjsr(th)=tjsr
     enddo
     !$omp end do
     !$omp end parallel
     
     !Hyper-priors for mistyping errors
     x_o=m_e
     bo(1)=m_e*(dble(1)-y_e)/y_e
     bo(2)=(dble(1)-m_e)*(dble(1)-y_e)/y_e
     l_o=dble(0)
     
     m_e=rectdev(x_o,d_m_e)
     bn(1)=m_e*(dble(1)-y_e)/y_e
     bn(2)=(dble(1)-m_e)*(dble(1)-y_e)/y_e
     l_n=dble(0)
     do i=1,nl
       if (mtyp(i)==0.D0) cycle
       xx(1)=mtyp(i)
       xx(2)=dble(1)-mtyp(i)
       l_o = l_o + log_DirichletD(xx,bo)
       l_n = l_n + log_DirichletD(xx,bn)
     enddo
     
     if (dlog(rnd()).ge.(l_n - l_o)) then
       m_e = x_o
       rr_m_e=rr_m_e+dble(1)
     else
       l_o=l_n
       ar_m_e=ar_m_e+dble(1)
     endif
     
     x_o=y_e
     y_e=rectdev(x_o,d_y_e)
     bn(1)=m_e*(dble(1)-y_e)/y_e
     bn(2)=(dble(1)-m_e)*(dble(1)-y_e)/y_e
     l_n=dble(0)
     do i=1,nl
       if (mtyp(i)==0.D0) cycle
       xx(1)=mtyp(i)
       xx(2)=dble(1)-mtyp(i)
       l_n = l_n + log_DirichletD(xx,bn)
     enddo
     
     if (dlog(rnd()).ge.(l_n - l_o)) then
       y_e = x_o
       rr_y_e=rr_y_e+dble(1)
     else
       ar_y_e=ar_y_e+dble(1)
     endif
   endif
   
   !Mistyping error for cytoplasmic DNA marker
   if (incl_mcytyp==1 .and. mcytyp>0.D0) then
     x_o=mcytyp
     l_o=full_logl_mf_cy()
     x_n=rectdev(x_o,d_ecy)
     mcytyp=x_n
     l_n=full_logl_mf_cy()
     if (dlog(rnd())>=l_n-l_o) then
       rr_ecy=rr_ecy+1.D0
       mcytyp=x_o
     else
       ar_ecy=ar_ecy+1.D0
     endif
   endif
   
   !Probability of maternal transmission of cyDNA
   if (incl_maternal_cy==1 .and. (maternal_cy>0.D0 .or. maternal_cy<1.D0)) then
     x_o=maternal_cy
     l_o=full_logl_mf_cy()
     x_n=rectdev(x_o,d_mcy)
     maternal_cy=x_n
     l_n=full_logl_mf_cy()
     if (dlog(rnd())>=l_n-l_o) then
       rr_mcy=rr_mcy+1.D0
       maternal_cy=x_o
     else
       ar_mcy=ar_mcy+1.D0
     endif
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine performs between-model jumps for regression models (female fitness)
 subroutine update_model_f
   integer :: i,m
   double precision :: l_o,l_n,p_o,p_n,x_o
   
   if (ndg==0 .or. nf==0 .or. nexcl_g==nf) return
   
   do 
     m=int(rnd()*dble(nf))+1
     if (excl(m)==0) exit
   enddo
   
   if (incl(11+m)==1) then
     x_o=theta(11+m)
     l_o=0.D0
     do i=1,np
       if (pfem(i)==0.D0) cycle
       l_o=l_o-((dlog(femfec(i))-mlogfemfec(i))**2)/(2.D0*femsigm**2)
     enddo
     p_o=-(x_o**2)*0.5D0/10.D0-dlog(10.D0)*0.5D0
     incl(11+m)=0
     theta(11+m)=0.D0
     call compute_mlogfemfec(m,x_o,theta(11+m))
     l_n=0.D0
     do i=1,np
       if (pfem(i)==0.D0) cycle
       l_n=l_n-((dlog(femfec(i))-mlogfemfec(i))**2)/(2.D0*femsigm**2)
     enddo
     p_n=-((theta(11+m)-mq_g(m))**2)*0.5D0/vq_g(m)-dlog(vq_g(m))*0.5D0
     if (dlog(rnd())>=l_n-l_o+p_n-p_o) then
       call compute_mlogfemfec(m,theta(11+m),x_o)
       theta(11+m)=x_o
       incl(11+m)=1
     endif
   else
     x_o=theta(11+m)
     l_o=0.D0
     do i=1,np
       if (pfem(i)==0.D0) cycle
       l_o=l_o-((dlog(femfec(i))-mlogfemfec(i))**2)/(2.D0*femsigm**2)
     enddo
     p_o=-((theta(11+m)-mq_g(m))**2)*0.5D0/vq_g(m)-dlog(vq_g(m))*0.5D0
     incl(11+m)=1
     theta(11+m)=normaldev(mq_g(m),dsqrt(vq_g(m)))
     call compute_mlogfemfec(m,x_o,theta(11+m))
     l_n=0.D0
     do i=1,np
       if (pfem(i)==0.D0) cycle
       l_n=l_n-((dlog(femfec(i))-mlogfemfec(i))**2)/(2.D0*femsigm**2)
     enddo
     p_n=-(theta(11+m)**2)*0.5D0/10.D0-dlog(10.D0)*0.5D0
     if (dlog(rnd())>=l_n-l_o+p_n-p_o) then
       call compute_mlogfemfec(m,theta(11+m),x_o)
       theta(11+m)=0.D0
       incl(11+m)=0
     endif
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine performs between-model jumps for regression models (male fitness)
 subroutine update_model_m
   integer :: i,m
   double precision :: l_o,l_n,p_o,p_n,x_o
   
   if (nf==0 .or. nexcl_b==nf) return
   
   do
     m=int(rnd()*dble(nf))+1
     if (excl(nf+m)==0) exit
   enddo
   
   if (incl(11+nf+m)==1) then
     x_o=theta(11+nf+m)
     l_o=0.D0
     do i=1,np
       if (pfem(i)==1.D0) cycle
       l_o=l_o-((dlog(malfec(i))-mlogmalfec(i))**2)/(2.D0*malsigm**2)
     enddo
     p_o=-(x_o**2)*0.5D0/10.D0-dlog(10.D0)*0.5D0
     incl(11+nf+m)=0
     theta(11+nf+m)=0.D0
     call compute_mlogmalfec(m,x_o,theta(11+nf+m))
     l_n=0.D0
     do i=1,np
       if (pfem(i)==1.D0) cycle
       l_n=l_n-((dlog(malfec(i))-mlogmalfec(i))**2)/(2.D0*malsigm**2)
     enddo
     p_n=-((theta(11+nf+m)-mq_b(m))**2)*0.5D0/vq_b(m)-dlog(vq_b(m))*0.5D0
     if (dlog(rnd())>=l_n-l_o+p_n-p_o) then
       call compute_mlogmalfec(m,theta(11+nf+m),x_o)
       theta(11+nf+m)=x_o
       incl(11+nf+m)=1
     endif
   else
     x_o=theta(11+nf+m)
     l_o=0.D0
     do i=1,np
       if (pfem(i)==1.D0) cycle
       l_o=l_o-((dlog(malfec(i))-mlogmalfec(i))**2)/(2.D0*malsigm**2)
     enddo
     p_o=-((theta(11+nf+m)-mq_b(m))**2)*0.5D0/vq_b(m)-dlog(vq_b(m))*0.5D0
     incl(11+nf+m)=1
     theta(11+nf+m)=normaldev(mq_b(m),dsqrt(vq_b(m)))
     call compute_mlogmalfec(m,x_o,theta(11+nf+m))
     l_n=0.D0
     do i=1,np
       if (pfem(i)==1.D0) cycle
       l_n=l_n-((dlog(malfec(i))-mlogmalfec(i))**2)/(2.D0*malsigm**2)
     enddo
     p_n=-(theta(11+nf+m)**2)*0.5D0/10.D0-dlog(10.D0)*0.5D0
     if (dlog(rnd())>=l_n-l_o+p_n-p_o) then
       call compute_mlogmalfec(m,theta(11+nf+m),x_o)
       theta(11+nf+m)=0.D0
       incl(11+nf+m)=0
     endif
   endif
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine adjusts parameters of proposal distributions in order to get a mixing rate of about 25-45%
 subroutine adjust_proposals(fct)
   double precision, intent(in) :: fct
   double precision :: ar
   integer :: i
   
   if (ar_g+rr_g>0.D0) then
     ar = ar_g / (ar_g+rr_g)
     if (ar < dble(0.25)) then
       v_g = v_g/fct
     else
       if (ar > dble(0.45)) then
         v_g = v_g*fct
       endif
     endif
     ar_g = 0.D0
     rr_g = 0.D0
   endif
   
   if (ar_b+rr_b>0.D0) then
     ar = ar_b / (ar_b+rr_b)
     if (ar < dble(0.25)) then
       v_b = v_b/fct
     else
       if (ar > dble(0.45)) then
         v_b = v_b*fct
       endif
     endif
     ar_b = 0.D0
     rr_b = 0.D0
   endif
   
   if (ar_ds+rr_ds>0.D0) then
     ar = ar_ds / (ar_ds+rr_ds)
     if (ar < dble(0.25)) then
       v_ds= v_ds/fct
     else
       if (ar > dble(0.45)) then
         v_ds = v_ds*fct
       endif
     endif
     ar_ds = 0.D0
     rr_ds = 0.D0
   endif
   
   if (ar_dp+rr_dp>0.D0) then
     ar = ar_dp / (ar_dp+rr_dp)
     if (ar < dble(0.25)) then
       v_dp= v_dp/fct
     else
       if (ar > dble(0.45)) then
         v_dp = v_dp*fct
       endif
     endif
     ar_dp = 0.D0
     rr_dp = 0.D0
   endif
   
   if (ar_bs+rr_bs>0.D0) then
     ar = ar_bs / (ar_bs+rr_bs)
     if (ar < dble(0.25)) then
       v_bs= v_bs/fct
     else
       if (ar > dble(0.45)) then
         v_bs = v_bs*fct
       endif
     endif
     ar_bs = 0.D0
     rr_bs = 0.D0
   endif
   
   if (ar_bp+rr_bp>0.D0) then
     ar = ar_bp / (ar_bp+rr_bp)
     if (ar < dble(0.25)) then
       v_bp= v_bp/fct
     else
       if (ar > dble(0.45)) then
         v_bp = v_bp*fct
       endif
     endif
     ar_bp = 0.D0
     rr_bp = 0.D0
   endif
   
   if (ar_ks+rr_ks>0.D0) then
     ar = ar_ks / (ar_ks+rr_ks)
     if (ar < dble(0.25)) then
       v_ks= v_ks/fct
     else
       if (ar > dble(0.45)) then
         v_ks = v_ks*fct
       endif
     endif
     ar_ks = 0.D0
     rr_ks = 0.D0
   endif
   
   if (ar_kp+rr_kp>0.D0) then
     ar = ar_kp / (ar_kp+rr_kp)
     if (ar < dble(0.25)) then
       v_kp= v_kp/fct
     else
       if (ar > dble(0.45)) then
         v_kp = v_kp*fct
       endif
     endif
     ar_kp = 0.D0
     rr_kp = 0.D0
   endif
   
   if (ar_as+rr_as>0.D0) then
     ar = ar_as / (ar_as+rr_as)
     if (ar < dble(0.25)) then
       v_as= v_as/fct
     else
       if (ar > dble(0.45)) then
         v_as = v_as*fct
       endif
     endif
     ar_as = 0.D0
     rr_as = 0.D0
   endif
   
   if (ar_ap+rr_ap>0.D0) then
     ar = ar_ap / (ar_ap+rr_ap)
     if (ar < dble(0.25)) then
       v_ap= v_ap/fct
     else
       if (ar > dble(0.45)) then
         v_ap = v_ap*fct
       endif
     endif
     ar_ap = 0.D0
     rr_ap = 0.D0
   endif
   
   if (ar_ff+rr_ff>0.D0) then
     ar = ar_ff / (ar_ff+rr_ff)
     if (ar < dble(0.25)) then
       v_ff= v_ff/fct
     else
       if (ar > dble(0.45)) then
         v_ff = v_ff*fct
       endif
     endif
     ar_ff = 0.D0
     rr_ff = 0.D0
   endif
   
   if (ar_fm+rr_fm>0.D0) then
     ar = ar_fm / (ar_fm+rr_fm)
     if (ar < dble(0.25)) then
       v_fm= v_fm/fct
     else
       if (ar > dble(0.45)) then
         v_fm = v_fm*fct
       endif
     endif
     ar_fm = 0.D0
     rr_fm = 0.D0
   endif
   
   if (ar_sf+rr_sf>0.D0) then
     ar = ar_sf / (ar_sf+rr_sf)
     if (ar < dble(0.25)) then
       v_sf= v_sf/fct
     else
       if (ar > dble(0.45)) then
         v_sf = v_sf*fct
       endif
     endif
     ar_sf = 0.D0
     rr_sf = 0.D0
   endif
   
   if (ar_sm+rr_sm>0.D0) then
     ar = ar_sm / (ar_sm+rr_sm)
     if (ar < dble(0.25)) then
       v_sm= v_sm/fct
     else
       if (ar > dble(0.45)) then
         v_sm = v_sm*fct
       endif
     endif
     ar_sm = 0.D0
     rr_sm = 0.D0
   endif
   
   do i=1,nl
     if (ar_e(i)+rr_e(i)>0.D0) then
       ar = ar_e(i) / (ar_e(i)+rr_e(i))
       if (ar < dble(0.25)) then
         d_e(i)= d_e(i)/fct
       else
         if (ar > dble(0.45)) then
           d_e(i) = d_e(i)*fct
         endif
       endif
       ar_e(i) = 0.D0
       rr_e(i) = 0.D0
     endif
   enddo
   
   if (ar_m_e+rr_m_e>0.D0) then
     ar = ar_m_e / (ar_m_e+rr_m_e)
     if (ar < dble(0.25)) then
       d_m_e = d_m_e/fct
     else
       if (ar > dble(0.45)) then
         d_m_e = d_m_e*fct
       endif
     endif
     ar_m_e = 0.D0
     rr_m_e = 0.D0
   endif
     
   if (ar_y_e+rr_y_e>0.D0) then
     ar = ar_y_e / (ar_y_e+rr_y_e)
     if (ar < dble(0.25)) then
       d_y_e = d_y_e/fct
     else
       if (ar > dble(0.45)) then
         d_y_e = d_y_e*fct
       endif
     endif
     ar_y_e = 0.D0
     rr_y_e = 0.D0
   endif
   
   if (ar_ecy+rr_ecy>0.D0) then
     ar = ar_ecy / (ar_ecy+rr_ecy)
     if (ar < dble(0.25)) then
       d_ecy= d_ecy/fct
     else
       if (ar > dble(0.45)) then
         d_ecy = d_ecy*fct
       endif
     endif
     ar_ecy = 0.D0
     rr_ecy = 0.D0
   endif
   
   if (ar_mcy+rr_mcy>0.D0) then
     ar = ar_mcy / (ar_mcy+rr_mcy)
     if (ar < dble(0.25)) then
       d_mcy= d_mcy/fct
     else
       if (ar > dble(0.45)) then
         d_mcy = d_mcy*fct
       endif
     endif
     ar_mcy = 0.D0
     rr_mcy = 0.D0
   endif
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine implements the main algorithm for RJ-MCMC (burn-in stage + pilot tuning)
 subroutine run_pilotmcmc(ncycles)
   integer :: i,j,ncycles,burnin,keepevery,nq,prg
   real(8) :: tim_a,tim_b
   
   fixedeffects=.false.
   keepevery=20
   allocate(mq_g(nf))
   allocate(vq_g(nf))
   allocate(mq_b(nf))
   allocate(vq_b(nf))
   mq_g=0.D0
   vq_g=0.D0
   mq_b=0.D0
   vq_b=0.D0
   burnin=ncycles/2
   nq=0
   prg=0
   write(*,"(a)") '*** Markov Chain Monte Carlo approximation to the posterior distribution ***'
   write(*,"('MCMC progress (pilot tuning) ')",advance="no")
   tim_a=omp_get_wtime()
   do i=1,ncycles
     call genealogies
     call Bayes_update_parmeters
     
     if (mod(i,keepevery)==0) then
       if (i*100/ncycles==prg) then
         write (*,"('.')",advance="no")
         call flush()
       else
         tim_b=(omp_get_wtime()-tim_a)/dble(i)*dble(ncycles-i)/dble(60)
         !tim_b=(omp_get_wtime()-tim_a)/dble(i)*dble(120000)/dble(60)
         prg=i*100/ncycles
         write(*,"(i3,'% (est. time to complete pilot run: ',i0,' min.)')",advance="yes") prg,int(tim_b)
         if (prg<100) write(*,"('MCMC progress (pilot tuning) ')",advance="no")
         call flush()
       endif
       
       if (i>burnin) then
         nq=nq+1
         do j=1,nf
           mq_g(j)=mq_g(j)+theta(11+j)
           vq_g(j)=vq_g(j)+theta(11+j)**2
           mq_b(j)=mq_b(j)+theta(11+nf+j)
           vq_b(j)=vq_b(j)+theta(11+nf+j)**2
         enddo
       endif
     endif
     if (mod(i,1000)==0) call adjust_proposals(dsqrt(2.D0))
   enddo
   
   do j=1,nf
     mq_g(j)=mq_g(j)/dble(nq)
     mq_b(j)=mq_b(j)/dble(nq)
     vq_g(j)=(vq_g(j)/dble(nq)-mq_g(j)**2)*dble(nq)/dble(nq-1)
     vq_b(j)=(vq_b(j)/dble(nq)-mq_b(j)**2)*dble(nq)/dble(nq-1)
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine implements the main algorithm for RJ-MCMC (sampling stage + post-processing)
 subroutine run_mcmc(ncycles,keepevery)
   integer :: i,j,k,s,ncycles,nsamples,keepevery,mod_f,mod_m,prg,fecke
   double precision :: x,y,z,mlogl,vlogl,hdi_p,sumsina,sumcosa,lppd,pw
   double precision, allocatable :: mtheta(:),mfemfec(:),mmalfec(:),mmtyp(:),mmcytyp,mmat_cy !posterior means
   logical :: mixing_ok
   real(8) :: tim_a,tim_b
   
   hdi_p=9.5D-1
   fixedeffects=.false.
   allocate(mtheta(11+2*nf))
   allocate(mfemfec(np))
   allocate(mmalfec(np))
   allocate(mmtyp(nl))
   mtheta=0.D0
   mfemfec=0.D0
   mmalfec=0.D0
   mmtyp=0.D0
   mmcytyp=0.D0
   mmat_cy=0.D0
   mlogl=0.D0
   vlogl=0.D0
   nsamples=ncycles/keepevery
   fecke=ncycles/nfecsaved
   s=0
   allocate(lg(nf,nsamples))
   allocate(lb(nf,nsamples))
   allocate(lmodg(nsamples))
   allocate(lmodb(nsamples))
   allocate(lr2g(nsamples))
   allocate(lr2b(nsamples))
   allocate(lvg(nsamples))
   allocate(lvb(nsamples))
   allocate(llogfemfec(np,nsamples))
   allocate(llogmalfec(np,nsamples))
   allocate(ltheta(11,nsamples))
   allocate(lproi(no,nsamples))
   allocate(lmtyp(nl,nsamples))
   allocate(lmcytyp(nsamples))
   allocate(lmat_cy(nsamples))
   allocate(lm_e(nsamples))
   allocate(ly_e(nsamples))
   allocate(frg(nf))
   allocate(frb(nf))
   allocate(geneal_m(no,nsamples))
   allocate(geneal_f(no,nsamples))
   frg=0
   frb=0
   
   j=0
   do i=1,len(trim(fo))
     if (fo(i:i)=='.') then
       j=i
       write(fm,"(a,'.mcmc')") fo(1:i-1)
       write(fmf,"(a,'.ffec')") fo(1:i-1)
       write(fmm,"(a,'.mfec')") fo(1:i-1)
       exit
     endif
   enddo
   if (j==0) then
     write(fm,"(a,'.mcmc')") fo(1:len(trim(fo)))
     write(fmf,"(a,'.ffec')") fo(1:len(trim(fo)))
     write(fmm,"(a,'.mfec')") fo(1:len(trim(fo)))
   endif
   
   if (writecov==1) then
     open(unit=789,file=trim(fm),action="write",status="replace")
     open(unit=78,file=trim(fmf),action="write",status="replace")
     open(unit=79,file=trim(fmm),action="write",status="replace")
   endif
   
   prg=0
   write(*,"('MCMC progress (sampling) ')",advance="no")
   tim_a=omp_get_wtime()
   do i=1,ncycles
     call genealogies
     call Bayes_update_parmeters
     call update_model_f
     call update_model_m
     
     if (mod(i,keepevery)==0) then
       
       mtheta=mtheta+theta/dble(nsamples)
       mfemfec=mfemfec+dlog(femfec)/dble(nsamples)
       mmalfec=mmalfec+dlog(malfec)/dble(nsamples)
       mmtyp=mmtyp+mtyp/dble(nsamples)
       mmcytyp=mmcytyp+mcytyp/dble(nsamples)
       mmat_cy=mmat_cy+maternal_cy/dble(nsamples)
       
       frg=frg+incl(12:11+nf)
       frb=frb+incl(12+nf:12+2*nf)
       
       new_logl=full_logl()
       mlogl=mlogl+new_logl/dble(nsamples)
       vlogl=vlogl+dble(nsamples)*(new_logl/dble(nsamples))**2
       call bayes_r2
       call ytoint(mod_f,mod_m)
       
       s=s+1
       lproi(:,s)=proi(:)
       if (nf>0) then
         lg(:,s)=theta(12:11+nf)
         lb(:,s)=theta(12+nf:11+2*nf)
       endif
       lmodg(s)=mod_f
       lmodb(s)=mod_m
       lr2g(s)=femr2
       lr2b(s)=malr2
       lvg(s)=femsigm
       lvb(s)=malsigm
       llogfemfec(:,s)=dlog(femfec)
       llogmalfec(:,s)=dlog(malfec)
       ltheta(:,s)=theta(1:11)
       lmtyp(:,s)=mtyp
       lmcytyp(s)=mcytyp
       lmat_cy(s)=maternal_cy
       lm_e(s)=m_e
       ly_e(s)=y_e
       
       geneal_m(:,s)=parass(:)%mo1
       geneal_f(:,s)=parass(:)%fa1
       
       if (i*100/ncycles==prg) then
         write (*,"('.')",advance="no")
         call flush()
       else
         prg=i*100/ncycles
         tim_b=(omp_get_wtime()-tim_a)/dble(i)*dble(ncycles-i)/dble(60)
         write(*,"(i3,'% (est. time to complete sampling run: ',i0,' min.)')",advance="yes") prg,int(tim_b)
         if (prg<100) write(*,"('MCMC progress (sampling) ')",advance="no")
         call flush()
       endif
       
       if (writecov==1) then
         write(789,"(i0,a,e16.8,2(a,i0),*(a,f9.6))") i,tab,new_logl,tab,mod_f,tab,mod_m,(tab,theta(j),j=1,11+2*nf) &
         ,tab,femsigm,tab,malsigm,tab,femr2,tab,malr2,tab,m_e,tab,y_e,(tab,mtyp(j),j=1,nl),tab,mcytyp,tab,maternal_cy
         flush(789)
       endif
       
       if (mod(i,fecke)==0 .and. writecov==1) then
         write(78,"(i0,*(a,e14.6))") i,(tab,dlog(femfec(j)),j=1,np)
         flush(78)
         write(79,"(i0,*(a,e14.6))") i,(tab,dlog(malfec(j)),j=1,np)
         flush(79)
       endif
     endif
     
   enddo
   if (writecov==1) then
     close(789)
     close(78)
     close(79)
   endif
   
   j=0
   do i=1,len(trim(fo))
     if (fo(i:i)=='.') then
       j=i
       write(fm,"(a,'.hnm')") fo(1:i-1)
       exit
     endif
   enddo
   if (j==0) write(fm,"(a,'.hnm')") fo(1:len(trim(fo)))
   open(unit=789,file=trim(fm),action="write",status="replace")
   
   !Final estimates of model parameters
   
   if (incl(4)==1) ltheta(4,:)=1.D0/ltheta(4,:)
   if (incl(5)==1) ltheta(5,:)=1.D0/ltheta(5,:)
   
   !Posterior mean dispersal directions
   sumsina=0.D0
   sumcosa=0.D0
   do i=1,nsamples
     sumsina=sumsina+sin(ltheta(10,i))
     sumcosa=sumcosa+cos(ltheta(10,i))
   enddo
   mtheta(10)=atan2(sumsina,sumcosa)
   sumsina=0.D0
   sumcosa=0.D0
   do i=1,nsamples
     sumsina=sumsina+sin(ltheta(11,i))
     sumcosa=sumcosa+cos(ltheta(11,i))
   enddo
   mtheta(11)=atan2(sumsina,sumcosa)
   
   theta=mtheta
   mtyp=mmtyp
   mcytyp=mmcytyp
   maternal_cy=mmat_cy
   femfec=dexp(mfemfec)
   malfec=dexp(mmalfec)
   new_logl=full_logl()
   vlogl=dble(nsamples)*(vlogl-mlogl**2)/dble(nsamples-1)
   call compute_lppd(lppd,pw,nsamples)
   
   !Output formatting:
   write(789,"('# Model support')")
   write(789,"('LogL.avg.theta',a,e16.8)") tab,new_logl
   write(789,"('avg.LogL',a,e16.8)") tab,mlogl
   write(789,"('var.LogL',a,e16.8)") tab,vlogl
   write(789,"('WAIC',a,e16.8)") tab,-2.D0*(lppd-pw)
   write(789,"('effect.params',a,e16.8)") tab,pw
   
   write(789,*)
   write(789,"('# Neighborhood model assumptions')")
   if (linked_migr==1) write (789,"('Linked seed and pollen migration rates')")
   write(789,"(3a)") 'Seed dispersal kernel: ',tab,kernel_type(seed_kernel)
   write(789,"(3a)") 'Pollen dispersal kernel: ',tab,kernel_type(pollen_kernel)
   if (neighborhood<dble(0)) then
     write(789,"(3a)") 'Neighborhood radius:',tab,'unlimited'
   else 
     write(789,"(2a,f10.2)") 'Neighborhood radius:',tab,neighborhood
   endif
   
   !Here are the parameters that enter directly to the likelihood function:
   write(789,*)
   write(789,"('# Posterior distribution of model parameters')")
   write(789,"(9a)") 'Param.',tab,'Incl',tab,'Median',tab,'HPDl(95%)',tab,'HPDh(95%)'
   do i=1,9
     call hdi(ltheta(i,:),nsamples,hdi_p,x,y)
     write(789,"(2a,i0,3(a,e16.8))") trim(adjustl(ntheta(i))),tab,incl(i),tab,ltheta(i,nsamples/2),tab,x,tab,y
   enddo
   do i=10,11
     call circular_estims(ltheta(i,:),nsamples,hdi_p,x,y,z)
     write(789,"(a,'*',a,i0,3(a,e16.8))") trim(adjustl(ntheta(i))),tab,incl(i),tab,x,tab,y,tab,z
   enddo
   do i=1,np
     call hdi(llogfemfec(i,:),nsamples,hdi_p,x,y)
     if (pfem(i)>0.D0 .and. ndg>0) then
       write(789,"('log.fem.fec.',i0,a,i0,3(a,e16.8))") pid(i),tab,1,tab,llogfemfec(i,nsamples/2),tab,x,tab,y
     else
       write(789,"('log.fem.fec.',i0,a,i0,3(a,e16.8))") pid(i),tab,0,tab,llogfemfec(i,nsamples/2),tab,x,tab,y
     endif
   enddo
   do i=1,np
     call hdi(llogmalfec(i,:),nsamples,hdi_p,x,y)
     if (pfem(i)<1.D0) then
       write(789,"('log.mal.fec.',i0,a,i0,3(a,e16.8))") pid(i),tab,1,tab,llogmalfec(i,nsamples/2),tab,x,tab,y
     else
       write(789,"('log.mal.fec.',i0,a,i0,3(a,e16.8))") pid(i),tab,0,tab,llogmalfec(i,nsamples/2),tab,x,tab,y
     endif
   enddo
   do i=1,nl
     if (mtyp(i)>0.D0 .and. incl_mtyp==1) then
       call hdi(lmtyp(i,:),nsamples,hdi_p,x,y)
       write(789,"('mtyp.nu.',i0,a,i0,3(a,e16.8))") i,tab,1,tab,lmtyp(i,nsamples/2),tab,x,tab,y
     else
       x=mtyp(i)
       y=mtyp(i)
       write(789,"('mtyp.nu.',i0,a,i0,3(a,e16.8))") i,tab,0,tab,mtyp(i),tab,x,tab,y
     endif
   enddo
   
   call hdi(lmcytyp(:),nsamples,hdi_p,x,y)
   write(789,"('mtyp.cy',a,i0,3(a,e16.8))") tab,incl_mcytyp,tab,lmcytyp(nsamples/2),tab,x,tab,y
   call hdi(lmat_cy(:),nsamples,hdi_p,x,y)
   write(789,"('maternal.cy',a,i0,3(a,e16.8))") tab,incl_maternal_cy,tab,lmat_cy(nsamples/2),tab,x,tab,y
   
   !Here are the hyper-parameters for genotyping error rates
   write(789,*)
   write(789,"('# Posterior distribution of hyper-parameters for genotyping errors')")
   write(789,"(9a)") 'Param.',tab,'Incl',tab,'Median',tab,'HPDl(95%)',tab,'HPDh(95%)'
   call hdi(lm_e,nsamples,hdi_p,x,y)
   write(789,"('m_e.nu',a,i0,3(a,e16.8))") tab,incl_mtyp,tab,lm_e(nsamples/2),tab,x,tab,y
   call hdi(ly_e,nsamples,hdi_p,x,y)
   write(789,"('y_e.nu',a,i0,3(a,e16.8))") tab,incl_mtyp,tab,ly_e(nsamples/2),tab,x,tab,y
   
   !Here are the hyper-parameters of fecundity models, separately for female and male reproductive function
   call identify_mods(nsamples)
   write(789,*)
   write(789,"('# Posterior distribution of female fecundity models')")
   if (ndg>0) then
     do k=1,nmodg
       write(789,"('Posterior frequency of model.',i0,a,f6.4)") modng(k),tab,dble(modfg(k))/dble(nsamples)
     enddo
     write(789,*)
     write(789,"('# Posterior inclusion frequency of phenotypic variables (female fec.)')")
     do k=1,nf
       write(789,"('Posterior frequency of var.',i0,a,f6.4)") k,tab,dble(frg(k))/dble(nsamples)
     enddo
     write(789,*)
     write(789,"('# Regression parameter estimates for the best female fecundity model')")
     write(789,"(9a)") 'Param.',tab,'Incl',tab,'Median',tab,'HPDl(95%)',tab,'HPDh(95%)'
     do k=1,nf
       call hdi(lg(k,1:modfg(1)),modfg(1),hdi_p,x,y)
       write(789,"('gamma.',i0,a,i0,*(a,e16.8))") k,tab,1-excl(k),tab,lg(k,modfg(1)/2),tab,x,tab,y
     enddo
     
     call hdi(lvg(1:modfg(1)),modfg(1),hdi_p,x,y)
     write(789,"('sigm.g',a,i0,*(a,e16.8))") tab,1,tab,lvg(modfg(1)/2),tab,x,tab,y
     call hdi(lr2g(1:modfg(1)),modfg(1),hdi_p,x,y)
     write(789,"('R^2.g',a,i0,*(a,e16.8))") tab,1,tab,lr2g(modfg(1)/2),tab,x,tab,y
   else
     write(789,"('Results not available since data do not contain information on female fecundity...')")
   endif
   
   write(789,*)
   write(789,"('# Posterior distribution of male fecundity models')")
   do k=1,nmodb
     write(789,"('Posterior frequency of model.',i0,a,f6.4)") modnb(k),tab,dble(modfb(k))/dble(nsamples)
   enddo
   
   write(789,*)
   write(789,"('# Posterior inclusion frequency of phenotypic variables (male fec.)')")
   do k=1,nf
     write(789,"('Posterior frequency of var.',i0,a,f6.4)") k,tab,dble(frb(k))/dble(nsamples)
   enddo
   
   write(789,*)
   write(789,"('# Regression parameter estimates for the best male fecundity model')")
   write(789,"(9a)") 'Param.',tab,'Incl',tab,'Median',tab,'HPDl(95%)',tab,'HPDh(95%)'
   do k=1,nf
     call hdi(lb(k,1:modfb(1)),modfb(1),hdi_p,x,y)
     write(789,"('beta.',i0,a,i0,*(a,e16.8))") k,tab,1-excl(nf+k),tab,lb(k,modfb(1)/2),tab,x,tab,y
   enddo
   
   call hdi(lvb(1:modfb(1)),modfb(1),hdi_p,x,y)
   write(789,"('sigm.b',a,i0,*(a,e16.8))") tab,1,tab,lvb(modfb(1)/2),tab,x,tab,y
   call hdi(lr2b(1:modfb(1)),modfb(1),hdi_p,x,y)
   write(789,"('R^2.b',a,i0,*(a,e16.8))") tab,1,tab,lr2b(modfb(1)/2),tab,x,tab,y
   
   !Here are the final acceptance rates for parameters estimated with the Metropolis-Hastings algorithm
   mixing_ok=.true.
   write(789,*)
   write(789,"('# Metropolis-Hastings acceptance rates')")
   
   if (ar_ds+rr_ds>0.D0) then
     write(789,"('ds',a,i0,'%')", advance="no") tab,int(100*(ar_ds/(ar_ds+rr_ds)))
     if (ar_ds/(ar_ds+rr_ds)<0.25D0 .or. ar_ds/(ar_ds+rr_ds)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_dp+rr_dp>0.D0) then
     write(789,"('dp',a,i0,'%')", advance="no") tab,int(100*(ar_dp/(ar_dp+rr_dp)))
     if (ar_dp/(ar_dp+rr_dp)<0.25D0 .or. ar_dp/(ar_dp+rr_dp)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_bs+rr_bs>0.D0) then
     write(789,"('bs',a,i0,'%')", advance="no") tab,int(100*(ar_bs/(ar_bs+rr_bs)))
     if (ar_bs/(ar_bs+rr_bs)<0.25D0 .or. ar_bs/(ar_bs+rr_bs)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_bp+rr_bp>0.D0) then
     write(789,"('bp',a,i0,'%')", advance="no") tab,int(100*(ar_bp/(ar_bp+rr_bp)))
     if (ar_bp/(ar_bp+rr_bp)<0.25D0 .or. ar_bp/(ar_bp+rr_bp)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_ks+rr_ks>0.D0) then
     write(789,"('ks',a,i0,'%')", advance="no") tab,int(100*(ar_ks/(ar_ks+rr_ks)))
     if (ar_ks/(ar_ks+rr_ks)<0.25D0 .or. ar_ks/(ar_ks+rr_ks)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_kp+rr_kp>0.D0) then
     write(789,"('kp',a,i0,'%')", advance="no") tab,int(100*(ar_kp/(ar_kp+rr_kp)))
     if (ar_kp/(ar_kp+rr_kp)<0.25D0 .or. ar_kp/(ar_kp+rr_kp)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_as+rr_as>0.D0) then
     write(789,"('as',a,i0,'%')", advance="no") tab,int(100*(ar_as/(ar_as+rr_as)))
     if (ar_as/(ar_as+rr_as)<0.25D0 .or. ar_as/(ar_as+rr_as)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write (789,"('')", advance="yes")
     endif
   endif
   
   if (ar_ap+rr_ap>0.D0) then
     write(789,"('ap',a,i0,'%')", advance="no") tab,int(100*(ar_ap/(ar_ap+rr_ap)))
     if (ar_ap/(ar_ap+rr_ap)<0.25D0 .or. ar_ap/(ar_ap+rr_ap)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_ff+rr_ff>0.D0) then
     write(789,"('ff',a,i0,'%')", advance="no") tab,int(100*(ar_ff/(ar_ff+rr_ff)))
     if (ar_ff/(ar_ff+rr_ff)<0.25D0 .or. ar_ff/(ar_ff+rr_ff)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_fm+rr_fm>0.D0) then
     write(789,"('fm',a,i0,'%')", advance="no") tab,int(100*(ar_fm/(ar_fm+rr_fm)))
     if (ar_fm/(ar_fm+rr_fm)<0.25D0 .or. ar_fm/(ar_fm+rr_fm)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_sf+rr_sf>0.D0) then
     write(789,"('sf',a,i0,'%')", advance="no") tab,int(100*(ar_sf/(ar_sf+rr_sf)))
     if (ar_sf/(ar_sf+rr_sf)<0.25D0 .or. ar_sf/(ar_sf+rr_sf)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_sm+rr_sm>0.D0) then
     write(789,"('sm',a,i0,'%')", advance="no") tab,int(100*(ar_sm/(ar_sm+rr_sm)))
     if (ar_sm/(ar_sm+rr_sm)<0.25D0 .or. ar_sm/(ar_sm+rr_sm)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_g+rr_g>0.D0) then
     write(789,"('g',a,i0,'%')", advance="no") tab,int(100*(ar_g/(ar_g+rr_g)))
     if (ar_g/(ar_g+rr_g)<0.25D0 .or. ar_g/(ar_g+rr_g)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_b+rr_b>0.D0) then
     write(789,"('b',a,i0,'%')", advance="no") tab,int(100*(ar_b/(ar_b+rr_b)))
     if (ar_b/(ar_b+rr_b)<0.25D0 .or. ar_b/(ar_b+rr_b)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   do i=1,nl
     if (ar_e(i)+rr_e(i)>0.D0) then
       write(789,"('e',i0,a,i0,'%')", advance="no") i,tab,int(100*(ar_e(i)/(ar_e(i)+rr_e(i))))
       if (ar_e(i)/(ar_e(i)+rr_e(i))<0.25D0 .or. ar_e(i)/(ar_e(i)+rr_e(i))>0.45D0) then
         write(789,"(a,'(!)')", advance="yes") tab
         mixing_ok=.false.
       else
         write(789,"('')", advance="yes")
       endif
     endif
   enddo
   
   if (ar_ecy+rr_ecy>0.D0) then
     write(789,"('ecy',a,i0,'%')", advance="no") tab,int(100*(ar_ecy/(ar_ecy+rr_ecy)))
     if (ar_ecy/(ar_ecy+rr_ecy)<0.25D0 .or. ar_ecy/(ar_ecy+rr_ecy)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_m_e+rr_m_e>0.D0) then
     write(789,"('me',a,i0,'%')", advance="no") tab,int(100*(ar_m_e/(ar_m_e+rr_m_e)))
     if (ar_m_e/(ar_m_e+rr_m_e)<0.25D0 .or. ar_m_e/(ar_m_e+rr_m_e)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   
   if (ar_y_e+rr_y_e>0.D0) then
     write(789,"('ye',a,i0,'%')", advance="no") tab,int(100*(ar_y_e/(ar_y_e+rr_y_e)))
     if (ar_y_e/(ar_y_e+rr_y_e)<0.25D0 .or. ar_y_e/(ar_y_e+rr_y_e)>0.45D0) then
       write(789,"(a,'(!)')", advance="yes") tab
       mixing_ok=.false.
     else
       write(789,"('')", advance="yes")
     endif
   endif
   if (mixing_ok .eqv. .false.) write(789,"('There might be some mixing issues in the MCMC chain...')")
   close(789)
   
   write(*,*)
   write(*,"(a)") 'Analysis finished normally...'
   write(*,"(a)") 'Results written to the output file: '//trim(fm)
   
   write(*,*)
   write(*,"(a)") 'Performing parentage assignment...'
   call post_genealogies(nsamples)
   write(*,"(a)") 'Parentage results written to the output file: '//trim(fp)
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes model indicators (single integers) based on inclusion indicators
 subroutine ytoint(mod_f,mod_m)
   integer :: mod_f,mod_m,i
   
   mod_f=0
   mod_m=0
   do i=1,nf
     mod_f=mod_f+incl(11+i)*(2**(i-1))
     mod_m=mod_m+incl(11+nf+i)*(2**(i-1))
   enddo
   
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine lists distinct models that appear during MCMC and computes their relative frequencies
 ! Model lists and frequencies are stored in global variables: modng and modnb (model indicators),
 ! nmodg abd nmodb (number of distinct models), modfg and modfb (mode frequencies)
 subroutine identify_mods(nsamp)
   integer :: nsamp,i,b,k
   integer, allocatable :: tmpmodn(:)
   double precision :: x
   double precision, allocatable :: ax(:)
   
   allocate(modng(0))
   allocate(tmpmodn(0))
   allocate(ax(nf))
   
   !identify all regression models for female fecundity that appeared in MCMC chain
   nmodg=0
   do i=1,nsamp
     b=0
     do k=1,nmodg
       if (lmodg(i)==modng(k)) b=b+1
     enddo
         
     if (b==0) then
       tmpmodn=modng
       nmodg=nmodg+1
       deallocate(modng)
       allocate(modng(nmodg))
       do k=1,nmodg-1
         modng(k)=tmpmodn(k)
       enddo
       modng(nmodg)=lmodg(i)
       deallocate(tmpmodn)
       allocate(tmpmodn(nmodg))
     endif
   enddo
   
   !compute model distribution
   allocate(modfg(nmodg))
   modfg=0
   do i=1,nsamp
     do k=1,nmodg
       if (lmodg(i)==modng(k)) modfg(k)=modfg(k)+1
     enddo
   enddo
   
   !sort models according to their posterior frequencies
   do i=1,nmodg-1
     do k=i+1,nmodg
       if (modfg(i)<modfg(k)) then
         b=modfg(i)
         modfg(i)=modfg(k)
         modfg(k)=b
         
         b=modng(i)
         modng(i)=modng(k)
         modng(k)=b
       endif
     enddo
   enddo
   
   !assign a regression model to a given MCMC sample
   do i=1,nsamp
     do k=1,nmodg
       if (lmodg(i)==modng(k)) then
         lmodg(i)=k
         exit
       endif
     enddo
   enddo
   
   !sort samples according to models' frequencies
   do i=1,nsamp-1
     do k=i+1,nsamp
       if (modfg(lmodg(i))<modfg(lmodg(k))) then
         b=lmodg(i)
         lmodg(i)=lmodg(k)
         lmodg(k)=b
         
         ax=lg(:,i)
         lg(:,i)=lg(:,k)
         lg(:,k)=ax
         
         x=lr2g(i)
         lr2g(i)=lr2g(k)
         lr2g(k)=x
         
         x=lvg(i)
         lvg(i)=lvg(k)
         lvg(k)=x
       endif
     enddo
   enddo
   
   deallocate(tmpmodn)
   allocate(modnb(0))
   allocate(tmpmodn(0))
   
   !identify all regression models for male fecundity that appeared in MCMC chain
   nmodb=0
   do i=1,nsamp
     b=0
     do k=1,nmodb
       if (lmodb(i)==modnb(k)) b=b+1
     enddo
         
     if (b==0) then
       tmpmodn=modnb
       nmodb=nmodb+1
       deallocate(modnb)
       allocate(modnb(nmodb))
       do k=1,nmodb-1
         modnb(k)=tmpmodn(k)
       enddo
       modnb(nmodb)=lmodb(i)
       deallocate(tmpmodn)
       allocate(tmpmodn(nmodb))
     endif
   enddo
   
   !compute model distribution
   allocate(modfb(nmodb))
   modfb=0
   do i=1,nsamp
     do k=1,nmodb
       if (lmodb(i)==modnb(k)) modfb(k)=modfb(k)+1
     enddo
   enddo
   
   !sort models according to their posterior frequencies
   do i=1,nmodb-1
     do k=i+1,nmodb
       if (modfb(i)<modfb(k)) then
         b=modfb(i)
         modfb(i)=modfb(k)
         modfb(k)=b
         
         b=modnb(i)
         modnb(i)=modnb(k)
         modnb(k)=b
       endif
     enddo
   enddo
   
   !assign a regression model to a given MCMC sample
   do i=1,nsamp
     do k=1,nmodb
       if (lmodb(i)==modnb(k)) then
         lmodb(i)=k
         exit
       endif
     enddo
   enddo
   
   !sort samples according to models' frequencies
   do i=1,nsamp-1
     do k=i+1,nsamp
       if (modfb(lmodb(i))<modfb(lmodb(k))) then
         b=lmodb(i)
         lmodb(i)=lmodb(k)
         lmodb(k)=b
         
         ax=lb(:,i)
         lb(:,i)=lb(:,k)
         lb(:,k)=ax
         
         x=lr2b(i)
         lr2b(i)=lr2b(k)
         lr2b(k)=x
         
         x=lvb(i)
         lvb(i)=lvb(k)
         lvb(k)=x
       endif
     enddo
   enddo
   
 end subroutine identify_mods
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes two the most likely genealogies for the i-th offspring (Bayesian approach)
 subroutine post_genealogy(i,nsamp,p1m,p1f,pr1,p2m,p2f,pr2)
   integer :: nsamp,ngt,i,j,k,l
   integer,intent(out) :: p1m,p1f,p2m,p2f
   double precision, intent(out) :: pr1,pr2
   integer, allocatable :: tmpgfi(:),tmpgti(:,:),gti(:,:),gfi(:)
   
   allocate(gfi(1))   !posterior frequency
   allocate(gti(1,2)) !different genealogy samples
   gfi(1)=0
   gti(1,1)=geneal_m(i,1)
   gti(1,2)=geneal_f(i,1)
   ngt=1 !number of different genealogies
   
   do j=1,nsamp
     l=0
     do k=1,ngt
       if (geneal_m(i,j)==gti(k,1) .and. geneal_f(i,j)==gti(k,2)) then
         l=l+1
         gfi(k)=gfi(k)+1
       endif
     enddo
     
     if (l==0) then
       allocate(tmpgfi(ngt))
       allocate(tmpgti(ngt,2))
       tmpgti=gti
       tmpgfi=gfi
       ngt=ngt+1
       deallocate(gti)
       deallocate(gfi)
       allocate(gti(ngt,2))
       allocate(gfi(ngt))
       do k=1,ngt-1
         gti(k,1)=tmpgti(k,1)
         gti(k,2)=tmpgti(k,2)
         gfi(k)=tmpgfi(k)
       enddo
       gti(ngt,1)=geneal_m(i,j)
       gti(ngt,2)=geneal_f(i,j)
       gfi(ngt)=1
       deallocate(tmpgti)
       deallocate(tmpgfi)
     endif
   enddo
   
   do j=1,ngt-1
     do k=j+1,ngt
       if (gfi(j)<gfi(k)) then
         !swap genealogies
         l=gti(k,1)
         gti(k,1)=gti(j,1)
         gti(j,1)=l
         
         l=gti(k,2)
         gti(k,2)=gti(j,2)
         gti(j,2)=l
         
         !swap frequencies
         l=gfi(k)
         gfi(k)=gfi(j)
         gfi(j)=l
       endif
     enddo
   enddo
   
   p1m=-1
   p1f=-1
   p2m=-1
   p2f=-1
   if (ngt>1) then
     if (gti(1,1)>0) p1m=pid(gti(1,1))
     if (gti(1,2)>0) p1f=pid(gti(1,2))
     if (gti(2,1)>0) p2m=pid(gti(2,1))
     if (gti(2,2)>0) p2f=pid(gti(2,2))
     pr1=dble(gfi(1))/dble(nsamp)
     pr2=dble(gfi(2))/dble(nsamp)
   else
     if (gti(1,1)>0) p1m=pid(gti(1,1))
     if (gti(1,2)>0) p1f=pid(gti(1,2))
     pr1=dble(gfi(1))/dble(nsamp)
     pr2=0.D0
   endif
   
 end subroutine post_genealogy
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine creates a text file containing two the most likely offspring genealogies (Bayesian approach)
 subroutine post_genealogies(nsamp)
   integer :: i,nsamp
   
   if (allocated(parass)) deallocate(parass)
   allocate(parass(no))
   
   do i=1,no
     parass(i)%oid=oid(i)
     call post_genealogy(i,nsamp,parass(i)%mo1,parass(i)%fa1,parass(i)%pr1, &
          parass(i)%mo2,parass(i)%fa2,parass(i)%pr2)
   enddo
   
   outfile = 90
   open(unit=outfile,file=fp,action='write',status='replace')
   write(outfile,"(a1,7(a10))") ' ','Prog','Mo1','Fa1','Pr1','Mo2','Fa2','Pr2'   
   do i=1,no
     write(outfile,"(i10,2(2(i10),f10.4))") parass(i)%oid,parass(i)%mo1,parass(i)%fa1,parass(i)%pr1, &
          parass(i)%mo2,parass(i)%fa2,parass(i)%pr2
     flush(outfile)
   enddo
   close(outfile)
   
 end subroutine post_genealogies
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes limits of the highest posterior density interval as the shortest range containing
 ! (1-p)*n elements
 subroutine hdi(x,n,p,xl,xu)
   integer, intent(in) :: n
   double precision, intent(inout) :: x(n)
   double precision, intent(in) :: p
   double precision, intent(inout) :: xl,xu
   integer :: i,maxi,r
   double precision :: dul
   
   maxi=int(n*(1.D0-p))
   r=int(n*p)
   call sort(x)
   xu=x(r)
   xl=x(1)
   dul=xu-xl
   do i=1,maxi
     if (x(i+r-1)-x(i)<dul) then
       xu=x(i+r-1)
       xl=x(i)
       dul=xu-xl
     endif
   enddo
   
 end subroutine hdi
 !----------------------------------------------------------------------------------------------------------
  
 !----------------------------------------------------------------------------------------------------------
 ! This routine computes a statistical support for the model  WAIC (lppd)
 subroutine compute_lppd(lppd,pw,ns)
   double precision, intent(out) :: lppd,pw
   integer :: ns,i,j
   double precision :: vi,mi
   
   lppd=0.D0
   pw=0.D0
   do i=1,no
     mi=0.D0
     vi=0.D0
     do j=1,ns
       mi=mi+lproi(i,j)/dble(ns)
       vi=vi+dlog(lproi(i,j))/dble(ns)
     enddo
     lppd=lppd+dlog(mi)
     !pw=pw+2.D0*(dlog(mi)-vi) !alternative definition of pWAIC (the default expression for pw needs to be commented out)
     mi=vi
     vi=0.D0
     do j=1,ns
       vi=vi+(dlog(lproi(i,j))-mi)**2/dble(ns-1)
     enddo
     pw=pw+vi !use this definition by default
   enddo
 end subroutine compute_lppd
 !----------------------------------------------------------------------------------------------------------
  
 !----------------------------------------------------------------------------------------------------------
 ! This routine sorts real-type vector of length n
 subroutine sort(ra)
   integer :: n
   double precision, intent(inout) :: ra(:)
   integer :: l,j,ir,i
   double precision :: rra
   
   n = size(ra)
   
   l = (n / 2) + 1
   ir = n

   do
     if (l > 1) then
       l = l - 1
       rra = ra(l)
     else
       rra = ra(ir)
       ra(ir) = ra(1)
       ir = ir - 1
       if (ir == 1) then
         ra(1) = rra
         exit
       endif
     endif
   
     i = l
     j = 2 * l
     do while (j <= ir)
       if (j < ir) then
         if (ra(j) < ra(j+1)) then
           j = j + 1
         endif
       endif
     
       if (rra < ra(j)) then
         ra(i) = ra(j)
         i = j
         j = 2 * j
       else
         j = ir + 1 
       endif
     enddo
     ra(i) = rra
   enddo
 end subroutine sort
 !----------------------------------------------------------------------------------------------------------
 
 !----------------------------------------------------------------------------------------------------------
 ! This routine is similar to the routine "hdi()" but is designed specifically for a circular variable
 ! such as the prevailing dispersal direction
 subroutine circular_estims(x,n,p,mq,lq,hq)
   integer i,j,mi,n
   double precision :: x(n),d(n),p,devi,mdev,xi,di
   double precision, intent(inout) :: mq,lq,hq
   
   !determination of median angle as an angle with minimum absolute difference to any angle:
   x=x*my2pi
   mi=1
   mdev=0.D0
   do i=1,n
     devi=0.D0
     do j=1,n
       !minimum distance between two angles (clockwise vs. counter-clockwise)
       devi=devi+min(my2pi-dabs(x(j)-x(i)),dabs(x(j)-x(i)))
     enddo
     devi=devi/dble(n)
     if (i==1) then
       mdev=devi
     else if (devi<mdev) then
       mdev=devi
       mi=i
     endif
   enddo
   
   do i=1,n
     d(i)=x(i)-x(mi)
     if (d(i)<-mypi) d(i)=my2pi+d(i)
     if (d(i)>mypi) d(i)=d(i)-my2pi
   enddo
   
   do i=1,n
     do j=i+1,n
       if (d(i)>d(j)) then
         xi=x(i)
         di=d(i)
         x(i)=x(j)
         d(i)=d(j)
         x(j)=xi
         d(j)=di
       endif
     enddo
   enddo
   
   mq=x(n/2)
   call hdi(d,n,p,lq,hq)
   lq=mq+lq
   hq=mq+hq
 end subroutine
 !----------------------------------------------------------------------------------------------------------
 
end program nmpi2
